/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define ASM_FILE

#include "shared.h"

#if 0
	.file	"asm.S"
#endif

	.text

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

ENTRY(start)
lstart:
	/*
	 *  "start" should be loaded at 0x0:$STARTADDR.  This next
	 *  is to guarantee that.
	 */

	/* ljmp $0, $codestart */
	.byte	0xea
#ifndef	PC9800
	.word	0x8070, 0
#else
	.word	0x8080, 0
#endif

	/* padding */
	.byte	0

	/*
	 *  Compatibility version number
	 *
	 *  These MUST be at byte offset 6 and 7 of the executable
	 *
	 *  DO NOT MOVE !!!
	 */
	.byte	COMPAT_VERSION_MAJOR, COMPAT_VERSION_MINOR

	/*
	 *  This is a special data area 8 bytes from the beginning.
	 */

	. = EXT_C(start) + 0x8

VARIABLE(install_partition)
	.long	0x020000
VARIABLE(stage2_size)
	/* GAS 2.13.90.0.10 rejects ``_edata - EXT_C(start)'' for
	   unknown reason.  We work around this by defining a local symbol.  */
	.long	_edata - lstart
VARIABLE(version_string)
#ifdef PC9800
	.string	"0.5.98.3"
#else
	.string "0.5"
#endif
VARIABLE(config_file)
#ifndef CONFIG_FILE_ASM
	.string "/boot/grub/menu.lst"
#else   /* CONFIG_FILE_ASM */
	CONFIG_FILE_ASM
#endif  /* CONFIG_FILE_ASM */

	/*
	 *  Leave some breathing room for the config file name.
	 */

	. = EXT_C(start) + 0x70

#ifdef	PC9800
	/*
	 *  Leave more room for some BIOS-like parameters.
	 */
VARIABLE(vram)
	.long	0x000A0000

	/*
	 *  The putchar_*'s order is important; some codes handle
	 *  these variables at once by long-word access.
	 */
VARIABLE(putchar_y)
	.byte	0
VARIABLE(putchar_x)
	.byte	0
VARIABLE(putchar_flags)
	.byte	0
VARIABLE(putchar_save)
	.byte	0
VARIABLE(putchar_attribute)
	.byte	0xE1		/* normal white face */
VARIABLE(force_check_disk_change)
	.byte	0
#ifdef DEBUG
VARIABLE(debug_flags)
	.byte	0x00		/* all debug features are disabled */
#endif

	. = EXT_C(start) + 0x80
#endif

/* the code continues... */
codestart:
#ifndef	PC9800
	cli		/* we're not safe here! */

	/* set up %ds, %ss, and %es */
	xorw	%ax, %ax
	movw	%ax, %ds
	movw	%ax, %es

	/* set up the real mode/BIOS stack */
	movw	%ax, %ss
	movl	$STACKOFF, %esp
	movl	$STACKOFF, %ebp

	sti		/* we're safe again */

	/* save boot drive reference */
	movb	%dl, EXT_C(boot_drive)

	/* reset disk system (%ah = 0) */
	int	$13
#else
	/* set up %ds, %ss, and %es */
	xorw	%bx, %bx		/* %al holds boot device information */
	movw	%bx, %ds
	movw	%bx, %es

	/* set up the real mode/BIOS stack */
	movw	%bx, %ss
	movl	$ STACKOFF, %esp
	movl	%esp, %ebp

	movb	%al, DISK_BOOT

	/*
	 * Determine boot partition in case we are booted by MBR-loader.
	 * There is no magic; so strict checking.
	 */
	cmpb	$0xFF, EXT_C(install_partition)+2
	jne	1f
	testb	$0x50, %al
	jnz	1f		/* not hard disk */
#if 0
	testb	$0x80, %al
	jz	1f		/* not CHS access */
#endif
	xorw	%bx, %bx
	movb	$0x84, %ah
	int	$0x1B		/* get sector size in %bx */
	xchgw	%si, %ax
	testb	$0x1F, %al
	jnz	1f		/* invalid partition table offset */
	subw	%bx, %ax
	jb	1f		/* invalid partition table offset */
	cmpw	%bx, %ax
	jae	1f		/* invalid partition table offset */
	shrw	$5, %ax
	movb	%al, EXT_C(install_partition) + 2
   1:	/*
	 * Adjust VRAM address for high-resolution mode.
	 */
	testb	$8, BIOS_FLAG
	jz	1f		/* normal mode */
	movb	$0x0E, EXT_C(vram) + 2
    1:	/*
	 * Other BIOS initializations will go here
	 */
#if 0	/* screen mode initialization is now done in cls() */
	movw	$0x0A00, %ax
	int	$0x18		/* initialize text display mode */
#endif
#endif	/* PC9800 */

	/* transition to protected mode */
	call EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	/*
	 *  Call the start of main body of C code, which does some
	 *  of it's own initialization before transferring to "cmain".
	 */
	call EXT_C(init_bios_info)


#ifdef PC9800
ENTRY(panic)
	movb	$0xC1, EXT_C(putchar_attribute)
	pushl	$ L(panic_string)
	call	EXT_C(printf)
	popl	%eax
	popl	%eax			/* discard return address */
	call	EXT_C(printf)
	/* fall through */

	.section .rodata
L(panic_string):
	.string	"panic:	"
	.previous
#endif

/*
 *  This call is special...  it never returns...  in fact it should simply
 *  hang at this point!
 */

ENTRY(stop)
	call	EXT_C(prot_to_real)

	/*
	 * This next part is sort of evil.  It takes advantage of the
	 * byte ordering on the x86 to work in either 16-bit or 32-bit
	 * mode, so think about it before changing it.
	 */

#ifdef	PC9800
	movb	$0x17, %ah
	int	$0x18			/* beep on */
#endif
ENTRY(hard_stop)
	hlt
	jmp EXT_C(hard_stop)


#ifndef	PC9800
/*
 * stop_floppy()
 *
 * Stops the floppy drive from spinning, so that other software is
 * jumped to with a known state.
 */
ENTRY(stop_floppy)
	movw	$0x3F2, %dx
	xorb	%al, %al
	outb	%al, %dx
	ret
#endif

#ifndef NO_FANCY_STUFF

#ifndef PC9800
/*
 * chain_stage1(segment, offset, part_table_addr)
 *
 *  This starts another stage1 loader, at segment:offset.
 */
#else
/*
 * chain_stage1()
 *
 *  On PC-9800 this function takes no argument. Segment and offset are
 *  computed internally.
 */
#endif

ENTRY(chain_stage1)
#ifndef PC9800
	/* no need to save anything, just use %esp */

	/* store %ESI, presuming %ES is 0 */
	movl	0xc(%esp), %esi

#if 0
	/* store new offset */
	movl	0x8(%esp), %eax
	movl	%eax, offset

	/* store new segment */
	movw	0x4(%esp), %ax
	movw	%ax, segment
#else
	movl	8(%esp), %edi
	movl	4(%esp), %ecx
#endif

	/* set up to pass boot drive */
	movb	EXT_C(boot_drive), %dl

#else /* PC9800 */

	movl	EXT_C(saved_drive), %eax
	movb	%al, %ch	/* fallback geometry */
	movb	$9, %cl
	call	EXT_C(get_geometry)
	testl	%eax, %eax	/* == NULL ? */
	jz	1f
	movl	(%eax), %ecx
    1:	xorl	%ebx, %ebx	/* %ebx = 0 */
	incl	%ebx		/* %ebx = 1 */
	shll	%cl, %ebx	/* %ebx = sector size */
	movzbl	EXT_C(saved_partition) + 2, %esi	/* get slice # */
	testb	$0x50, %ch
	jnz	2f		/* -> not hard disk (ie. floppy) */
	/* hard disk */
	shll	$5, %esi
	addl	%ebx, %esi	/* %esi is boot partition information */
	orb	$0x80, %ch	/* force C/H/S addressing */
	addl	%ebx, %ebx	/* IPL is 2 sectors */
	/*
	 *  Note: %si got junk value if saved_partition indicates 
	 *  `entire disk', but it is ok because booted program can
	 *  determine whether it should use %si.
	 */
    2:
#endif /* PC9800 */

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

#ifndef PC9800
#if 0
	ljmp	(offset)
#else
	pushw	%cx
	pushw	%di
	lret
#endif
#else /* PC9800 */

	movw	$0x03F0, %sp		/* `standard' stack area (%ss = 0) */
	decw	%ax			/* %ax = 0xFFFF */
	movw	%ax, %ds		/* %ds = 0xFFFF */
	movw	%bx, %ax
	shrw	$4, %ax			/* to segment-based */
	negw	%ax
	addw	$0x2000, %ax
	cs; testb $0x08, BIOS_FLAG
	jz	1f
	addw	$0x2000, %ax		/* high-resolution mode/machine */
    1:	pushw	%ax			/* for last `lret' */
	movw	%ax, %es
	movb	%ch, %dl		/* %dl = boot device */
	pushw	%si
	movw	$0x0010, %si		/* %ds:%si = 0xFFFF:0x0010 (HMA) */
	xorw	%di, %di		/* %es:%di = IPL address */
	movw	%bx, %cx
	shrw	%cx
	cld
	rep; movsw
	popw	%si			/* %si = boot partition information */
	movw	%cx, %ds		/* %ds = 0 */
	movb	$0x03, %al
	outb	%al, $0xF6		/* wire A20 line to 0 */
	xchgw	%ax, %dx		/* %al = boot device */
	movb	%al, DISK_BOOT
	pushw	%cx			/* pushw $0 */
	xorw	%bp, %bp
	lret

#endif /* PC9800 */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

#endif /* NO_FANCY_STUFF */

#if 0	/* inlined by K.Takai */
/*
 * chain_stage2(segment, offset)
 *
 *  This starts another stage2 loader, at segment:offset.  It presumes
 *  that the other one starts with this same "asm.S" file, and passes
 *  parameters by writing the embedded install variables.
 */

ENTRY(chain_stage2)
	/* no need to save anything, just use %esp */

#if 0
	/* store new offset */
	movl	0x8(%esp), %eax
	movl	%eax, offset
	movl	%eax, %ebx

	/* store new segment */
	movw	0x4(%esp), %ax
	movw	%ax, segment
	shll	$4, %eax

	/* generate linear address */
	addl	%eax, %ebx
#else
	movl	8(%esp), %edi
	movl	4(%esp), %ecx
	shll	%ecx			/* save 1 byte... */
	leal	(%edi,%ecx,8), %ebx
#endif

	/* store install info */
	movl	EXT_C(install_partition), %eax
	movl	%eax, EXT_C(install_partition)-EXT_C(start)(%ebx)

#ifndef PC9800
	/* set up to pass boot drive */
	movb	EXT_C(boot_drive), %dl
#endif

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

#ifdef PC9800
	movb	DISK_BOOT, %al
#endif
#if 0
	ljmp	(offset)
#else
	pushw	%cx
	pushw	%di
	lret
#endif

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32
#endif /* #if 0 */

/*
 *  These next two routines, "real_to_prot" and "prot_to_real" are structured
 *  in a very specific way.  Be very careful when changing them.
 *
 *  NOTE:  Use of either one messes up %eax and %ebp.
 */

ENTRY(real_to_prot)
	cli

	/* load the GDT register */
	addr32
	data32
	lgdt	gdtdesc

	/* turn on protected mode */
#if 1
	movl	%cr0, %eax
#if 0
	data32
	orl	$CR0_PE_ON, %eax
#else
	orb	$CR0_PE_ON, %al
#endif
	movl	%eax, %cr0
#else
	.code16

	movw	$CR0_PE_ON, %ax	
	lmsw	%ax

	.code32
#endif

	/* jump to relocation, flush prefetch queue, and reload %cs */
	data32
	ljmp	$PROT_MODE_CSEG, $1f
  1:	/* reload other segment registers */
	movl	$PROT_MODE_DSEG, %eax
	/* GNU assember (2.8.1) complains about `movl r/m32, seg'... */
	.byte	0x8E, MODRM(3,3,0)	/* movl	%eax, %ds */
	.byte	0x8E, MODRM(3,0,0)	/* movl %eax, %es */
	.byte	0x8E, MODRM(3,4,0)	/* movl	%eax, %fs */
	.byte	0x8E, MODRM(3,5,0)	/* movl	%eax, %gs */
	.byte	0x8E, MODRM(3,2,0)	/* movl	%eax, %ss */

#if 0
	/* put the return address in a known safe location */
	movl	(%esp), %eax
	movl	%eax, STACKOFF

	/* get protected mode stack */
	movl	protstack, %eax
	movl	%eax, %esp
	movl	%eax, %ebp

	/* get return address onto the right stack */
	movl	STACKOFF, %eax
	movl	%eax, (%esp)
#else
	/* save return address to register */
	xorl	%eax,%eax
	popw	%ax

	/* get protected mode stack */
	movl	protstack, %esp
	movl	%esp, %ebp

	/* restore return address to stack */
	pushl	%eax
#endif

	/* zero %eax */
	xorl	%eax, %eax

	/* return on the old (or initialized) stack! */
	ret


ENTRY(prot_to_real)
	/* just in case, set GDT */
	lgdt	gdtdesc

	/* get the return address */
	popl	%eax

	/* save the protected mode stack */
	movl	%esp, protstack

	/* set up new stack */
	movl	$STACKOFF, %esp
	pushw	%ax
	movl	%esp, %ebp

	/* set up segment limits */
	movl	$PSEUDO_RM_DSEG, %eax
	/* GNU assember (2.8.1) complains about `movl r/m32, seg'... */
	.byte	0x8E, MODRM(3,3,0)	/* movl	%eax, %ds */
	.byte	0x8E, MODRM(3,0,0)	/* movl %eax, %es */
	.byte	0x8E, MODRM(3,4,0)	/* movl	%eax, %fs */
	.byte	0x8E, MODRM(3,5,0)	/* movl	%eax, %gs */
	.byte	0x8E, MODRM(3,2,0)	/* movl	%eax, %ss */

	/* this might be an extra step */
	ljmp	$PSEUDO_RM_CSEG, $1f		/* jump to a 16 bit segment */

   1:	/* clear the PE bit of CR0 */
	movl	%cr0, %eax
#if 0
	data32
	andl 	$CR0_PE_OFF, %eax
#else
	andb	$CR0_PE_OFF, %al
#endif
	movl	%eax, %cr0

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	/* flush prefetch queue, reload %cs */
	ljmp	$0, $1f

   1:	/* we are in real mode now
	 * set up the real mode segment registers : DS, SS, ES
	 */
	/* zero %eax */
	xorl	%eax, %eax

	movw	%ax, %ds
	movw	%ax, %es
	movw	%ax, %fs
	movw	%ax, %gs
	movw	%ax, %ss

	/* restore interrupts */
	sti

	/* return on new stack! */
	ret

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32


/*
 * biosdisk(subfunc, drive, geometry, sector, nsec, segment)
 *	Read/write "nsec" sectors from disk to real segment "segment"
 *		offset zero
 *
 *   If it will fit into the BIOS geometry, it tries the INT 0x13
 *   AH=<subfunction> interface, else if it is a hard disk, it will
 *   try the hard disk LBA calls (not yet implemented).  If these
 *   won't work, or otherwise it is out of the translated area, then
 *   it returns BIOS_GEOMETRY_ERROR.
 */

ENTRY(biosdisk)
#ifndef	PC9800
	push	%ebp
	mov	%esp, %ebp

	push	%ebx
	push	%ecx
	push	%edx
	push	%esi

	/*
	 * "geometry" is a longword representing the BIOS geometry:
	 *       6 bit zero
	 *	10 bit cylinder   (bytes 2 & 3)
	 *	 8 bit head       (byte 1)
	 *	 8 bit sector     (byte 0)
	 */

	/* set up original sector number */
	xorl	%edx, %edx
	movl	0x14(%ebp), %eax

	/* get sector offset, place in %ecx */
	xorl	%ebx, %ebx
	movb	0x10(%ebp), %bl
	divl	%ebx
	movl	%edx, %ecx

	/* get track offset (head number) in %edx,
	   and cylinder offset in %eax */
	xorl	%edx, %edx
	xorl	%ebx, %ebx
	movb	0x11(%ebp), %bl
	inc	%ebx
	divl	%ebx

	/* check cylinder offset, is there a geometry problem here? */
	movl	0x10(%ebp), %ebx
	shrl	$16, %ebx
	cmpl	%ebx, %eax

	/* if not, go onto standard read function */
	jle	disk_use_standard_bios

	/* XXX else we better use LBA or generate a geometry error */
	movl	$BIOSDISK_ERROR_GEOMETRY, %eax
	jmp	disk_exit_32

	/*
	 *  This portion implements the BIOS standardized
	 *  INT 0x13/AH=<subfunc> interface.
	 */
disk_use_standard_bios:

	shll	$8, %edx		/* get head number to %dh */
	xchgl	%eax, %ecx		/* cylinder to %cx, sector to %al */
	/* cylinder; the highest 2 bits of cyl is in %cl */
	xchgb	%ch, %cl
	rorb	$2, %cl
	incb	%al			/* sector; sec starts from 1, not 0 */
	orb	%al, %cl
	movb	0xc(%ebp), %dl		/* drive */

	/* prot_to_real will set %es to 0, so must set it ourselves
	   after it is called */
	movb	0x18(%ebp), %bl		/* number of sectors */
	movb	0x8(%ebp), %bh		/* bios disk subfunction */
	shll	$16, %ebx		/* shift num sect. and subfunction */
	movw	0x1c(%ebp), %bx		/* segment */

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movw	%bx, %es		/* load segment */
	/*
	 *  Shift num sectors and subfunction back
	 */
	shrl	$16, %ebx
	pushw	%bx			/* save num sectors */
	xorw	%bx, %bx

	movw	$3, %si

disk_loop:
	popw	%ax			/* restore num sectors */
	pushw	%ax			/* save num sectors */

	/* BIOS call for reading/writing */
	int	$0x13

	/* set success return value */
	movb	$0, %bl

	/* did we actually succeed? */
	jnc	disk_exit_16

	/* do we try again? */
	decw	%si
	cmpw	$0, %si

	/* if this isn't the third try, go again */
	jne	disk_loop

	/* save return value */
	movb	%ah, %bl

disk_exit_16:
	call	EXT_C(real_to_prot)	/* back to protected mode */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	movb	%bl, %al		/* return value in %eax */

disk_exit_32:
	pop	%esi
	pop	%edx
	pop	%ecx
	pop	%ebx
	pop	%ebp
	ret

#else	/* PC9800 */

	pushl	%ebp
	movl	%esp, %ebp
	pushl	%esi
	pushl	%ebx

	movl	12(%ebp), %esi		/* get disk_geometry structure */
	movl	16(%ebp), %eax		/* get sector # */
	movl	20(%ebp), %ebx		/* get # of sectors to read */

#ifdef DEBUG
	testb	$ DEBUG_BIOSDISK, EXT_C(debug_flags)
	jz	0f
	movl	24(%ebp), %edx		/* segment */
	pushl	%eax			/* save nsec */
	pushl	%edx			/*  1 */
	pushl	%ebx			/*  2 */
	pushl	%eax			/*  3 */
	movl	4(%esi), %eax
	movzbl	3(%esi), %edx
	pushl	%eax			/*  4 */
	pushl	%edx			/*  5 */
	movzbl	2(%esi), %eax
	movzbl	1(%esi), %edx
	movzbl	(%esi), %ecx
	pushl	%eax			/*  6 */
	pushl	%edx			/*  7 */
	movl	8(%ebp), %eax		/* drive */
	pushl	%ecx			/*  8 */
	pushl	%esi			/*  9 */
	pushl	%eax			/* 10 */
	pushl	$1f			/* 11 */
	call	EXT_C(dprintf)
	addl	$11 * 4, %esp
	popl	%eax			/* restore nsec */
	.section .rodata
    1:	.string	"biosdisk(0x%2x, 0x%8x { %u, 0x%2x, %u, %u, %u }, %u, %u, 0x%4x)\n"
	.previous
    0:
#endif

	movl	%eax, %edx		/* save sector # */
	addl	%ebx, %eax		/* add # of sectors to read */
	jc	L(b_geometry_error)
	cmpl	4(%esi), %eax		/* exceed total # of sectors? */
	ja	L(b_geometry_error)
	shll	$16, %ebx
	movl	(%esi), %ecx		/* %cl = log2 of sector size */
	shll	%cl, %ebx		/* %ebx(H) = transfer length */
	movb	8(%ebp), %bh		/* get subfunction */
	movb	%ch, %bl		/* DA/UA */
	movl	24(%ebp), %esi		/* get segment address */

#ifdef DEBUG
	testb	$ DEBUG_BIOSDISK, EXT_C(debug_flags)
	jz	0f
	movb	$'\n', %al
	call	EXT_C(putchar)
	call	EXT_C(putchar)
    0:
#endif

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movw	%si, %es
	movw	%bx, %ax
	testb	$0x50, %al		/* hard disk ? */
	jz	L(b_harddisk)		/* yes */
	testb	$0x10, %al
	jz	L(b_setup_lba)
	andb	$0x70, %al
	cmpb	$0x50, %al
	je	L(b_setup_lba)
	orb	$0xD0,%bh		/* MFM, seek, retry */
	rorl	$16, %ecx		/* %cl = #sector / %ch = #head */
	movw	%dx, %ax
	shrl	$16, %edx		/* %dx:%ax is sector # */
	movzbw	%cl, %si
	divw	%si			/* %dx = sector # */
	incw	%dx			/* sector # starts from 1 */
	divb	%ch			/* %al = cylinder #, %ah = head # */
	movb	%ah, %dh		/* %dl = sector #, %dh = head # */
	shrl	$8, %ecx		/* %ch = log2 of sector size */
	subb	$7, %ch
	movb	%al, %cl		/* %cl = cylinder #, %ch = N for FDC */
	jmp	L(b_final_setup)
L(b_harddisk):	
	andb	$0x7F, %bl		/* force LBA addressing */
	testb	$0x70, %al
	jz	L(b_setup_lba)		/* IDE disk */
	movw	%bx, %si
	andw	$0x0007, %si		/* %si = target SCSI ID */
	subb	$8, %cl			/* %cl = block-size code for BIOS */
	shlw	$2, %si
	addw	$(SCSI_PARAM + 3), %si
	movb	(%si), %al
	pushw	%ax
	shlb	$4, %cl			/* modify block-size field */
	andb	$0xCF, %al
	orb	%cl, %al
	movb	%al, (%si)
L(b_setup_lba):
	movw	%dx, %cx		/* %dx:%cx is LBA */
	shrl	$16, %edx
L(b_final_setup):
	movw	%bx, %ax		/* %ah = function #, %al = DA/UA */
	shrl	$16, %ebx		/* %bx = transfer length */
	xorw	%bp, %bp		/* %bp = transfer address offset */

#ifdef DEBUG
	testb	$ DEBUG_BIOSDISK, EXT_C(debug_flags)
	jz	0f
	call	L(real_register_dump)
    0:
#endif

	int	$0x1B			/* go! */

	movzbl	%ah, %ebx		/* status */
	andb	$0x70, %al
	cmpb	$0x20, %al		/* SCSI disk */
	jne	4f
	popw	%ax
	movb	%al, (%si)		/* restore BIOS work area */
    4:

	call	EXT_C(real_to_prot)	/* back to protected mode */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32
	
	movl	%ebx, %eax
    5:	popl	%ebx
	popl	%esi
	popl	%ebp

#ifdef DEBUG

	testb	$ DEBUG_BIOSDISK, EXT_C(debug_flags)
	jnz	1f

ENTRY(print_disk_bios_error)
	testb	$ DEBUG_BIOS_ERROR, EXT_C(debug_flags)
	jz	0f
	testl	%eax, %eax
	jz	0f
    1:	movl	%eax, %ecx
	shrl	$3, %eax
	xorl	%edx, %edx
	cmpl	$0x100 >> 3, %eax
	ja	1f
	movb	L(bios_error_list)(%eax), %dl
    1:	leal	L(bios_error_UN)(,%edx,2), %edx
	pushl	%edx		/* push error string */
	pushl	%ecx		/* push error code */
	pushl	$1f		/* push format string */
	call	EXT_C(dprintf)
	popl	%ecx		/* pop off format string */
	popl	%eax		/* restore error code */
	popl	%edx		/* pop off error string */

	.section .rodata

    1:	.string	"Disk BIOS status = 0x%2x (%s)\n"

#define	_(mnemonic, str) .balign 2,0; L(bios_error_ ## mnemonic): .string str

	_(UN, "Unknown");		_(NT, "Normal Termination")
	_(DB, "DMA Boundary");		_(EN, "End of Cylinder")
	_(IA, "Illegal Disk Address");	_(EC, "Equipment Check")
	_(OR, "Overrun");		_(NR, "Not Ready")
	_(UA, "Unit Attension");	_(NW, "Not Writable")
	_(ER, "Error");		_(DA, "Direct Access an Alternate Track")
	_(TO, "Time Out");		_(DL, "Defect List Error")
	_(DE, "Data Error in ID");	_(NS, "No Defect Spare Location")
	_(DD, "Data Error in Data");	_(ND, "No Data")
	_(BC, "Bad Cylinder");		_(AT, "Alternate Track Read Error")
	_(SE, "Seek Error");		_(MA, "Missing Address Mark in ID")
	_(MD, "Missing Address Mark in Data")
	_(GE, "Geometry Translation Error")

#undef _
#define	_(mnemonic) ((L(bios_error_ ## mnemonic)-L(bios_error_UN)) >> 1)

L(bios_error_list):
	/*	 +00	+08    +10    +18    +20    +28    +30	  +38 */
/* 00 */ .byte	_(NT), _(NT), _(UN), _(UN), _(DB), _(DB), _(EN), _(IA)
/* 40 */ .byte	_(EC), _(EC), _(OR), _(OR), _(NR), _(UA), _(NW), _(NW)
/* 80 */ .byte	_(ER), _(DA), _(TO), _(DL), _(DE), _(NS), _(DD), _(DD)
/* C0 */ .byte	_(ND), _(SE), _(BC), _(AT), _(MA), _(MA), _(MD), _(MD)
	 .byte	_(GE)

#undef _

	.previous
    0:
#endif /* DEBUG */
	ret

#endif	/* PC9800 */

L(b_geometry_error):
	movl	$ BIOSDISK_ERROR_GEOMETRY, %eax
	jmp	5b
	
#ifndef	PC9800
/*
 *
 * get_diskinfo(drive):  return a word that represents the
 *	max number of sectors and heads and cylinders for this drive
 *
 */

ENTRY(get_diskinfo)
	pushl	%ebp
	movl	%esp, %ebp
	pushl	%ebx
	pushl	%ecx
	pushl	%edx
	pushl	%edi
	pushl	%esi

	movw	0x8(%ebp), %dx		/* diskinfo(drive #) */

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movb	%dl, %al
	andb	$0x80, %al
	jnz	hard_drive

/*
 *  Perform floppy probe!
 */

	movl	$probe_values-1, %esi

probe_loop:
	/* reset floppy controller INT 13h AH=0 */
	xorw	%ax, %ax
	int	$0x13

	incw	%si

	/* movb	  (%si), %cl */
	.byte	0x8A, 0x0C

	/* if number of sectors is 0, display error and die */
	cmpb	$0, %cl

	je	probe_failed

	/* perform read */
	movw	$SCRATCHSEG, %ax
	movw	%ax, %es
	xorw	%bx, %bx
	movw	$0x201, %ax
	movb	$0, %ch
	movb	$0, %dh
	int	$0x13

	jc	probe_loop

	/* %cl is already the correct value! */
	movb	$1, %dh
	movb	$79, %ch

	jmp	probe_success

probe_values:
	.byte	36, 18, 15, 9, 0

hard_drive:
	movb	$0x8, %ah		/* ask for disk info */
	int	$0x13

	jc	probe_failed

	/* es:di = parameter table */
	/* carry = 0 */

probe_success:
	/*
	 * form a longword representing all this gunk:
	 *       6 bit zero
	 *	10 bit cylinder
	 *	 8 bit head
	 *	 8 bit sector
	 */
	movb	%cl, %al		/* Upper two bits of cylinder count */
	andl	$192,%eax	
	leal	0(,%eax,4),%eax		/* << 2 */
	movb	%ch, %al		/* Lower 8 bits */
	sall	$16,%eax		/* << 16 */
	movb	%dh, %ah		/* max head */
	andb	$0x3f, %cl		/* mask of cylinder gunk */
	movb	%cl, %al		/* max sector (and # sectors) */

	movl	%eax, %ebx		/* save return value */
	jmp	got_drive

probe_failed:
	/*
	 * Urk.  Call failed.  It is not supported for floppies by old
	 * BIOSes, but it should work for all hard drives!!
	 *
	 * Return a 0 here...  presume there is no drive present.  ????
	 */

	movl	$0, %ebx		/* not present value */

got_drive:
	call	EXT_C(real_to_prot)	/* back to protected mode */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	/* set up return in correct register */
	movl	%ebx, %eax

	popl	%esi
	popl	%edi
	popl	%edx
	popl	%ecx
	popl	%ebx
	popl	%ebp

	ret

#else	/* PC9800 */
/*
 * int get_bios_geometry(struct disk_geometry *geom): Inquire to BIOS
 *	 about geometry of a hard disk. (REGPARM)
 */
ENTRY(get_bios_geometry)
	pushl	%ebp
	pushl	%ebx
	pushl	%eax
	movl	(%eax), %ecx		/* get DA/UA */
	xorl	%ebx, %ebx

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movb	%ch, %al

#if 0
	testb	$0x70, %al
	jnz	1f
#endif
	movb	$0x8E, %ah		/* ??? */
	int	$0x1B
    1:
	movb	$0x84, %ah
	int	$0x1B

	call	EXT_C(real_to_prot)	/* back to protected mode */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	orl	%ebx, %eax		/* %eax was zeroed in real_to_prot */
	popl	%ebx
	jz	1f			/* probe failed... */
	testw	$0xF0FF, %ax
	jnz	1f
	movb	$7, %al
    2:	incb	%al
	shrb	%ah
	jnc	2b
	jnz	1f			/* sector size is not power of 2 */
	movb	%al, (%ebx)
	movl	%edx, %eax
	movw	%ax, 2(%ebx)		/* # of sectors/track, # of heads */
	mulb	%ah
	mulw	%cx			/* %dx:%ax = total # of sectors */
	orb	$0x80, 1(%ebx)
	xorl	%ecx, %ecx
	jmp	3f
    1:	movl	$-1, %ecx
	movl	%ecx, %eax
	movl	%ecx, %edx
    3:	movw	%ax, 4(%ebx)
	movw	%dx, 6(%ebx)
	movl	%ecx, %eax

	popl	%ebx
	popl	%ebp
	ret
	
/*
 * read_capacity(geom):	Do READ CAPACITY command to a SCSI device.
 */
ENTRY(read_capacity)
	pushl	%ebp
	pushl	%esi
	pushl	%ebx
	pushl	%eax
	movl	(%eax), %ecx		/* get DA/UA */

#ifdef DEBUG
	testb	$ DEBUG_GEOMETRY, EXT_C(debug_flags)
	jz	0f
	movb	$'\n', %al
	call	EXT_C(putchar)
	call	EXT_C(putchar)
    0:
#endif
	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	decw	%ax
	pushw	%ax			/* make room of 8 bytes */
	pushw	%ax			/* initialize to (-1, -1)	*/
	pushw	%ax			/* to distinguish the case that	*/
	pushw	%ax			/* this BIOS is not implemented */
	movb	%ch, %al
	movw	$ L(read_capacity_cdb), %si
	movw	$8, %bx
	movw	%sp, %bp
	movw	$10, %dx
	movb	$0xB0, %ah
#ifdef DEBUG
	testb	$ DEBUG_GEOMETRY, EXT_C(debug_flags)
	jz	0f
	call	L(real_register_dump)
    0:
#endif
	int	$0x1B
	popl	%ecx		/* # of blocks */
	popw	%si		/* block size (MSW) */
	popw	%dx		/* block size (LSW) */
	movzbl	%ah, %ebx	/* BIOS status */

	call	EXT_C(real_to_prot)	/* back to protected mode */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	decl	%eax		/* %eax == -1 */
	testl	%ebx, %ebx
	jnz	L(rc_failed)	/* BIOS call failed */

	xchgb	%cl, %ch	/* swap byte order in %ecx */
	roll	$16, %ecx
	xchgb	%cl, %ch

#ifdef DEBUG
	testb	$ DEBUG_GEOMETRY, EXT_C(debug_flags)
	jz	0f
	pushl	%eax
	movl	4(%esp), %eax
	pushl	%edx
	rolw	$8, %si
	pushw	%si
	xchgb	%dl, %dh
	pushw	%dx
	movzbl	1(%eax), %eax
	pushl	%ecx
	pushl	%eax
	pushl	$1f
	call	EXT_C(dprintf)
	popl	%eax
	popl	%edx
	popl	%ecx
	popl	%eax
	popl	%edx
	popl	%eax		/* = -1 */

	.section .rodata
    1:	.string	"read_capacity(0x%2x) = { %u, %u }\n"
	.previous
    0:
#endif	
	testw	%si, %si
	jnz	1f		/* block size >= 65536 */
	cmpw	$0x10, %dx
	jae	1f		/* block size not in 256 - 2048 bytes */
	movb	$7, %al
    3:	incb	%al
	shrb	%dl
	jnc	3b
	jz	4f		/* check whether block size is power of 2... */
    1:	xorl	%ebx, %ebx
	decl	%ebx		/* %ebx = -1 */
L(rc_failed):
	xorl	%ecx, %ecx
	decl	%ecx		/* %ecx = -1 */
	movb	$0xFF, %al
    4:	popl	%esi
	movb	%al, (%esi)
	movl	%ecx, 4(%esi)
	andb	$0x7F, 1(%esi)
	movl	%ebx, %eax	/* return value */
	
	popl	%ebx
	popl	%esi
	popl	%ebp
#ifdef DEBUG
	jmp	EXT_C(print_disk_bios_error)
#else
	ret
#endif

L(read_capacity_cdb):
	/* READ CAPACITY is 10-byte command */
	.byte	0x25, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

ENTRY(test_unit_ready)
	pushl	%ebp
	pushl	%esi
	pushl	%ebx
	movl	(%eax), %ecx		/* get DA/UA */

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movb	%ch, %al
	movw	$ L(test_unit_ready_cdb), %si
	xorw	%bx, %bx
	movw	$6, %dx
	movb	$0xB0, %ah
	int	$0x1B
	movzbl	%ah, %ebx	/* BIOS status */

	call	EXT_C(real_to_prot)	/* back to protected mode */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32
	movl	%ebx, %eax	/* return value */

	popl	%ebx
	popl	%esi
	popl	%ebp
#ifdef DEBUG
	jmp	EXT_C(print_disk_bios_error)
#else
	ret
#endif

L(test_unit_ready_cdb):
	/* TEST_UNIT_READY is 6-byte command */
	.byte	0x00, 0x00, 0x00, 0x00, 0x00, 0x00

/*
 *  void check_interrupt(void): check BIOS interrupt and process it if any.
 */
ENTRY(check_interrupt)
	xorl	%eax, %eax
	decl	%eax		/* %eax = -1 */
	/* fall through */

/*
 *  int simple_int1B(int ax): call "INT 1Bh". (REGPARM)
 */
ENTRY(simple_int1B)
	pushl	%ebp
	movl	%eax, %edx

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	testl	%edx, %edx
	js	1f
	xchgw	%dx, %ax
	int	$0x1B
	xchgw	%dx, %ax
	
    1:	call	EXT_C(real_to_prot)	/* back to protected mode */

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	movzbl	%dh, %eax
	popl	%ebp
	ret

#endif	/* PC9800 */		

/*
 * putchar(c)   : Puts character on the screen, interpreting '\n' as in the
 *                UNIX fashion.
 *
 * BIOS call "INT 10H Function 0Eh" to write character to console
 *      Call with       %ah = 0x0e
 *                      %al = character
 *                      %bh = page
 *                      %bl = foreground color ( graphics modes)
 */


ENTRY(putchar)
#ifndef	PC9800
	push	%ebp
	push	%eax
	push	%ebx

	movb	0x10(%esp), %bl

	/* if not '\n', just print the character */
	cmpb	$0xa, %bl
	jne	pc_notnewline

	/* if newline, print CR as well */
	pushl	$0xd
	call	EXT_C(putchar)
	popl	%eax

pc_notnewline:
	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movb	%bl, %al
	movb	$0xe, %ah
	movw	$1, %bx
	int	$0x10

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	pop	%ebx
	pop	%eax
	pop	%ebp
	ret
#else
	pushl	%eax
	pushl	%ecx
	pushl	%edx
	pushl	%edi
	pushl	%esi

	cld
	movl	$ EXT_C(putchar_y), %esi	/* to make addressing short */

	movb	$0, %ah
	movl	(%esi), %edx			/* %dl = row / %dh = column */
	movl	-4(%esi), %edi			/* vram */
	xchgb	%ah, 3(%esi)			/* putchar_save */
	testb	%ah, %ah
	jnz	L(pc_2byte)
	testb	%al, %al
	jns	L(pc_ascii)
	testb	$ PUTCHAR_SJIS, 2(%esi)		/* putchar_flags */
	jz	0f
	cmpb	$0xA0, %al
	jb	0f
	cmpb	$0xE0, %al
	jb	L(pc_1byte)
    0:	movb	%al, 3(%esi)
	jmp	L(pc_exit_1)
L(pc_2byte):
	testb	$ PUTCHAR_SJIS, 2(%esi)
	jnz	L(pc_shift_jis)
	cmpb	$0x8E, %ah		/* EUC Single Shift 2 (Katakana) */
	jne	2f
L(pc_1byte):
	movb	$0, %ah
	jmp	L(pc_write)
L(pc_ascii):
	cmpb	$'\n', %al
	je	L(pc_newline)
	cmpb	$'\t', %al
	jne	L(pc_1byte)
    1:	movb	$' ', %al
	call	EXT_C(putchar)
	testb	$7, 1(%esi)			/* putchar_x */
	jnz	1b
L(pc_exit_1):
	jmp	L(pc_exit)
L(pc_shift_jis):
	/*
	 * Convert Shift JIS (aka Microsoft Kanji Encoding Scheme) to JIS code
	 */
	addb	%ah, %ah
	subb	$0x1F, %al
	js	1f
	cmpb	$0x61, %al
	adcb	$0xDE, %al
    1:	addw	$0x1FA1, %ax
    2:	andw	$0x7F7F, %ax
	xchgb	%al, %ah
	subb	$0x20, %al
L(pc_write):
	xchgl	%eax, %edi
	/* %ax = (%dl<row> * 80 + %dh<column>) * 2 */
	movb	$80, %al
	mulb	%dl
	addb	%dh, %al
	adcb	$0, %ah
	addw	%ax, %ax
	/* high word of %eax is already VRAM address */
	xchgl	%eax, %edi
	movb	4(%esi), %cl		/* putchar_attribute */
	testb	%ah, %ah
	jz	L(pc_hankaku)
	cmpb	$0x08, %al
	jbe	1f
	cmpb	$0x0C, %al
	jb	L(pc_hankaku)
    1:	testb	$ PUTCHAR_XZENKAKU, 2(%esi)
	jnz	L(pc_exit)
	cmpb	$79, %dh
	jb	1f
	pushl	%eax
	movb	$' ', %al
	call	EXT_C(putchar)
	popl	%eax
	movl	(%esi),%edx
	jmp	L(pc_write)

    1:	stosw
	movb	%cl, 0x2000-2(%edi)
	incb	%dh
	orb	$0x80, %al
L(pc_hankaku):
	stosw
	movb	%cl, 0x2000-2(%edi)
	incb	%dh
	cmpb	$80, %dh
	jb	L(pc_finish)
L(pc_newline):
	incb	%dl
	movb	$0, %dh
	cmpb	$25, %dl
	jb	L(pc_finish)
	decb	%dl

	xorw	%di, %di		/* clear lower 16 bits */
	leal	160(%edi), %esi
	movl	$(80 * 24 * 2) >> 2, %ecx
	rep; movsl
	movl	$0x00200020, %eax
	movb	$(80 * 2) >> 2, %cl
	rep; stosl
	addl	$(0x2000 - 80 * 25 * 2), %edi
	addl	$(0x2000 - 80 * 25 * 2 + 80 * 2), %esi
	movw	$(80 * 24 * 2) >> 2, %cx
	rep; movsl
	movl	$0x00E100E1, %eax
	movb	$(80 * 2) >> 2, %cl
	rep; stosl
L(pc_finish):
	movw	%dx, EXT_C(putchar_y)
L(pc_exit):
	popl	%esi
	popl	%edi
	popl	%edx
	popl	%ecx
	popl	%eax
	ret
#endif

#ifndef	PC9800
/*
 *
 * get_memsize(i) :  return the memory size in KB. i == 0 for conventional
 *		memory, i == 1 for extended memory
 *	BIOS call "INT 12H" to get conventional memory size
 *	BIOS call "INT 15H, AH=88H" to get extended memory size
 *		Both have the return value in AX.
 *
 */

ENTRY(get_memsize)
	push	%ebp
	push	%ebx

	mov	0xc(%esp), %ebx

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	cmpb	$0x1, %bl
	je	xext
	
	int	$0x12
	jmp	xdone

xext:
	movb	$0x88, %ah
	int	$0x15

xdone:
	movw	%ax, %bx

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	movw	%bx, %ax
	pop	%ebx
	pop	%ebp
	ret

#endif	/* !PC9800 */

#ifndef NO_FANCY_STUFF

#ifndef	PC9800
/*
 *
 * get_eisamemsize() :  return packed EISA memory map, lower 16 bits is
 *		memory between 1M and 16M in 1K parts, upper 16 bits is
 *		memory above 16M in 64K parts.  If error, return -1.
 *	BIOS call "INT 15H, AH=E801H" to get EISA memory map,
 *		AX = memory between 1M and 16M in 1K parts.
 *		BX = memory above 16M in 64K parts.
 *
 */

ENTRY(get_eisamemsize)
	push	%ebp
	push	%ebx
	push	%ecx
	push	%edx

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movw	$0xe801, %ax
	int	$0x15

	shll	$16, %ebx
	movw	%ax, %bx

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	movl	$0xFFFFFFFF, %eax
	cmpb	$0x86, %bh
	je	xnoteisa

	movl	%ebx, %eax

xnoteisa:
	pop	%edx
	pop	%ecx
	pop	%ebx
	pop	%ebp
	ret


/*
 *
 * get_mem_map(addr, cont) :  address and old continuation value (zero to
 *		start), for the Query System Address Map BIOS call.
 *
 *  Sets the first 4-byte int value of "addr" to the size returned by
 *  the call.  If the call fails, sets it to zero.
 *
 *	Returns:  new (non-zero) continuation value, 0 if done.
 *
 * NOTE: Currently hard-coded for a maximum buffer length of 1024.
 */

ENTRY(get_mem_map)
	push	%ebp
	push	%ebx
	push	%ecx
	push	%edx
	push	%edi
	push	%esi

	/* place address (+4) in ES:DI */
	movl	0x1c(%esp), %eax
	addl	$4, %eax
	movl	%eax, %edi
	andl	$0xf, %edi
	shrl	$4, %eax
	movl	%eax, %esi

	/* set continuation value */
	movl	0x20(%esp), %ebx

	/* set default maximum buffer size */
	movl	$0x14, %ecx

	/* set EDX to 'SMAP' */
	movl	$0x534d4150, %edx

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movw	%si, %es
	movw	$0xe820, %ax
	int	$0x15

	jc	xnosmap

	cmpl	$0x534d4150, %eax
	jne	xnosmap

	cmpl	$0x14, %ecx
	jl	xnosmap

	cmpl	$0x400, %ecx
	jg	xnosmap

	jmp	xsmap

xnosmap:
	movl	$0, %ecx

xsmap:
	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	/* write length of buffer (zero if error) into "addr" */
	movl	0x1c(%esp), %eax
	movl	%ecx, (%eax)

	/* set return value to continuation */
	movl	%ebx, %eax

	pop	%esi
	pop	%edi
	pop	%edx
	pop	%ecx
	pop	%ebx
	pop	%ebp
	ret

/*
 * gateA20(int linear)
 *
 * Gate address-line 20 for high memory.
 *
 * This routine is probably overconservative in what it does, but so what?
 *
 * It also eats any keystrokes in the keyboard buffer.  :-(
 */

ENTRY(gateA20)
	pushl	%eax

	call    gloop1

	movb	$KC_CMD_WOUT, %al
	outb	$K_CMD

gloopint1:
	inb	$K_STATUS
	andb	$K_IBUF_FUL, %al
	jnz	gloopint1

	movb	$KB_OUTPUT_MASK, %al
	cmpb	$0, 0x8(%esp)
	jz	gdoit

	orb	$KB_A20_ENABLE, %al
gdoit:
	outb	$K_RDWR

	call	gloop1

	popl	%eax
	ret

gloop1:
	inb	$K_STATUS
	andb	$K_IBUF_FUL, %al
	jnz	gloop1

gloop2:
	inb	$K_STATUS
	andb	$K_OBUF_FUL, %al
	jz	gloop2ret
	inb	$K_RDWR
	jmp	gloop2

gloop2ret:
	ret

#endif	/* !PC9800 */

#ifdef DEBUG

	.code16

ENTRY(patch_code)	/* labels start with "pc_" */
#ifdef PC9800
	.word	L(pc_start), 0
L(pc_start):
#endif
	mov	%cs, %ax
	mov	%ax, %ds
	mov	%ax, %es
	mov	%ax, %fs
	mov	%ax, %gs
	movl	$0, 0
pc_stop:
	hlt
	jmp	pc_stop
ENTRY(patch_code_end)

	.code32

#endif /* DEBUG */

/*
 * linux_boot()
 *
 * Does some funky things (including on the stack!), then jumps to the
 * entry point of the Linux setup code.
 */

ENTRY(linux_boot)
	/* don't worry about saving anything, we're committed at this point */
	cld	/* forward copying */

	/* copy kernel */
	movl	$LINUX_SETUP, %eax
	movl	LINUX_KERNEL_LEN_OFFSET(%eax), %ecx
	shll	$2, %ecx
	movl	$LINUX_STAGING_AREA, %esi
	movl	$LINUX_KERNEL, %edi

	rep
	movsl

ENTRY(big_linux_boot)
	/* XXX new stack pointer in safe area for calling functions */
	movl	$0x4000, %esp
#ifndef	PC9800
	call	EXT_C(stop_floppy)
#endif
#if 0
	/* final setup for linux boot */

	movw	$LINUX_SETUP_SEG, %ax
	movw	%ax, segment

	xorl	%eax, %eax
	movl	%eax, offset
#endif

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

#if 0
	/* final setup for linux boot */
	movw	$LINUX_SETUP_STACK, %sp

	movw	$LINUX_INIT_SEG, %ax
	movw	%ax, %ss

	/* jump to start */
	ljmp	(offset)
#else
#ifdef PC9800
	call	L(rm_cursor_on)
#endif
	movw	$ LINUX_INIT_SEG, %ax
	movw	%ax, %ss
	movw	$ LINUX_SETUP_STACK, %sp

	ljmp	$ LINUX_SETUP_SEG, $0
#endif

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32


/*
 * multi_boot(int start, int mbi)
 *
 *  This starts a kernel in the manner expected of the multiboot standard.
 */

ENTRY(multi_boot)
	/* no need to save anything */
#ifndef	PC9800
	call	EXT_C(stop_floppy)
#endif

	movl	$0x2BADB002, %eax
	movl	0x8(%esp), %ebx

	/* boot kernel here (absolute address call) */
	call	*0x4(%esp)

	/* error */
	call	EXT_C(stop)


/*
 * cls()
 * BIOS call "INT 10H Function 0Fh" to get current video mode
 *	Call with	%ah = 0x0f
 *      Returns         %al = (video mode)
 *                      %bh = (page number)
 * BIOS call "INT 10H Function 00h" to set the video mode (clears screen)
 *	Call with	%ah = 0x00
 *                      %al = (video mode)
 */


ENTRY(cls)
#ifndef	PC9800
	push	%ebp
	push	%eax
	push	%ebx                    /* save EBX */

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movb	$0xf, %ah
	int	$0x10			/* Get Current Video mode */
        xorb	%ah, %ah
        int	$0x10                   /* Set Video mode (clears screen) */

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	pop	%ebx
	pop	%eax
	pop	%ebp
	ret
#else
	pushl	%ebp		/* prot_to_real messes %ebp */

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movw	$0x0A00, %ax		/* set screen mode to 80x25
					 * (this call will turn off the cursor
					 *  but will not clear the screen)
					 */
	int	$0x18
	movw	$0xE120, %dx		/* %dh = attribute (white face)
					 * %dl = filling character (whitespace)
					 */
	movb	$0x16, %ah		/* clear text VRAM */
	int	$0x18

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	/* real_to_prot zeroes %eax */
	movl	%eax, EXT_C(putchar_y)	/* cursor to (0, 0), clear flags */

	popl	%ebp
	ret
#endif

#ifndef	PC9800
/*
 * getxy()
 * BIOS call "INT 10H Function 03h" to get cursor position
 *	Call with	%ah = 0x03
 *			%bh = page
 *      Returns         %ch = starting scan line
 *                      %cl = ending scan line
 *                      %dh = row (0 is top)
 *                      %dl = column (0 is left)
 */

ENTRY(getxy)
	push	%ebp
	push	%ebx                    /* save EBX */
	push	%ecx                    /* save ECX */
	push	%edx

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

        xorb	%bh, %bh                /* set page to 0 */
	movb	$0x3, %ah
	int	$0x10			/* get cursor position */

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	movb	%dl, %ah
	movb	%dh, %al

	pop	%edx
	pop	%ecx
	pop	%ebx
	pop	%ebp
	ret


/*
 * gotoxy(x,y)
 * BIOS call "INT 10H Function 02h" to set cursor position
 *	Call with	%ah = 0x02
 *			%bh = page
 *                      %dh = row (0 is top)
 *                      %dl = column (0 is left)
 */


ENTRY(gotoxy)
	push	%ebp
	push	%eax
	push	%ebx                    /* save EBX */
	push	%edx

	movb	0x14(%esp), %dl          /* %dl = x */
	movb	0x18(%esp), %dh          /* %dh = y */

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

        xorb	%bh, %bh                /* set page to 0 */
	movb	$0x2, %ah
	int	$0x10			/* set cursor position */

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	pop	%edx
	pop	%ebx
	pop	%eax
	pop	%ebp
	ret

/*
 * set_attrib(attr) :  Sets the character attributes for character at
 *		current cursor position.
 *
 *  Bitfields for character's display attribute:
 *  Bit(s)	Description
 *   7		foreground blink
 *   6-4	background color
 *   3		foreground bright
 *   2-0	foreground color
 *
 *  Values for character color:
 *		Normal		Bright
 *   000b	black		dark gray
 *   001b	blue		light blue
 *   010b	green		light green
 *   011b	cyan		light cyan
 *   100b	red		light red
 *   101b	magenta		light magenta
 *   110b	brown		yellow
 *   111b	light gray	white
 *
 * BIOS call "INT 10H Function 08h" to read character and attribute data
 *	Call with	%ah = 0x08
 *			%bh = page
 *	Returns		%ah = character attribute
 *			%al = character value
 * BIOS call "INT 10H Function 09h" to write character and attribute data
 *	Call with	%ah = 0x09
 *			%al = character value
 *			%bh = page
 *			%bl = character attribute
 *			%cx = count to display (???, possible side-effects!!)
 */

ENTRY(set_attrib)
	push	%ebp
	push	%eax
	push	%ebx
	push	%ecx

	movl	0x14(%esp), %ecx
	xorl	%ebx, %ebx

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movb	$0x8, %ah
	int	$0x10
	movb	$0x9, %ah
	movb	%cl, %bl
	movw	$1, %cx
	int	$0x10

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	pop	%ecx
	pop	%ebx
	pop	%eax
	pop	%ebp
	ret
#endif


/*
 * getrtsecs()
 *	if a seconds value can be read, read it and return it (BCD),
 *      otherwise return 0xFF
 * BIOS call "INT 1AH Function 02H" to check whether a character is pending
 *	Call with	%ah = 0x2
 *	Return:
 *		If RT Clock can give correct values
 *			%ch = hour (BCD)
 *			%cl = minutes (BCD)
 *                      %dh = seconds (BCD)
 *                      %dl = daylight savings time (00h std, 01h daylight)
 *			Carry flag = clear
 *		else
 *			Carry flag = set
 *                         (this indicates that the clock is updating, or
 *                          that it isn't running)
 */
ENTRY(getrtsecs)
	push	%ebp
#if 0	/* gcc does not expect preservation of these registers */
	push	%ecx
	push	%edx
#endif

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

#ifndef	PC9800
	movb	$0x2, %ah
	int	$0x1a

	jnc	gottime
	movb	$0xff, %dh

gottime:
#else /* PC9800 */
	pushw	%bx
	subw	$6, %sp
	movw	%sp, %bx		/* %es:%bx = address of buffer */
	movb	$0x00, %ah		/* read calender/clock */
	int	$0x1C
	popw	%ax			/* discard year, month, day of week */
	popw	%ax			/* discard day of month, hour */
	popw	%dx			/* get minute in %dl, second in %dh */
	popw	%bx
#endif /* PC9800 */
	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	movb	%dh, %al

#if 0
	pop	%edx
	pop	%ecx
#endif
	pop	%ebp
	ret

/*
 * asm_getkey()
 * BIOS call "INT 16H Function 00H" to read character from keyboard
 *	Call with	%ah = 0x0
 *	Return:		%ah = keyboard scan code
 *			%al = ASCII character
 */

#ifndef PC9800
ENTRY(asm_getkey)
#else
ENTRY(getkey)
#endif
	push	%ebp

	call	EXT_C(prot_to_real)

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

#ifndef	PC9800
	int	$0x16
	movw	%ax, %dx		/* real_to_prot uses %eax */
#else
	call	L(rm_cursor_on)
	movb	$0x00, %ah		/* key input */
	int	$0x18
	xchgw	%ax, %dx
	movb	$0x12, %ah		/* cursor turn off */
	int	$0x18
	movb	$1, EXT_C(force_check_disk_change)
#endif

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	movw	%dx,%ax

	pop	%ebp
	ret


/*
 * checkkey()
 *	if there is a character pending, return it; otherwise return -1
 * BIOS call "INT 16H Function 01H" to check whether a character is pending
 *	Call with	%ah = 0x1
 *	Return:
 *		If key waiting to be input:
 *			%ah = keyboard scan code
 *			%al = ASCII character
 *			Zero flag = clear
 *		else
 *			Zero flag = set
 */
ENTRY(checkkey)
	push	%ebp
	push	%ebx

#ifndef	PC9800
	xorl	%ebx, %ebx
#endif

	call	EXT_C(prot_to_real)	/* enter real mode */

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	movb	$0x1, %ah
#ifndef	PC9800
	int	$0x16

	jz	notpending
	movw	%ax, %bx
	jmp	pending

notpending:
	movl	$0xFFFFFFFF, %ebx

pending:
#else
	int	$0x18
	subb	$1, %bh
	sbbl	%ebx, %ebx
	jne	1f
	movw	%ax, %bx
    1:
#endif

	call	EXT_C(real_to_prot)

	/*
	 *  The ".code32" directive only works in GAS, the GNU assembler!
	 *  This gets out of "16-bit" mode.
	 */
	.code32

	mov	%ebx, %eax

	pop	%ebx
	pop	%ebp
	ret

#endif /* NO_FANCY_STUFF */

#ifdef	PC9800
	.code16
	/*
	 *  Emulate IBM PC's Ctrl-Alt-Del
	 */

ENTRY(int09h_handler)
	pushf
	.byte	0x9A			/* lcall */
VARIABLE(int09h_orig_handler)
	.word	0,0
	pushw	%ds
#if 0	/* GAS's bug? it can't generate shorter form */
	pushw	$0
#else
	.byte	0x6A, 0x00
#endif
	popw	%ds
	testb	$0x02, KB_KY_STS+7	/* check DEL key is pressed	*/
	jz	1f			/* -> not pressed		*/
	testb	$0x18, KB_SHFT_STS	/* check CTRL, GRPH is pressed	*/
	jz	1f			/* -> none is pressed		*/
	jpe	L(real_machine_reset)	/* -> both is pressed		*/
1:	popw	%ds
#if 0	/* GAS's bug? `iret' generates operand-size override prefix */
	iret
#else
	.byte	0xCF
#endif

L(real_machine_reset):
	movb	$0, %al
	outb	%al,$0xF0		/* reset CPU */
	ljmp	$0xFFFF, $0x0000	/* jump to BIOS */

#ifndef NO_FANCY_STUFF

L(rm_cursor_on):
	movb	$80, %al
	mulb	EXT_C(putchar_y)	/* %ax = <row> * 80 */
	addb	EXT_C(putchar_x), %al
	adcb	$0, %ah			/* %ax = <row> * 80 + <column> */
	shlw	%ax			/* GDC address to CPU address */
	xchgw	%ax, %dx
	movb	$0x13, %ah		/* move cursor */
	int	$0x18
	movb	$0x11, %ah		/* cursor turn on */
	int	$0x18
	ret

#endif /* NO_FANCY_STUFF */

#ifdef DEBUG

L(real_register_dump):
	pushaw
	movw	%sp,%bp
	pushw	%ds
	pushw	%es
	pushw	%ss
	pushw	%cs
	pushw	16(%bp)
	pushfw

	movw	$1f, %si
	movw	EXT_C(vram)+1, %ax
	shlw	$4, %ax
	movw	%ax, %es
	movb	$160, %al
	mulb	EXT_C(putchar_y)
	subw	$320, %ax
	xorw	%di, %di
	xchgw	%ax, %di
	movw	%ax, %ds		/* just in case */
    4:	lodsb; stosw
	lodsb; stosw
	movb	$'=', %al
	stosw
	movw	14(%bp), %dx
	movw	$4, %cx
    5:	rolw	$4, %dx
	movb	%dl, %al
	andb	$0x0F, %al
	/*
	 *  following code to translate a nibble to hexadecimal digit is
	 *  quite clever:
	 */			/* 0) 00 01 ... 08 09 0A 0B 0C 0D 0E 0F	*/
	cmpb	$10, %al	/* 1) |<-- CF:=1 -->| |<--- CF:=0 --->| */
	sbbb	$0x69, %al	/* 2) |<--- -70 --->| |<---- -69 ---->| */
				/*	      *IN PACKED BCD*		*/
	das			/* 3) 30 31 ... 38 39 41 42 43 44 45 46 */
				/*    These are hexadecimal ascii codes.*/
	stosw
	loop	5b
	movb	$' ', %al
	stosw
	decw	%bp
	decw	%bp
	cmpw	$3f, %si
	jae	6f
	cmpw	$2f, %si
	jne	4b
	movb	$16, %cl
	rep; stosw
	jmp	4b
    6:	movw	%di, %dx
	addw	$0x2000 - (80+48) * 2, %di
	movb	$ DEBUG_ATTRIBUTE, %al
	movb	$80+48, %cl
	rep; stosw

	testb	$0x02, KB_SHFT_STS
	jnz	7f		/* do not wait if CAPS key is locked */
	movb	$0x13, %ah	/* move cursor */
	int	$0x18
	movb	$0x11, %ah	/* cursor turn on */
	int	$0x18
	movb	$0x00, %ah	/* key input wait */
	int	$0x18
	movb	$0x12, %ah	/* cursor turn off */
	int	$0x18
    7:
	popfw
	popw	%ax		/* IP */
	popw	%ax		/* CS */
	popw	%ax		/* SS */
	popw	%es
	popw	%ds
	popaw
	ret

    1:	.ascii	"AXCXDXBXSPBPSIDI"
    2:	.ascii	"DSESSSCSIPFL"
    3:
#endif /* DEBUG */

	.code32
#endif	/* PC9800 */

/*
 *  This is the area for all of the special variables.
 */

	.p2align	2	/* force 4-byte alignment */

protstack:
	.long	PROTSTACKINIT

#ifndef PC9800
VARIABLE(boot_drive)
	.long	0
#endif

#if 0
	/* an address can only be long-jumped to if it is in memory, this
	   is used by multiple routines */
offset:
	.long	0x8000
segment:
	.word	0
#endif

/*
 * This is the Global Descriptor Table
 *
 *  An entry, a "Segment Descriptor", looks like this:
 *
 * 31          24         19   16                 7           0
 * ------------------------------------------------------------
 * |             | |B| |A|       | |   |1|0|E|W|A|            |
 * | BASE 31..24 |G|/|0|V| LIMIT |P|DPL|  TYPE   | BASE 23:16 |
 * |             | |D| |L| 19..16| |   |1|1|C|R|A|            |
 * ------------------------------------------------------------
 * |                             |                            |
 * |        BASE 15..0           |       LIMIT 15..0          |
 * |                             |                            |
 * ------------------------------------------------------------
 *
 *  Note the ordering of the data items is reversed from the above
 *  description.
 */

	.p2align	2	/* force 4-byte alignment */
gdt:
	.word	0, 0
	.byte	0, 0, 0, 0

	/* code segment */
	.word	0xFFFF, 0
	.byte	0, 0x9A, 0xCF, 0

	/* data segment */
	.word	0xFFFF, 0
	.byte	0, 0x92, 0xCF, 0

	/* 16 bit real mode CS */
	.word	0xFFFF, 0
	.byte	0, 0x9E, 0, 0

	/* 16 bit real mode DS */
	.word	0xFFFF, 0
	.byte	0, 0x92, 0, 0


/* this is the GDT descriptor */
gdtdesc:
	.word	0x27			/* limit */
	.long	gdt			/* addr */

#ifdef PC9800

#ifdef DEBUG
ENTRY(dprintf)
	popl	%eax			/* pop return address */
	movl	%eax, 1f		/* save return address */
	movb	$ DEBUG_ATTRIBUTE, %al
	xchgb	%al, EXT_C(putchar_attribute)
	movb	%al, 2f			/* save text attribute */
	call	EXT_C(printf)
	movb	2f, %al
	movb	%al, EXT_C(putchar_attribute)	/* restore text attribute */
	jmp	*1f

	.section .bss
	.balign	4
    1:	.space	4
    2:	.space	1
	.previous
#endif /* DEBUG */

/*
 *  struct disk_geometry * get_geometry(int drive)	(drive is %eax)
 *	-- Get disk geometry structure for given DRIVE.
 *	   If no drive can be attached to DRIVE, it returns NULL.
 *	   *This function does not check if physical device is actually
 *	   detected.*
 *	-- This function can be written in C quite easily, but in assembly
 *	   to deflate code size. C compilers will optimize code for its speed
 *	   rather than its size.
 */
	.balign	4
ENTRY(get_geometry)
	testl	$~0xFF,%eax
	jnz	0f
	cmpb	$0x04,%al
	jb	1f
	cmpb	$0x10,%al
	jb	0f
	cmpb	$0x14,%al
	jb	2f
	cmpb	$0x20,%al
	jb	0f
	cmpb	$0x27,%al
	jb	3f
	cmpb	$0x30,%al
	jb	0f
	cmpb	$0x34,%al
	jb	4f
	cmpb	$0x70,%al
	jb	0f
	cmpb	$0x74,%al
	jb	5f
	cmpb	$0x80,%al
	jb	0f
	cmpb	$0x84,%al
	jb	6f
	cmpb	$0x90,%al
	jb	0f
	cmpb	$0x94,%al
	jb	7f
	cmpb	$0xA0,%al
	jb	0f
	cmpb	$0xA7,%al
	jb	8f
	cmpb	$0xB0,%al
	jb	0f
	cmpb	$0xB4,%al
	jb	9f
	cmpb	$0xF0,%al
	jb	0f
	cmpb	$0xF4,%al
	jae	0f
	leal	GEOMETRY_TABLE+( 4-0xF0)*8(,%eax,8),%eax
	ret
	.balign	4
0:	xorl	%eax,%eax
	ret
	.balign	4
	/*
	 * following labels are all aligned because
	 * `leal <disp32>(,<r32>,<scale>),<r32>' and `ret' is 8 bytes long.
	 */
1:	leal	GEOMETRY_TABLE+( 8-0x00)*8(,%eax,8),%eax
	ret
2:	leal	GEOMETRY_TABLE+( 0-0x10)*8(,%eax,8),%eax
	ret
3:	leal	GEOMETRY_TABLE+(12-0x20)*8(,%eax,8),%eax
	ret
4:	leal	GEOMETRY_TABLE+( 0-0x30)*8(,%eax,8),%eax
	ret
5:	leal	GEOMETRY_TABLE+( 4-0x70)*8(,%eax,8),%eax
	ret
6:	leal	GEOMETRY_TABLE+( 8-0x80)*8(,%eax,8),%eax
	ret
7:	leal	GEOMETRY_TABLE+( 0-0x90)*8(,%eax,8),%eax
	ret
8:	leal	GEOMETRY_TABLE+(12-0xA0)*8(,%eax,8),%eax
	ret
9:	leal	GEOMETRY_TABLE+( 4-0xB0)*8(,%eax,8),%eax
	ret

#define DEFVAR(var, addr, size)	.globl EXT_C(var); EXT_C(var) = (addr)

	DEFVAR(geometry_table,		0x600, 0x100)
	DEFVAR(drive_lookup_table,	0x700, 0x100)
	DEFVAR(partition_table,		0x800, 0x800)
#endif

/* entry point (no effect in fact; only to suppress warning from ld) */
	.global	_start
	_start	= EXT_C(start)
