/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#define _CHAR_IO_C

#include "shared.h"

#ifndef NO_FANCY_STUFF
#ifndef PC9800
int
getkey(void)
{
  buf_drive = -1;
  return asm_getkey();
}
#endif
#endif  /* NO_FANCY_STUFF */


void
print_error(void)
{
  if (errnum > ERR_NONE && errnum < MAX_ERR_NUM)
#ifndef NO_FANCY_STUFF
    /* printf("\7\n %s\n", err_list[errnum]); */
    printf("\n %s\n", err_list[errnum]);
#else /* NO_FANCY_STUFF */
    printf("Error: %d\n", errnum);
#endif  /* NO_FANCY_STUFF */

  errnum = ERR_NONE;
}

#ifdef PC9800
__asm__(".align	4
convert_to_ascii_subr:
	xorl %edx,%edx
	divl %ecx
	testl %eax,%eax
	pushl %edx
	jz 1f
	call convert_to_ascii_subr
1:	popl %eax
	addb $'0',%al
	cmpb $'0'+10,%al
	jb 2f
	addb $'a'-'0'-10,%al
2:	stosb
	ret");
#endif

char *
convert_to_ascii(char *buf, int c, ...)
{
  unsigned long num = *((&c) + 1), mult = 10;
  char *ptr = buf;

#ifndef PC9800
  if (c == 'x')
    mult = 16;

  if ((num & 0x80000000uL) && c == 'd')
    {
      num = (~num)+1;
      *(ptr++) = '-';
      buf++;
    }

  do
    {
      int dig = num % mult;
      *(ptr++) = ( (dig > 9) ? dig + 'a' - 10 : '0' + dig );
    }
  while (num /= mult);

  /* reorder to correct direction!! */
  {
    char *ptr1 = ptr-1;
    char *ptr2 = buf;
    while (ptr1 > ptr2)
      {
	int c = *ptr1;
	*ptr1 = *ptr2;
	*ptr2 = c;
	ptr1--;
	ptr2++;
      }
  }
#else
  if (c == 'x')
    mult = 16;

  __asm__("call convert_to_ascii_subr"
	  : "=D" (ptr), "=a" (num), "=c" (mult)
	  : "0"(buf), "1"(num), "2"(mult));
#endif
  return ptr;
}


void
printf(const char *format, ... )
{
  int *dataptr = (int *) &format;
  char c, *ptr, str[16];

  dataptr++;

  while ((c = *(format++)) != 0)
    {
      if (c != '%')
	putchar(c);
      else
#ifndef PC9800
	switch (c = *(format++))
	  {
	  case 'd': case 'u': case 'x':
	    *convert_to_ascii(str, c, *((unsigned long *) dataptr++)) = 0;

	    ptr = str;

	    while (*ptr)
	      putchar(*(ptr++));
	    break;

	  case 'c': putchar((*(dataptr++))&0xff); break;

	  case 's':
	    ptr = (char *)(*(dataptr++));

	    while ((c = *(ptr++)) != 0)
	      putchar(c);
	    break;
	  }
#else
        {
	  int width = 0;
	  char *end;
	  unsigned long val = *(unsigned long *)dataptr++;

#if 0
	  safe_parse_maxint(&format, &width);
	  switch((c = *format++))
#else
	  while ('0' <= (c = *format++) && c <= '9')
	    width = width * 10 + (c - '0');
	  switch (c)
#endif
	    {
	    case 'd':
	      if ((int)val < 0)
		{
		  putchar('-');
		  val = -(int)val;
		}
	    case 'u':
	    case 'x':
	      end = convert_to_ascii(str, c, val);
	      *end = 0;
	      if (width)
		{
		  ptr = end - width;
		  while (ptr < str)
		    {
		      putchar((c == 'x' ? '0' : ' '));
		      ptr++;
		    }
		  goto string;
		}
	      ptr = str;
	      goto string;
	    case 's':
	      ptr = (char *)val;
	    string:
	      while (*ptr)
		putchar(*ptr++);
	      break;
	    case 'c':
	      putchar(val);
	      break;
	    case '%':
	      putchar(c);
	      break;
	    }
	}
#endif
    }
}


#ifndef NO_FANCY_STUFF

void
init_page(void)
{
  cls();

  printf("\n    GRUB  version %s  (%dK lower / %dK upper memory)\n\n",
	 version_string,
	 (unsigned int)mbi.mem_lower, (unsigned int)mbi.mem_upper);
}


/* don't use this with a maxlen greater than 1600 or so!  the problem
   is that it depends on the whole thing fitting on the screen at once,
   and the whole screen is about 2000 characters, minus the prompt...
   so want to leave space for an error line, etc.
   maxlen should be at least 1, and we shouldn't have a NULL prompt or
   cmdline...  the cmdline must be a valid string at the start */

int
get_cmdline(char *prompt, char *commands, char *cmdline, int maxlen)
{
  int ystart, yend, xend, lpos, c;
  int plen = 0;
  int llen = 0;

  /* nested function definition for code simplicity XXX GCC only, I think */
  void cl_print(char *str)
    {
      while (*str != 0)
	{
#ifndef PC9800
	  putchar(*(str++));
	  if (++xend > 78)
	    {
	      xend = 0;
	      putchar(' ');
	      if (yend == (getxy() & 0xFF))
		ystart--;
	      else
		yend++;
	    }
#else
	  Byte orig_y = putchar_y;
	  putchar(*str++);
	  if (putchar_x == 0)
	    {
	      if (orig_y == putchar_y)
		ystart--;
	      else
		yend++;
	    }
#endif
	}
#ifdef PC9800
      xend = putchar_x;
#endif
    }
  /* nested function definition for code simplicity XXX GCC only, I think */
  void cl_setcpos(void)
    {
#ifndef PC9800
      yend = ((lpos+plen) / 79) + ystart;
      xend = ((lpos+plen) % 79);
      gotoxy(xend, yend);
#else
      putchar_y = yend = ((lpos+plen) / 80) + ystart;
      putchar_x = xend = ((lpos+plen) % 80);
#endif
    }

  /* nested function definition for initial command-line printing */
  void cl_init()
    {
      /* distinguish us from other lines and error messages! */
      putchar('\n');

      /* print full line and set position here */
#ifndef PC9800
      ystart = (getxy() & 0xFF);
      yend = ystart;
#else
      ystart = yend = putchar_y;
#endif
      xend = 0;
      cl_print(prompt);
      cl_print(cmdline);
      cl_setcpos();
    }

  /* nested function definition for erasing to the end of the line */
  void cl_kill_to_end()
    {
      int i;
      cmdline[lpos] = 0;
      for (i = lpos; i <= llen; i++)
	{
#ifndef PC9800
	  if (i && ((i + plen) % 79) == 0)
	    putchar(' ');
#endif
	  putchar(' ');
	}
      llen = lpos;
      cl_setcpos();
    }

  while (prompt[plen])
    plen++;
  while (cmdline[llen])
    llen++;
  /* XXX limiting maxlen to 1600 */
  if (maxlen > 1600)
    {
      maxlen = 1600;
      if (llen > 1599)
	{
	  llen = 1599;
	  cmdline[1600] = 0;
	}
    }
  lpos = llen;

  cl_init();

  while (ASCII_CHAR(c = getkey()) != '\n' &&  ASCII_CHAR(c) != '\r')
    {
      switch (c)
	{
	case KEY_LEFT:
	  c = 2;
	  break;
	case KEY_RIGHT:
	  c = 6;
	  break;
	case KEY_HOME:
	  c = 1;
	  break;
#ifndef	PC9800	/* PC-9800 have no END key. */
	case KEY_END:
	  c = 5;
	  break;
#endif
	case KEY_DELETE:
#ifndef	PC9800
	  c = 8;
#else
	  c = 4;
#endif
	  break;
	default:
	}

      c = ASCII_CHAR(c);

      switch (c)
	{
	case 27:  /* ESC immediately return 1*/
	  return 1;
	case 9:   /* TAB lists completions */
	  {
	    int i, j = 0, llen_old = llen;

	    while (cmdline[j] && cmdline[j] != '=')
	      j++;

	    /* since the command line cannot have a '\n', we're OK to use c */
	    c = cmdline[lpos];

	    cl_kill_to_end();

	    /* goto part after line here */
#ifndef PC9800
	    yend = ((llen+plen) / 79) + ystart;
#else
	    yend = ((llen+plen) / 80) + ystart;
#endif
	    gotoxy(0, yend); putchar('\n');

	    if (lpos > j)
	      {
		for (i = lpos; i > 0 && cmdline[i-1] != ' '; i--);
		if (i <= j)
		  i = j+1;
		/* print possible completions */
		print_completions(cmdline+i);
	      }
	    else if (commands)
	      printf(commands);
	    else
	      break;

	    /* restore command-line */
	    cmdline[lpos] = c;
	    llen = llen_old;
	    cl_init();
	  }
	  break;
	case 1:   /* C-a go to beginning of line */
	  lpos = 0;
	  cl_setcpos();
	  break;
	case 5:   /* C-e go to end of line */
	  lpos = llen;
	  cl_setcpos();
	  break;
	case 6:   /* C-f forward one character */
	  if (lpos < llen)
	    {
	      lpos++;
	      cl_setcpos();
	    }
	  break;
	case 2:   /* C-b backward one character */
	  if (lpos > 0)
	    {
	      lpos--;
	      cl_setcpos();
	    }
	  break;
	case 4:   /* C-d delete character under cursor */
	  if (lpos == llen)
	    break;
	  lpos++;
	  /* fallthrough is on purpose! */
	case 8:   /* C-h backspace */
	  if (lpos > 0)
	    {
	      int i;
	      for (i = lpos-1; i < llen; i++)
		cmdline[i] = cmdline[i+1];
	      i = lpos;
	      lpos = llen - 1;
	      cl_setcpos();
	      putchar(' ');
	      lpos = i - 1;  /* restore lpos and decrement */
	      llen--;
	      cl_setcpos();
	      if (lpos != llen)
		{
		  cl_print(cmdline+lpos);
		  cl_setcpos();
		}
	    }
	  break;
	case 21:   /* C-u kill to beginning of line */
	  if (lpos == 0)
	    break;
	  {
	    int i;
	    for (i = 0; i < (llen-lpos); i++)
	      cmdline[i] = cmdline[lpos+i];
	  }
	  lpos = llen-lpos;
	  cl_setcpos();
	  /* fallthrough on purpose! */
	case 11:   /* C-k kill to end of line */
	  if (lpos < llen)
	    {
	      cl_kill_to_end();
	      if (c == 21)
		{
		  lpos = 0;
		  cl_setcpos();
		  cl_print(cmdline);
		  cl_setcpos();
		}
	    }
	  break;
	default:   /* insert printable character into line */
	  if (llen < (maxlen-1) && c >= ' ' && c <= '~')
	    {
	      if (lpos == llen)
		{
		  cmdline[lpos] = c;
		  cmdline[lpos+1] = 0;
		  cl_print(cmdline+lpos);
		  lpos++;
		}
	      else
		{
		  int i;
		  for (i = llen; i >= lpos; i--)
		    cmdline[i+1] = cmdline[i];
		  cmdline[lpos] = c;
		  cl_setcpos();
		  cl_print(cmdline+lpos);
		  lpos++;
		  cl_setcpos();
		}
	      llen++;
	    }
	}
    }

  /* goto part after line here */
#ifndef PC9800
  yend = ((llen+plen) / 79) + ystart;
#else
  yend = ((llen+plen) / 80) + ystart;
#endif
  gotoxy(0, yend); putchar('\n');

  /* remove leading spaces */
  /* use c and lpos as indexes now */
  for (lpos = 0; cmdline[lpos] == ' '; lpos++);

  if (lpos != 0)
    {
      c = 0;
      do
	{
	  cmdline[c] = cmdline[lpos];
	  c++;
	  lpos++;
	}
      while (cmdline[lpos]);
    }

  cmdline[c] = 0;

  return 0;
}

#endif /* NO_FANCY_STUFF */

#if 0
int
get_based_digit(int c, int base)
{
  int digit = -1;

  /* make sure letter in the the range we can check! */
  c = tolower(c);

  /*
   *  Is it in the range between zero and nine?
   */
  if (base > 0 && c >= '0' && c <= '9' && c < (base + '0'))
    {
      digit = c - '0';
    }

  /*
   *  Is it in the range used by a letter?
   */
  if (base > 10 && c >= 'a' && c <= 'z' && c < ((base - 10) + 'a'))
    {
      digit = c - 'a' + 10;
    }

  return digit;
}
#endif

int
safe_parse_maxint(char **str_ptr, int *myint_ptr)
{
  register char *ptr = *str_ptr;
  register int myint = 0, digit;
  int mult = 10, found = 0;

  /*
   *  Is this a hex number?
   */
  if (*ptr == '0' && (*(ptr+1) == 'x' || *(ptr+1) == 'X') )
    {
      ptr += 2;
      mult = 16;
    }

#if 0
  while ((digit = get_based_digit(*ptr, mult)) != -1)
#else
  while (1)
#endif
    {
#if 0
#else
      register unsigned char c __asm__("al") = *ptr - '0';
      if (9 < c)
	{
	  c -= 'A' - '0';
	  if ('a' - 'A' <= c)
	    c -= 'a' - 'A';
	  if ('z' - 'a' < c)
	    break;
	  c += 10;
	}
      digit = c;
#endif
      found = 1;
      if (myint > ((MAXINT - digit)/mult))
	{
	  errnum = ERR_NUMBER_PARSING;
	  return 0;
	}
      myint = (myint * mult) + digit;
      ptr++;
    }

  if (!found)
    {
      errnum = ERR_NUMBER_PARSING;
      return 0;
    }

  *str_ptr = ptr;
  *myint_ptr = myint;

  return 1;
}


int
tolower(int c)
{
  if (c >= 'A' && c <= 'Z')
    return (c + ('a' - 'A'));

  return c;
}


int
isspace(int c)
{
  if (c == ' ' || c == '\t' || c == '\n')
    return 1;

  return 0;
}


int
strncat(char *s1, char *s2, int n)
{
  int i = -1;

  while (++i < n && s1[i] != 0);

  while (i < n && (s1[i++] = *(s2++)) != 0);

  s1[n-1] = 0;

  if (i >= n)
    return 0;

  s1[i] = 0;

  return 1;
}


int
strcmp(char *s1, char *s2)
{
  while (*s1 == *s2)
    {
      if (!*(s1++))
	return 0;
      s2++;
    }

  if (*s1 == 0)
    return -1;

  return 1;
}


char *
strstr(char *s1, char *s2)
{
  char *ptr, *tmp;

  while (*s1)
    {
      ptr = s1;
      tmp = s2;

      while (*s1 && *s1++ == *tmp++);

      if (tmp > s2 && !*(tmp-1))
	return ptr;
    }

  return 0;
}


int
memcheck(int start, int len)
{
#ifndef PC9800
  if ( (start < 0x1000) || (start < 0x100000
			   && (mbi.mem_lower * 1024) < (start+len))
       || (start >= 0x100000
	   && (mbi.mem_upper * 1024) < ((start-0x100000)+len)) )
#else
  if ((start < 0x600) || (start < 0x100000
			   && (mbi.mem_lower * 1024) < (start+len))
       || (start >= 0x100000
	   && (mbi.mem_upper * 1024) < ((start-0x100000)+len)))
#endif
    errnum = ERR_WONT_FIT;

  return (!errnum);
}


int
bcopy(const char *from, char *to, int len)
{
  if (memcheck((int)to, len))
    {
      if ((to >= from+len) || (to <= from))
	{
#if 0
	  while (len > 3)
	    {
	      len -= 4;
	      *(((unsigned long *)to)++) = *(((unsigned long *)from)++);
	    }
	  while (len-- > 0)
	    *(to++) = *(from++);
#else
	  asm volatile ("cld\n\
	rep\n\
	movsl\n\
	test%B3 %4,%b3\n\
	jz 1f\n\
	movsw\n\
     1:	test%B3 %5,%b3\n\
	jz 2f\n\
	movsb\n\
     2:"
			: "=S" (from), "=D" (to), "=c" (len)
			: "qm" (len), "i" (2), "i" (1),
			  "0" (from), "1" (to), "2" (len >> 2));
#endif
	}
      else
	{
#if 0
	  while (len-- > 0)
	    to[len] = from[len];
#else
	  asm volatile ("std\n	rep\n	movsb\n	cld"
			: "=S" (from), "=D" (to), "=c" (len)
			: "0" (from + len - 1), "1" (to + len - 1), "2" (len));
#endif
	}
    }

  return (!errnum);
}


int
bzero(char *start, int len)
{
  if (memcheck((int)start, len))
    {
      while (len-- > 0)
	*(start++) = 0;
    }

  return (!errnum);
}


/* Forward compatibility with GNU GRUB.
   I think, however, these are preferrable names over original GRUB's ones,
   since they can better cope with ANSI C  -- tak  */

#define STRING1(x)	#x
#define STRING(x)	STRING1 (x)
#define ALIAS(fun)	__attribute__ ((alias (STRING (EXT_C (fun)))))

int substring (const char *s1, const char *s2) ALIAS (strcmp);
