/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _COMMON_C

#include "shared.h"


/*
 *  Shared BIOS/boot data.
 */

struct multiboot_info mbi;
unsigned long saved_drive;
unsigned long saved_partition;
unsigned long saved_mem_upper;

/*
 *  Error code stuff.
 */

int errnum = 0;

#ifndef NO_FANCY_STUFF

char *err_list[] =
{
  0,
  MSG2("Selected item won\'t fit into memory",
       "礭뤿Υޤ"),
  MSG3("Selected disk doesn\'t exist",
       "ꤵ줿ǥǧǤޤ",
       "ʥǥϤ!"),
  MSG2("Disk read error",
       "ǥɤ߹˥顼ȯޤ"),
  MSG2("Disk write error",
       "ǥ񤭹˥顼ȯޤ"),
  MSG2("Disk geometry error",
       "ǥΥȥǤʤѴ˥顼ȯޤ"),
  MSG2("Attempt to access block outside partition",
       "ѡƥΰ賰򥢥褦Ȥޤ"),
  MSG2("Partition table invalid or corrupt",
       "ѡƥơ֥뤬ޤ"),
  MSG3("No such partition",
       "ꤵ줿ѡƥϤޤ",
       "ʥѡƥ"),
  MSG2("Bad filename (must be absolute pathname or blocklist)",
       "ե̾Ǥ (Хѥ֥åꥹȤǻꤷƤ)"),
  MSG2("Bad file or directory type",
       "ꤵ줿եˤбƤޤ"),
  MSG3("File not found",
       "ե뤬Ĥޤ",
       "ե뤬"),
  MSG2("Cannot mount selected partition",
       "ꤵ줿ѡƥޥȤǤޤ"),
  MSG3("Inconsistent filesystem structure",
       "ե륷ƥ̷⤬ޤ",
       "ե륷ƥबѤޤ"),
  MSG3("Filesystem compatibility error, can\'t read whole file",
       "ե뤬礭бǤޤ (Ǹޤɤ߹ޤ)",
       "ե뤬礭ޤ (Ǹޤɤ߹ޤ)"),
  MSG2("Error while parsing number",
       "ͤλ˸꤬ޤ"),
  MSG2("Device string unrecognizable",
       "ǥХ꤬Ǥ"),
  MSG2("Invalid device requested",
       "ꤵ줿ǥХ̵Ǥ"),
  MSG2("Invalid or unsupported executable format",
       "ޤбƤʤ¹ԷǤ"),
  MSG2("Loading below 1MB is not supported",
       "1MB̤Υ˥ɤ뤳ȤϤǤޤ"),
  MSG2("Unsupported Multiboot features requested",
       "ꤵ줿ޥ֡ȵǽˤбƤޤ"),
  MSG3("Unknown boot failure",
       "֡Ȥ˼Ԥޤ",
       "!"),
  MSG2("Must load Multiboot kernel before modules",
       "⥸塼ɤ˥ޥ֡ȥͥɤƤ"),
  MSG2("Must load Linux kernel before initrd",
       "initrd ɤ Linux ͥɤƤ"),
  MSG2("Cannot boot without kernel loaded",
       "ͥɤʤȥ֡ȤǤޤ"),
  MSG3("Unrecognized command",
       "ޥɤǤ",
       "ޥɤѤޤ"),
  MSG2("Bad or incompatible header on compressed file",
       "̥եΥإåʤбƤʤǤ"),
  MSG2("Bad or corrupt data while decompressing file",
       "̥եŸ˥ǡ顼ȯޤ"),
  MSG2("Bad or corrupt version of stage1/stage2",
       "stage1/stage2 ΥС󤬹äƤޤ"),
  /* GNU GRUB compatibility */
  MSG2 ("Too many symbolic links",
	"ե̾Υܥå󥯤¿ޤ"),
  MSG2 ("Invalid argument",
	"̵Ǥ"),
  0
};


/* static for BIOS memory map fakery */
static struct AddrRangeDesc fakemap[3] =
{
  { 20, 0, 0, 0, 0, MB_ARD_MEMORY },
  { 20, 0x100000, 0, 0, 0, MB_ARD_MEMORY },
  { 20, 0x1000000, 0, 0, 0, MB_ARD_MEMORY }
};
#endif  /* NO_FANCY_STUFF */

#ifdef	PC9800

#define INIT_GEOMETRY(disk_equip, daua, index)			({	\
	int __dummy1, __dummy2;						\
	Byte __disk_equip = (disk_equip);				\
	if (__disk_equip)						\
	  __asm__ __volatile__("1:	inc%L1 %1\n"			\
			       "	shr%B0 %b0\n"			\
			       "	jnc 2f\n"			\
			       "	mov%B1 %b1,%P2(,%1,8)\n"	\
			       "2:	jnz 1b"				\
			       : "=q"(__dummy1), "=q"(__dummy2)		\
			       : "g"(&(geometry_table			\
				       [(index) - (daua)].da_ua)),	\
			         "0"(__disk_equip), "1"((daua) - 1));	\
})

#define	INIT_SCSI_GEOMETRY(disk_equip, daua, index)		({	\
	int __dummy1, __dummy2;						\
	__asm__ __volatile__("1:mov%B1 %b1,%P2(,%1,8)\n"		\
			     "	shr%B0 %b0\n"				\
			     "	jnc 2f\n"				\
			     "	or%B4 %4,%P2(,%1,8)\n"			\
			     "2:\tinc%L1 %1\n"				\
			     "	cmp%B1 %b3,%b1\n"			\
			     "	jb 1b"					\
			     : "=q"(__dummy1), "=q"(__dummy2)		\
			     : "g"(&geometry_table			\
				   [(index) - (daua)].da_ua),		\
			       "ir"((daua) + 7), "ir"(0x80),		\
			       "0"((disk_equip)), "1"((daua)));		\
})

static __inline__ void
init_geometry(void)
{
  struct disk_geometry * boot_drive_geometry;

  /* clear geometry_table[]. */
  memset (geometry_table, -1, sizeof(struct disk_geometry[20]));

  INIT_GEOMETRY(*DISK_EQUIP & 0x000F, 0x90, 0);
  INIT_GEOMETRY((((*DISK_EQUIP >> 8) | *DISK_EQUIP2) >> 4) & 0xF, 0x70, 4);
  INIT_GEOMETRY((*DISK_EQUIP >> 8) & 0xF, 0x80, 8);
  if (DISK_XROM[0x2] && DISK_XROM[0xA])
    INIT_SCSI_GEOMETRY(*DISK_EQUIPS & 0x7F, 0x20, 12);
  if ((boot_drive & 0x50)
      && (boot_drive_geometry = get_geometry(boot_drive)) != NULL)
    boot_drive_geometry->da_ua = boot_drive;
}

#define	get_memsize(x) (((x) ? *EXPMMSZ : (*BIOS_FLAG & 7) + 1) << 7)
#endif /* PC9800 */

/*
 *  This queries for BIOS information.
 */

void
init_bios_info(void)
{
#ifndef	PC9800
  int cont, memtmp, addr;
#endif

  /*
   *  Get information from BIOS on installed RAM.
   */


  mbi.mem_lower = get_memsize(0);
  mbi.mem_upper = get_memsize(1);

#ifndef NO_FANCY_STUFF
  /*
   *  We need to call this somewhere before trying to put data
   *  above 1 MB, since without calling it, address line 20 will be wired
   *  to 0.  Not too desirable.
   */

  gateA20(1);

  /*
   *  The "mbi.mem_upper" variable only recognizes upper memory in the
   *  first memory region.  If there are multiple memory regions,
   *  the rest are reported to a Multiboot-compliant OS, but otherwise
   *  unused by GRUB.
   */

#ifndef	PC9800
  mbi.mmap_addr = (addr = (((int) end) & ~3) + 4);
  mbi.mmap_length = 0;
  cont = 0;

  do
    {
      cont = get_mem_map(addr, cont);

      if ( ! *((int *)addr) )
	break;

      mbi.mmap_length += *((int *)addr) + 4;
      addr += *((int *)addr) + 4;
    }
  while (cont);

  if (mbi.mmap_length)
    {
      /*
       *  This is to get the upper memory up to the first memory
       *  hole into the "mbi.mem_upper" element, for OS's that
       *  don't care about the memory map, but might care about
       *  RAM above 64MB.
       *
       *  A big problem is that the memory areas aren't guaranteed
       *  to be:  (1) contiguous, (2) sorted in ascending order, or
       *  (3) non-overlapping.
       */
      memtmp = 0x100000;

      do
	{
	  for (cont = 0, addr = mbi.mmap_addr;
	       addr < mbi.mmap_addr + mbi.mmap_length;
	       addr += *((int *) addr) + 4)
	    {
	      if (((struct AddrRangeDesc *)addr)->BaseAddrHigh == 0
		  && ((struct AddrRangeDesc *)addr)->Type == MB_ARD_MEMORY
		  && ((struct AddrRangeDesc *)addr)->BaseAddrLow <= memtmp
		  && (((struct AddrRangeDesc *)addr)->BaseAddrLow
		      + ((struct AddrRangeDesc *)addr)->LengthLow) > memtmp)
		{
		  memtmp = (((struct AddrRangeDesc *)addr)->BaseAddrLow
			    + ((struct AddrRangeDesc *)addr)->LengthLow);
		  cont++;
		}
	    }
	}
      while (cont);

      mbi.mem_upper = (memtmp - 0x100000) >> 10;
    }
  else if ((memtmp = get_eisamemsize()) != -1)
    {
      cont = memtmp & ~0xFFFF;
      memtmp = memtmp & 0xFFFF;

      if (!cont || (memtmp == 0x3c00))
	memtmp += (cont >> 10);
      else
	{
	  /* XXX should I do this at all ??? */

	  mbi.mmap_addr = (int)fakemap;
	  mbi.mmap_length = sizeof(fakemap);
	  fakemap[0].LengthLow = (mbi.mem_lower << 10);
	  fakemap[1].LengthLow = (memtmp << 10);
	  fakemap[2].LengthLow = cont;
	}

      mbi.mem_upper = memtmp;
    }

#else	/* PC9800 */

  fakemap[0].LengthLow = mbi.mem_lower << 10;
  fakemap[1].LengthLow = mbi.mem_upper << 10;
  if (fakemap[1].LengthLow == (15 << 20))
    {
      mbi.mem_upper += *EXPMMSZ16M << 10;
      fakemap[1].LengthLow += *EXPMMSZ16M << 20;
    }
  else
      fakemap[2].LengthLow = *EXPMMSZ16M << 20;
  mbi.mmap_length
    = fakemap[2].LengthLow ? sizeof(fakemap[0]) * 3 : sizeof(fakemap[0]) * 2;
  mbi.mmap_addr = (int)fakemap;

#endif	/* PC9800 */

  saved_mem_upper = mbi.mem_upper;

  /*
   *  Initialize other Multiboot Info flags.
   */

  mbi.flags = MB_INFO_MEMORY | MB_INFO_CMDLINE | MB_INFO_BOOTDEV;

#endif  /* NO_FANCY_STUFF */

#ifdef PC9800
  init_geometry();

  hook_int09h();		/* start trapping CTRL+GRPH+DEL */
#endif

  /*
   *  Set boot drive and partition.
   */

  saved_drive = boot_drive;
  saved_partition = install_partition;

  /*
   *  Start main routine here.
   */

#if 0
  {
    int i = 0x400;
    while (i < 0x600)
      {
	if (!(i & 31))
	  printf("%4x", i);
	if (!(i & 3))
	  putchar(' ');
	printf("%2x", *(Byte *)i++);
	if (!(i & 31))
	  putchar('\n');
      }
  }
  printf("getkey() ƤӽФޤ: ");
  printf("%4x\n", getkey());
#endif

  cmain();
}
