/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#define _DISK_IO_C

#include "shared.h"

#include "filesys.h"

/* XXX for evil hack */
#include "freebsd.h"

#ifndef NO_FANCY_STUFF
/* instrumentation variables */
void (*debug_fs)(int) = NULL;
void (*debug_fs_func)(int) = NULL;
#endif  /* NO_FANCY_STUFF */

/* these have the same format as "boot_drive" and "install_partition", but
   are meant to be working values */
unsigned long current_drive = 0xFF;
unsigned long current_partition;

/*
 *  Global variables describing details of the filesystem
 */

/* XXX BSD evil hack */
int bsd_evil_hack;

/* filesystem type */
int fsys_type = NUM_FSYS;
#ifndef NO_BLOCK_FILES
int block_file = 0;
#endif  /* NO_BLOCK_FILES */

/* these are the translated numbers for the open partition */
long part_start;
long part_length;

int current_slice;

/* disk buffer parameters */
int buf_drive = -1;
int buf_track;

#ifndef	PC9800

int buf_geom;

#else

int buf_sectors;
struct disk_geometry * buf_geom = NULL;

struct bios_disk_address {
	Byte	s, h;
	Word	c;
} __attribute__((packed));	/* no paddings are allowed. */

struct partition_descriptor {
	Word	id;
	Word	__unused;	/* ? */
	struct bios_disk_address ipl, sta, end;
	Byte	name[16];
} __attribute__((packed));

extern struct partition_descriptor partition_table[16];

Byte partition_table_da_ua = -1;

#define DRIVE_EXIST_P(drive) \
	((unsigned int)(drive) < 0x100 && drive_lookup_table[(drive)] != 0xFF)
#define DRIVE_FLOPPY_P(drive)	DAUA_FLOPPY_P(drive)
#define GEOM_FLOPPY_P(geom) (((geom)->da_ua & 0x50) != 0)
#define SET_GEOMETRY(geom, l2ss, daua, sects, heads, total)	({	\
	struct disk_geometry *__g = (geom);				\
	Long __l __asm__("eax") = *(Long *)__g & MAKELONG(0, 0xF);	\
	*(Long *)__g = __l | MAKELONG((l2ss),(daua),(sects),(heads));	\
	__g->total_sectors = (total);					\
})
static inline unsigned int
chs_to_lba(struct disk_geometry *geom,
	   unsigned int cylinder, unsigned int head, unsigned int sector)
{
  return (cylinder * geom->heads + head) * geom->sectors + sector;
}

static const struct disk_geometry floppy_formats[] = {
  {  9, 0x30, 18, 2, 2880 },
  {  9, 0x10,  9, 2, 1440 },
  { 10, 0x90,  8, 2, 1232 },
  {  9, 0x90, 15, 2, 2400 },

  {  9, 0xB0, 18, 2, 2880 },
  {  9, 0x70,  9, 2, 1440 },
  { 10, 0xF0,  9, 2, 1232 },
  {  9, 0xF0, 15, 2, 2400 },
};

#endif

/* filesystem common variables */
int filepos;
int filemax;

int
rawread(int drive, int sector, int byte_offset, int byte_len, int addr)
{
#ifndef PC9800
  int slen = (byte_offset+byte_len+511)/SECTOR_SIZE;

  if (byte_len <= 0)
    return 1;

  while (byte_len > 0 && !errnum)
    {
      int soff, num_sect, bufaddr, track, size = byte_len;

      /*
       *  Check track buffer.  If it isn't valid or it is from the
       *  wrong disk, then reset the disk geometry.
       */
      if (buf_drive != drive)
	{
	  buf_geom = get_diskinfo(drive);
	  buf_drive = drive;
	  buf_track = -1;
	}

      if (buf_geom == 0)
	{
	  errnum = ERR_NO_DISK;
	  return 0;
	}

      /*  Get first sector of track  */
      soff = sector % SECTORS(buf_geom);
      track = sector - soff;
      num_sect = SECTORS(buf_geom) - soff;
      bufaddr = BUFFERADDR + (soff * SECTOR_SIZE) + byte_offset;

      if (track != buf_track)
	{
	  int bios_err, read_start = track, read_len = SECTORS(buf_geom);

	  /*
	   *  If there's more than one read in this entire loop, then
	   *  only make the earlier reads for the portion needed.  This
	   *  saves filling the buffer with data that won't be used!
	   */
	  if (slen > num_sect)
	    {
	      read_start = sector;
	      read_len = num_sect;
	      bufaddr = BUFFERADDR + byte_offset;
	    }

	  if ((bios_err = biosdisk(BIOSDISK_SUBFUNC_READ, drive, buf_geom,
				   read_start, read_len, BUFFERSEG)) != 0)
	    {
	      buf_track = -1;

	      if (bios_err == BIOSDISK_ERROR_GEOMETRY)
		errnum = ERR_GEOM;
	      else
		{
		  /*
		   *  If there was an error, try to load only the
		   *  required sector(s) rather than failing completely.
		   */
		  if (slen > num_sect
		      || biosdisk(BIOSDISK_SUBFUNC_READ, drive, buf_geom,
				  sector, slen, BUFFERSEG))
		    errnum = ERR_READ;

		  bufaddr = BUFFERSEG + byte_offset;
		}
	    }
	  else
	    buf_track = track;
	}

#ifndef NO_FANCY_STUFF
      /*
       *  Instrumentation to tell which sectors were read and used.
       */
      if (debug_fs && debug_fs_func)
	{
	  int sector_end = sector + ((num_sect < slen) ? num_sect : slen);
	  int sector_num = sector;

	  while (sector_num < sector_end)
	    (*debug_fs_func)(sector_num++);
	}
#endif  /* NO_FANCY_STUFF */

      if (size > ((num_sect * SECTOR_SIZE(buf_geom)) - byte_offset))
	size = (num_sect * SECTOR_SIZE(buf_geom)) - byte_offset;

      bcopy((char *)bufaddr, (char *)addr, size);

      addr += size;
      byte_len -= size;
      sector += num_sect;
      slen -= num_sect;
      byte_offset = 0;
    }
#else /* PC9800 */
  struct disk_geometry *geom;	/* DRIVE's disk geometry */
  unsigned int slen;		/* total read size in unit of sectors */
  int buf_max;			/* maximum number of sectors in buffer */
  int count = 0;
  int format;

#if 0
  printf("rawread(0x%2x, %u, %u, %u, 0x%8x)\n",
	 drive, sector, byte_len, byte_offset, addr);
#endif

  /* byte_len = 0 has special meaning; it means to read exactly one sector */
  if (byte_len < 0)
    return 1;

  if ((geom = get_geometry(drive)) == NULL)
    {
      errnum = ERR_NO_DISK;
      return 0;
    }

  /*		  DA =  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
  format = (((char []){ -3, 1,-2, 0,-3,-3,-3, 5,-1, 2,-1, 4,-3,-3,-3, 6 })
	    [geom->da_ua >> 4]);

  if (TOTAL_SECTORS(geom) == -1)	/* initial call on this device */
    {
      buf_geom = geom;
    disk_changed:
      switch (format)
	{
	case -1:		/* IDE/SCSI disk supported by Disk BIOS */
	  if (get_bios_geometry(buf_geom))
	    {
	      errnum = ERR_GEOM;
	      return 0;
	    }
	  break;
	case -2:		/* SCSI device not fully supported by BIOS */
	  while (1)
	    {
	      int error;
	      if ((error = read_capacity(buf_geom)) < 0)
		{
		  errnum = ERR_GEOM;
		  return 0;
		}
	      if (error != BIOSDISK_ERROR_UNITATTENSION)
		{
		  if (error == BIOSDISK_ERROR_TIMEOUT)
		    errnum = ERR_NO_DISK;
		  else if (error)
		    errnum = ERR_GEOM;
		  break;
		}
	    }
	  break;
	default:
	  if (format < 0)
	    panic("rawread: Unknown DA/UA = 0x%2x", buf_geom->da_ua);
	  
	set_floppy_format:
	  {
	    register Long tmp __asm__("eax");
	    register struct disk_geometry * g = buf_geom;
	    tmp = *(Long *)g & MAKELONG(0, 0x0F);
	    *(Long *)g = tmp | *(Long *)&floppy_formats[format];
	    g->total_sectors = floppy_formats[format].total_sectors;
	    DPRINTF(DEBUG_FLOPPY_PROBE, "Trying %uKB floppy...\n",
		    NSECTORS_SIZE(g, g->total_sectors) >> 10);
	  }
	}

      buf_sectors = 0;			/* discard track buffer */
      force_check_disk_change = 0;
    }
  else
    {
      if (format >= 0)
	{
	  if ((format & 2) && geom->log2_sector_size == 9)
	    format++;
	  check_interrupt();
	  if ((*(!(format & 4) ? &DISK_RESULT[geom->da_ua & 3].st0
		 : &F2DD_RESULT[geom->da_ua & 3].st0) & 0xC0) == 0xC0)
	    {
	      DPRINTF(DEBUG_DISK_CHANGE,
		      "VFS (): Disk change detected on device 0x%2x\n",
		      buf_geom->da_ua);
	      buf_sectors = 0;
	    }
	}

      if (buf_geom != geom)		/* drive changed */
	{
	  buf_sectors = 0;		/* discard track buffer */
	  buf_geom = geom;
	}
    }

  slen = (byte_len
	  ? TO_NR_SECTORS(buf_geom,
			  byte_offset + byte_len + SECTOR_SIZE(buf_geom) - 1)
	  : 1);
  buf_max = (format >= 0
	     ? SECTORS(buf_geom) * 2 : TO_NR_SECTORS(buf_geom, BUFFERSIZE));

  while (!errnum)
    {
      int num_sect, size;

      if (buf_track <= sector && sector < buf_track + buf_sectors)
	{
	  /*
	   * target sector is probably on track buffer.
	   * check disk change and validate it.
	   */
	  if (format == -2	/* SCSI device */
	      && force_check_disk_change
	      && test_unit_ready(buf_geom) == BIOSDISK_ERROR_UNITATTENSION)
	    {
	    disk_changed1:
	      DPRINTF(DEBUG_DISK_CHANGE,
		      "VFS (): Disk change detected on device 0x%2x\n",
		      buf_geom->da_ua);
	      goto disk_changed;
	    }
	}
      else
	{
	  /*
	   * target sector is not on track buffer; call BIOS to read it.
	   */
	  int bios_err;

	  buf_track = sector;
	  buf_sectors = buf_max;

	  if (format >= 0)	/* floppy disk */
	    {
	      int soff = sector % buf_max;

	      if (buf_max - soff < slen)
		buf_sectors -= soff;
	      else
		buf_track -= soff;	/* read from beginning of cylinder */
	    }
	  else
	    {
	      /* adjust to avoid reading past the end of disk */
	      if (buf_track + buf_sectors > TOTAL_SECTORS(geom))
		buf_sectors = TOTAL_SECTORS(geom) - buf_track;
	    }

	retry:
	  if ((bios_err = biosdisk(BIOSDISK_SUBFUNC_READ, buf_geom,
				   buf_track, buf_sectors, BUFFERSEG)) != 0)
	    {
	      buf_sectors = 0;

	      if (bios_err == BIOSDISK_ERROR_GEOMETRY)
		errnum = ERR_GEOM;
	      else if (count == 0)
		{
		  /*
		   * Error in first try. It needs special handling...
		   */
		  if (format >= 0)	/* floppy disk */
		    goto floppy_try_next_format;
		  else
		    {	/* SCSI/IDE device */
		      if (bios_err == BIOSDISK_ERROR_UNITATTENSION)
			goto disk_changed1;
		    }
		}
	      else if (count < 0 && -4 <= count)
		{
		floppy_try_next_format:
		  /*
		   * Don't retry on Not Ready:
		   * This will be ``actual'' hardware errors.
		   */
		  if ((bios_err & 0xF0) != 0x60)
		    {
		      format++;
		      if (!(format & 3))
			format -= 4;
		      count--;
		      goto set_floppy_format;
		    }
		}

	      /*
	       *  If there was an error, try to load only the
	       *  required sector(s) rather than failing completely.
	       */
	      if (sector + slen - buf_track < buf_sectors)
		{
		  buf_sectors = sector + slen - buf_track;
		  goto retry;
		}

	      errnum = ERR_READ;
	    }
#ifdef DEBUG
# define LIVE()	do { dprintf("line %d\n", __LINE__); getkey(); } while (0)
#else
# define LIVE()	do { } while (0)
#endif
	  count = (count < 0 ? 0 : count) + 1;
	}

      num_sect = buf_track + buf_sectors - sector;

#ifndef NO_FANCY_STUFF
      /*
       *  Instrumentation to tell which sectors were read and used.
       */
      if (debug_fs && debug_fs_func)
	{
	  int sector_end = sector + (num_sect < slen ? num_sect : slen);
	  int sector_num = sector;

	  while (sector_num < sector_end)
	    (*debug_fs_func)(sector_num++);
	}
#endif  /* NO_FANCY_STUFF */

      size = NSECTORS_SIZE(buf_geom, num_sect) - byte_offset;
      if (size > byte_len)
	size = (byte_len > 0 ? byte_len : SECTOR_SIZE(buf_geom));

      bcopy((char *)(BUFFERADDR + NSECTORS_SIZE(buf_geom, sector - buf_track)
		     + byte_offset),
	    (char *)addr, size);

      addr += size;
      sector += num_sect;
      slen -= num_sect;
      byte_offset = 0;
      if ((byte_len -= size) <= 0)
	break;
    }
#endif /* PC9800 */

  return (!errnum);
}


int
devread(int sector, int byte_offset, int byte_len, int addr)
{
#ifdef PC9800
  struct disk_geometry *geom = get_geometry(current_drive);
#endif
  /*
   *  Check partition boundaries
   */
#ifndef PC9800
  if (sector < 0
      || (sector + ((byte_offset+byte_len-1)/SECTOR_SIZE)) >= part_length)
#else
  if (sector < 0
      || (sector + TO_NR_SECTORS(geom, byte_offset + byte_len - 1)
	  >= part_length))
#endif
    {
      errnum = ERR_OUTSIDE_PART;
      return 0;
    }

  /*
   *  Get the read to the beginning of a partition.
   */
#ifndef PC9800
  while (byte_offset >= SECTOR_SIZE)
    {
      byte_offset -= SECTOR_SIZE;
      sector++;
    }
#else
  sector += TO_NR_SECTORS(geom, byte_offset);
  byte_offset &= (SECTOR_SIZE(geom) - 1);
#endif

#if !defined(NO_FANCY_STUFF) && defined(DEBUG) && !defined(PC9800)
  if (debug_fs)
    printf("<%d, %d, %d>", sector, byte_offset, byte_len);
#endif  /* !NO_FANCY_STUFF && DEBUG */

  /*
   *  Call "rawread", which is very similar, but:
   *
   *    --  It takes an extra parameter, the drive number.
   *    --  It requires that "sector" is relative to the beginning
   *            of the disk.
   *    --  It doesn't handle offsets of more than 511 bytes into the
   *            sector.
   */
  return rawread(current_drive, part_start+sector, byte_offset,
		 byte_len, addr);
}


int
sane_partition(void)
{
#ifndef PC9800
  if ( !(current_partition & 0xFF000000uL)
       && (current_drive & 0xFFFFFF7F) < 8
       && (current_partition & 0xFF) == 0xFF
       && ( (current_partition & 0xFF00) == 0xFF00
	    || (current_partition & 0xFF00) < 0x800 )
       && ( (current_partition >> 16) == 0xFF
	    || (current_drive & 0x80) ) )
    return 1;
#else
  Byte * cp = (Byte *)&current_partition;
  struct disk_geometry *geom;

  if (!cp[3]
      && (geom = get_geometry(current_drive)) != NULL
      && cp[0] == 0xFF
      && (cp[1] == 0xFF || cp[1] < 8)
      && (cp[2] == 0xFF
	  || (DAUA_HARDDISK_P(geom->da_ua)
	      && cp[2] < (SECTOR_SIZE(geom) / sizeof(partition_table[0])))))
    return 1;
#endif

  errnum = ERR_DEV_VALUES;
  return 0;
}


static void
attempt_mount(void)
{
  for ( fsys_type = 0; fsys_type < NUM_FSYS
	  && (*(fsys_table[fsys_type].mount_func))() != 1; fsys_type++);

  if (fsys_type == NUM_FSYS && errnum == ERR_NONE)
    errnum = ERR_FSYS_MOUNT;
}


#ifndef NO_FANCY_STUFF
int
make_saved_active(void)
{
#ifndef PC9800
  if (saved_drive & 0x80)
    {
      int part = saved_partition >> 16;

      if (part > 3)
	{
	  errnum = ERR_NO_PART;
	  return 0;
	}

      if (!rawread(saved_drive, 0, 0, SECTOR_SIZE, SCRATCHADDR))
	return 0;

      if (PC_SLICE_FLAG(SCRATCHADDR, part) != PC_SLICE_FLAG_BOOTABLE)
	{
	  int i;

	  for (i = 0; i < 4; i++)
	    PC_SLICE_FLAG(SCRATCHADDR, i) = 0;

	  PC_SLICE_FLAG(SCRATCHADDR, part) = PC_SLICE_FLAG_BOOTABLE;

	  buf_track = -1;

	  if (biosdisk(BIOSDISK_SUBFUNC_WRITE, saved_drive, buf_geom,
		       0, 1, SCRATCHSEG))
	    {
	      errnum = ERR_WRITE;
	      return 0;
	    }
	}
    }
#else
  panic("makeactive: ʤ!!");
#endif

  return 1;
}


static void
check_and_print_mount(void)
{
  attempt_mount();
  if (errnum == ERR_FSYS_MOUNT)
    errnum = ERR_NONE;
  if (!errnum)
    print_fsys_type();
  print_error();
}
#endif  /* NO_FANCY_STUFF */

static int
check_BSD_parts(int flags)
{
#ifndef PC9800
  char label_buf[SECTOR_SIZE];
#else
  char label_buf[MAX_SECTOR_SIZE];
#endif
  int part_no, got_part = 0;

  if ( part_length < (BSD_LABEL_SECTOR+1) )
    {
      errnum = ERR_BAD_PART_TABLE;
      return 0;
    }

#ifndef PC9800
  if (!rawread(current_drive, part_start + BSD_LABEL_SECTOR,
	       0, SECTOR_SIZE, (int) label_buf))
    return 0;
#else
  /*
   * XXX - How FreeBSD(98) and NetBSD/pc98 handle BSD label sector
   * on 1024 byte/sector floppy ???
   */
  if (!rawread(current_drive, part_start + BSD_LABEL_SECTOR,
	       0, 0, (int) label_buf))
    return 0;
  current_slice = PC_SLICE_TYPE_BSD;
#endif

  if ( BSD_LABEL_CHECK_MAG(label_buf) )
    {
      for (part_no = 0; part_no < BSD_LABEL_NPARTS(label_buf); part_no++)
	{
	  if (BSD_PART_TYPE(label_buf, part_no))
	    {
	      /* XXX should do BAD144 sector remapping setup here */

#ifndef PC9800
	      current_slice = ((BSD_PART_TYPE(label_buf, part_no) << 8)
			       | PC_SLICE_TYPE_BSD);
#else
	      *((unsigned char *)&current_slice + 2)
		= BSD_PART_TYPE(label_buf, part_no);
#endif
	      part_start = BSD_PART_START(label_buf, part_no);
	      part_length = BSD_PART_LENGTH(label_buf, part_no);

#ifndef NO_FANCY_STUFF
	      if (flags)
		{
		  if (!got_part)
		    {
		      printf(MSG2("[BSD sub-partitions immediately follow]\n",
				  "[BSD֥ѡƥ]\n"));
		      got_part = 1;
		    }
		  printf(MSG2("     BSD Partition num: \'%c\', ",
			      "     BSDê '%c': "), part_no + 'a');
		  check_and_print_mount();
		}
	      else
#endif  /* NO_FANCY_STUFF */
		if (part_no == ((current_partition >> 8) & 0xFF))
		  break;
	    }
	}

      if (part_no >= BSD_LABEL_NPARTS(label_buf) && !got_part)
	{
	  errnum = ERR_NO_PART;
	  return 0;
	}

      if ((current_drive & 0x80)
	  && BSD_LABEL_DTYPE(label_buf) == DTYPE_SCSI)
	bsd_evil_hack = 4;

      return 1;
    }

  errnum = ERR_BAD_PART_TABLE;
  return 0;
}

#ifndef PC9800
static char cur_part_desc[16];
#endif

#ifdef PC9800
void
print_partition_name (const void *const name)
{
  int i;
  unsigned int saved = *(unsigned int *) &putchar_x;
  const int max = sizeof (((struct partition_descriptor *) 0)->name);

  for (i = 0; i < max; i++)
    {
      unsigned char c = ((const unsigned char *) name)[i];

      putchar_flags = PUTCHAR_SJIS;
      putchar_attribute = 0xE1;

      if (!c)
	{
	  c = 0xA5;
	  goto nonprintable;
	}
      else if (c < ' ' || c == 0x7F || c == 0x80
	       || (((0x81 <= c && c <= 0xBF) || (0xE0 <= c && c <= 0xFC))
		   && (i >= max - 1
		       || ((const unsigned char *) name)[i + 1] < 0x40
		       || 0xFC < ((const unsigned char *) name)[i + 1]
		       || ((const unsigned char *) name)[i + 1] == 0x7F)))
	{
	nonprintable:
	  *(unsigned short *)&putchar_flags = 0;
	  putchar_attribute = 0x81;
	  putchar('\x8E');
	}
      putchar(c);
    }
  *(unsigned int *) &putchar_x = saved + max;
}
#endif

static int
real_open_partition(int flags)
{
#ifndef	PC9800
  char mbr_buf[SECTOR_SIZE];
  int i, part_no, slice_no, ext = 0, part_offset = 0;
#else
  char mbr_buf[MAX_SECTOR_SIZE];
  struct disk_geometry * geom;
  int i, part_no, slice_no, ext = 0;
#endif

#ifndef	PC9800
  /*
   *  The "rawread" is probably unnecessary here, but it is good to
   *  know it works.
   */
  if ( !sane_partition()
       || !rawread(current_drive, 0, 0, SECTOR_SIZE, (int) mbr_buf) )
    return 0;
#else
  if (!sane_partition()
      || !rawread(current_drive, 0, 0, 0, (int)mbr_buf))
    return 0;
#endif

  bsd_evil_hack = 0;
  current_slice = 0;
  part_start = 0;
#ifndef	PC9800
  part_length = SECTORS(buf_geom) * HEADS(buf_geom) * CYLINDERS(buf_geom);
#else
  geom = get_geometry(current_drive);
  part_length = TOTAL_SECTORS(geom);
#endif

#ifndef PC9800
  if (current_drive & 0x80)
#else
  if (DAUA_HARDDISK_P(geom->da_ua))
#endif
    {
      /*
       *  We're looking at a hard disk
       */

#ifndef PC9800
      int ext_offset = 0, part_offset = 0;
#endif
      part_no = (current_partition >> 16);
      slice_no = 0;

      /* if this is the whole disk, return here */
      if (!flags && current_partition == 0xFFFFFFuL)
#ifndef PC9800
	return 1;
#else
	goto try_open_floppy;
#endif

#ifndef PC9800
      /*
       *  Load the current MBR-style PC partition table (4 entries)
       */
      while ( slice_no < 255 && ext >= 0
	      && (part_no == 0xFF || slice_no <= part_no)
	      && rawread(current_drive, part_offset,
			 0, SECTOR_SIZE, (int) mbr_buf) )
#else
      if (!(geom->da_ua & 0x80))
	{
	  /*
	   * BIOS geometry information not available: we cannot
	   * convert disk address to LBA.
	   */
	  errnum = ERR_GEOM;
	  return 0;
	}

      if (*(Word *)(mbr_buf + SECTOR_SIZE(geom) - 2) == 0xAA55
	  && rawread(current_drive, 1, 0, 0, (int)partition_table))
#endif
	{
#ifndef PC9800
	  /*
	   *  If the table isn't valid, we can't continue
	   */
	  if ( !PC_MBR_CHECK_SIG(mbr_buf) )
	    {
	      errnum = ERR_BAD_PART_TABLE;
	      return 0;
	    }
#endif

	  ext = -1;

	  /*
	   *  Scan table for partitions
	   */
#ifndef PC9800
	  for (i = 0; i < PC_SLICE_MAX; i++)
#else
	  for (slice_no = 0;
	       slice_no < SECTOR_SIZE(geom) / sizeof(struct partition_descriptor);
	       slice_no++)
#endif
	    {
#if 0
	      current_partition = ((slice_no << 16)
		                   | (current_partition & 0xFFFF));
#else
	      *((unsigned short *)&current_partition + 1) = slice_no;
#endif
#ifndef PC9800
	      current_slice = PC_SLICE_TYPE(mbr_buf, i);
	      part_start = part_offset + PC_SLICE_START(mbr_buf, i);
	      part_length = PC_SLICE_LENGTH(mbr_buf, i);
	      bcopy(mbr_buf+PC_SLICE_OFFSET+(i<<4), cur_part_desc, 16);
#else
	      current_slice = partition_table[slice_no].id;
	      part_start = chs_to_lba(geom,
				      partition_table[slice_no].sta.c,
				      partition_table[slice_no].sta.h,
				      partition_table[slice_no].sta.s);
	      part_length = chs_to_lba(geom,
				       partition_table[slice_no].end.c + 1,
				       0, 0)
		- part_start;
#endif

	      /*
	       *  Is this PC partition entry valid?
	       */
#ifndef PC9800
	      if (current_slice)
#else
	      if (part_start)
#endif
		{
#ifndef PC9800
		  /*
		   *  Is this an extended partition?
		   */
		  if (current_slice == PC_SLICE_TYPE_EXTENDED)
#else
		  if (0)
#endif
		    {
		      if (ext == -1)
			{
			  ext = i;
			}
		    }
#ifndef NO_FANCY_STUFF
		  /*
		   *  Display partition information
		   */
		  else if (flags)
		    {
		      current_partition |= 0xFFFF;
#ifndef PC9800
		      printf("   Partition num: %d, ", slice_no);
#else /* PC9800 */
		      printf(MSG2("   Partition %2u (",
				  "   ê %2u ("), slice_no);
		      print_partition_name (partition_table[slice_no].name);
		      printf("): ");
#endif /* PC9800 */
		      if (!PC_SLICE_TYPE_CHECK(current_slice, BSD))
			check_and_print_mount();
		      else
			check_BSD_parts(1);
		      errnum = ERR_NONE;
		    }
#endif  /* NO_FANCY_STUFF */
		  /*
		   *  If we've found the right partition, we're done
		   */
		  else if (part_no == slice_no
			   || (part_no == 0xFF
			       && PC_SLICE_TYPE_CHECK(current_slice, BSD)))
		    {
		      if ((current_partition & 0xFF00) != 0xFF00)
			{
			  if (PC_SLICE_TYPE_CHECK(current_slice, BSD))
			    check_BSD_parts(0);
			  else
			    errnum = ERR_NO_PART;
			}

		      ext = -2;
		      break;
		    }
		}
#ifndef PC9800
	      /*
	       *  If we're beyond the end of the standard PC partition
	       *  range, change the numbering from one per table entry
	       *  to one per valid entry.
	       */
	      if (slice_no < PC_SLICE_MAX
		  || (current_slice != PC_SLICE_TYPE_EXTENDED
		      && current_slice != PC_SLICE_TYPE_NONE))
		slice_no++;
#endif
	    }

#ifndef PC9800
	  part_offset = ext_offset + PC_SLICE_START(mbr_buf, ext);
	  if (!ext_offset)
	    ext_offset = part_offset;
#endif
	}
    }
  else
    {
#ifdef PC9800
    try_open_floppy:
#endif
      /*
       *  We're looking at a floppy disk
       */
      ext = -1;
      if ((flags || (current_partition & 0xFF00) != 0xFF00)
	  && check_BSD_parts(flags))
	ext = -2;
      else
	{
	  errnum = 0;
	  if (!flags)
	    {
	      if (current_partition == 0xFFFFFF
		  || current_partition == 0xFF00FF)
		{
		  current_partition == 0xFFFFFF;	/* ??? */
		  ext = -2;
		}
	    }
#ifndef NO_FANCY_STUFF
	  else
	    {
	      current_partition = 0xFFFFFF;
	      check_and_print_mount();
	      errnum = 0;
	    }
#endif  /* NO_FANCY_STUFF */
	}
    }

  if (!flags && (ext != -2) && (errnum == ERR_NONE))
    errnum = ERR_NO_PART;

  if (errnum != ERR_NONE)
    return 0;

  return 1;
}


int
open_partition(void)
{
  return real_open_partition(0);
}


/* XX used for device completion in 'set_device' and 'print_completions' */
static char incomplete, disk_choice, part_choice;


char *
set_device(char *device)
{
  incomplete = 0;
  disk_choice = 1;
  part_choice = 0;
  current_drive = saved_drive;
  current_partition = 0xFFFFFF;

  if (*device == '(' && *(++device))
    {
      if (*device != ',' && *device != ')')
	{
	  char ch = *device;

	  if ((ch == 'f' || ch == 'h'
#ifdef	PC9800
	       || ch == 's'
#endif
	       ) && (device += 2, (*(device-1) != 'd')))
	    errnum = ERR_NUMBER_PARSING;

	  safe_parse_maxint(&device, (int*)&current_drive);

	  disk_choice = 0;
#ifndef PC9800
	  if (ch == 'h')
	    current_drive += 0x80;
#else
	  if (current_drive > 0xFE)	/* 0xFF is a magical number. */
	    errnum = ERR_DEV_FORMAT;
	  else if (ch == 'f')
	    {
	      int i = current_drive;
	      struct disk_geometry * geom = geometry_table + 0;

	      while (1)
		{
		  if (geom->da_ua != 0xFF)
		    {
		      if (i == 0)
			{
			  current_drive = geom->da_ua;
			  break;
			}
		      i--;
		    }
		  if (++geom >= geometry_table + 8)
		    {
		      errnum = ERR_NO_DISK;
		      break;
		    }
		}
	    }
	  else if (ch == 'h')
	    {
	      int cnt = current_drive;
	      struct disk_geometry * geom = geometry_table + 8;
	      struct disk_geometry *const sd_start = geometry_table + 12;

	      for (geom = geometry_table + 8; geom < sd_start; geom++)
		if (geom->da_ua != 0xFF)
		  {
		    if (cnt == 0)
		      {
			current_drive = geom->da_ua;
			break;
		      }
		    cnt--;
		  }
	      if (geom >= sd_start)
		{
		  Byte disk_equips = *DISK_EQUIPS & 0x7F;
		  current_drive = 0xA0;
		  while (1)
		    {
		      if (!disk_equips)
			{
			  errnum = ERR_NO_DISK;
			  break;
			}
		      if (disk_equips & 1)
			{
			  if (cnt == 0)
			    break;
			  cnt--;
			}
		      current_drive++;
		      disk_equips >>= 1;
		    }
		}
	    }
	  else if (ch == 's')
	    {
	      if ((Byte)current_drive > 6)
		errnum = ERR_DEV_FORMAT;
	      else
		(Byte)current_drive |= 0xA0;
	    }
#endif
	}

      if (errnum)
	return NULL;

      if (*device == ')')
	{
	  part_choice = 2;
	  return device + 1;
	}
      if (*device == ',')
	{
	  disk_choice = 0;
	  part_choice = 1;

	  if (*(++device) >= '0' && *device <= '9')
	    {
	      part_choice = 2;
	      current_partition = 0;

#ifndef PC9800
	      if (!(current_drive & 0x80)
		  || !safe_parse_maxint(&device, (int*)&current_partition)
		  || current_partition > 254)
#else
	      if (DRIVE_FLOPPY_P(current_drive)
		  || !safe_parse_maxint(&device, (int*)&current_partition)
		  || current_partition > (MAX_SECTOR_SIZE / 32))
#endif
		{
		  errnum = ERR_DEV_FORMAT;
		  return 0;
		}

	      current_partition = (current_partition << 16) + 0xFFFF;

	      if (*device == ','
		  && *(device+1) >= 'a' && *(device+1) <= 'h')
		{
		  device++;
		  goto set_bsd_partition;
		}
	    }
	  else if (*device >= 'a' && *device <= 'h')
	    {
	      part_choice = 2;
	    set_bsd_partition:
	      *((unsigned char *)&current_partition + 1) = *(device++) - 'a';
	    }

	  if (*device == ')')
	    {
	      if (part_choice == 1)
		{
		  current_partition = saved_partition;
		  part_choice = 2;
		}

	      return device + 1;
	    }
	}
    }

  if (!*device)
    incomplete = 1;
  errnum = ERR_DEV_FORMAT;

  return NULL;
}


/*
 *  This performs a "mount" on the current device, both drive and partition
 *  number.
 */

int
open_device(void)
{
  if (open_partition())
    attempt_mount();

  if (errnum != ERR_NONE)
    return 0;

  return 1;
}


#ifndef NO_FANCY_STUFF
int
set_bootdev(int hdbias)
{
  int i, j;

#ifndef PC9800
  /*
   *  Set chainloader boot device.
   */
  bcopy(cur_part_desc, (char *)(BOOTSEC_LOCATION-16), 16);
#endif

  /*
   *  Set BSD boot device.
   */
  i = (saved_partition >> 16) + 2;
  if (saved_partition == 0xFFFFFF)
    i = 1;
  else if ((saved_partition >> 16) == 0xFF)
    i = 0;

  /* XXX extremely evil hack!!! */
  j = 2;
  if (saved_drive & 0x80)
    j = bsd_evil_hack;

  return MAKEBOOTDEV( j, (i >> 4), (i & 0xF),
		      ((saved_drive - hdbias) & 0x79),
		      ((saved_partition >> 8) & 0xFF) );
}
#endif /* NO_FANCY_STUFF */


static char *
setup_part(char *filename)
{
  if (*filename == '(')
    {
      if ( (filename = set_device(filename)) == 0 )
	{
	  current_drive = 0xFF;
	  return 0;
	}
#ifndef NO_BLOCK_FILES
      if (*filename != '/')
	open_partition();
      else
#endif /* NO_BLOCK_FILES */
	open_device();
    }
  else if (saved_drive != current_drive
	   || saved_partition != current_partition
	   || (*filename == '/' && fsys_type == NUM_FSYS)
	   || buf_drive == -1)
    {
      current_drive = saved_drive;
      current_partition = saved_partition;
      /* allow for the error case of "no filesystem" after the partition
	 is found.  This makes block files work fine on no filesystem */
#ifndef NO_BLOCK_FILES
      if (*filename != '/')
        open_partition();
      else
#endif /* NO_BLOCK_FILES */
        open_device();
    }

  if (errnum && (*filename == '/' || errnum != ERR_FSYS_MOUNT))
    return 0;
  else
    errnum = 0;

  return sane_partition() ? filename : 0;
}


#ifndef NO_FANCY_STUFF
/*
 *  This prints the filesystem type or gives relevant information.
 */

void
print_fsys_type(void)
{
  printf(MSG2(" Filesystem type ", " ե륷ƥ"));

  if (fsys_type != NUM_FSYS)
    printf(MSG2("is %s\n", " %s Ǥ\n"), fsys_table[fsys_type].name);
  else
    {
      printf(MSG2("unknown, ", "μǤ ("));
      if (current_partition == 0xFFFFFF)
	printf(MSG2("using whole disk\n", "ǥΤѤޤ)\n"));
      else
#ifndef PC9800
	printf("partition type 0x%x\n", current_slice);
#else
	printf(MSG2("partition type (%2x, %2x)\n",
		    "ѡƥID = %2x %2x)\n"),
	       current_slice, current_slice >> 8);
#endif
    }
}

/*
 *  This lists the possible completions of a device string, filename, or
 *  any sane combination of the two.
 */

void
print_completions(char *filename)
{
  char *ptr = filename;

  if (*filename == '/' || (ptr = set_device(filename)) || incomplete)
    {
      errnum = 0;

      if (*filename != '/' && (incomplete || !*ptr))
	{
	  if (!part_choice)
	    {
	      /* disk completions */
#ifndef PC9800
	      int disk_no, i, j;
#else
	      int i;
	      struct disk_geometry * geom;
#endif

	      printf(MSG2(" Possible disks are:", " Ǥǥ:"));

#ifndef PC9800
	      for (i = 0; i < 2; i++)
		{
		  for (j = 0; j < 8; j++)
		    {
		      disk_no = (i * 0x80) + j;
		      if ((disk_choice || disk_no == current_drive)
			  && get_diskinfo(disk_no))
			printf(" %cd%d", (i ? 'h' : 'f'), j);
		    }
		}
#else
	      for( i = 0, geom = geometry_table ;
		   geom < geometry_table + 8 ; geom++ )
		if (geom->da_ua != 0xFF)
		  printf(" fd%d", i++);
	      for( i = 0 ; geom < geometry_table + 12 + 7 ; geom++ )
		if (geom->da_ua != 0xFF && (geom->da_ua & 0x80))
		  printf(" hd%d", i++);
#if 0
	      for( i = 0 ; geom < geometry_table + 12 + 7 ; geom++, i++ )
		if (geom->da_ua & 0x80)
		  printf(" sd%d", i);
#endif
#endif

	      putchar('\n');
	    }
	  else
	    {
	      /* partition completions */
	      if (part_choice == 1)
		{
		  printf(MSG2(" Possible partitions are:\n",
			      " Ǥѡƥ:\n"));
		  real_open_partition(1);
		}
	      else
		{
		  if (open_partition())
		    check_and_print_mount();
		}
	    }
	}
      else if (*ptr == '/')
	{
	  /* filename completions */
	  printf(MSG2(" Possible files are:", " Ǥե:"));
	  dir(filename);
	}
      else
	errnum = ERR_BAD_FILENAME;
    }

  print_error();
}
#endif /* NO_FANCY_STUFF */


/*
 *  This is the generic file open function.
 */

int
open(char *filename)
{
#ifndef NO_DECOMPRESSION
  compressed_file = 0;
#endif  /* NO_DECOMPRESSION */

  /* if any "dir" function uses/sets filepos, it must
     set it to zero before returning if opening a file! */
  filepos = 0;

  if (!(filename = setup_part(filename)))
    return 0;

#ifndef NO_BLOCK_FILES
  block_file = 0;
#endif /* NO_BLOCK_FILES */

  /* XXX to account for partial filesystem implementations! */
  fsmax = MAXINT;

  if (*filename != '/')
    {
#ifndef NO_BLOCK_FILES
      char *ptr = filename;
      int tmp, list_addr = BLK_BLKLIST_START;
      filemax = 0;

      while (list_addr < BLK_MAX_ADDR)
	{
	  tmp = 0;
	  safe_parse_maxint(&ptr, &tmp);
	  errnum = 0;

	  if (*ptr != '+')
	    {
	      if ((*ptr && *ptr != '/' && !isspace(*ptr))
		  || tmp == 0 || tmp > filemax)
		errnum = ERR_BAD_FILENAME;
	      else
		filemax = tmp;

	      break;
	    }

	  /* since we use the same filesystem buffer, mark it to
	     be remounted */
	  fsys_type = NUM_FSYS;

	  BLK_BLKSTART(list_addr) = tmp;
	  ptr++;

	  if (!safe_parse_maxint(&ptr, &tmp)
	      || tmp == 0
	      || (*ptr && *ptr != ',' && *ptr != '/' && !isspace(*ptr)))
	    {
	      errnum = ERR_BAD_FILENAME;
	      break;
	    }

	  BLK_BLKLENGTH(list_addr) = tmp;

#ifndef PC9800
	  filemax += (tmp * SECTOR_SIZE);
#else
	  filemax += tmp * SECTOR_SIZE(buf_geom);
#endif
	  list_addr += BLK_BLKLIST_INC_VAL;

	  if (*ptr != ',')
	    break;

	  ptr++;
	}

      if (list_addr < BLK_MAX_ADDR && ptr != filename && !errnum)
	{
	  block_file = 1;
	  BLK_CUR_FILEPOS = 0;
	  BLK_CUR_BLKLIST = BLK_BLKLIST_START;
	  BLK_CUR_BLKNUM = 0;

#ifndef NO_DECOMPRESSION
	  return gunzip_test_header();
#else   /* NO_DECOMPRESSION */
	  return 1;
#endif  /* NO_DECOMPRESSION */
	}
#else   /* NO_BLOCK_FILES */
      errnum = ERR_BAD_FILENAME;
#endif  /* NO_BLOCK_FILES */
    }

  if (!errnum && fsys_type == NUM_FSYS)
    errnum = ERR_FSYS_MOUNT;

  /* set "dir" function to open a file */
  print_possibilities = 0;

  if (!errnum && (*(fsys_table[fsys_type].dir_func))(filename))
    {
#ifndef NO_DECOMPRESSION
      return gunzip_test_header();
#else   /* NO_DECOMPRESSION */
      return 1;
#endif  /* NO_DECOMPRESSION */
    }

  return 0;
}


int
grub_read(void * addr, int len)
{
  /* Make sure "filepos" is a sane value */
  if (filepos < 0 || filepos > filemax)
    filepos = filemax;

  /* Make sure "len" is a sane value */
  if (len < 0 || len > (filemax - filepos))
    len = filemax - filepos;

  /* if target file position is past the end of
     the supported/configured filesize, then
     there is an error */
  if (filepos+len > fsmax)
    {
      errnum = ERR_FILELENGTH;
      return 0;
    }

#ifndef NO_DECOMPRESSION
  if (compressed_file)
    return gunzip_read(addr, len);
#endif  /* NO_DECOMPRESSION */

#ifndef NO_BLOCK_FILES
  if (block_file)
    {
      int size, off, ret = 0;
#ifdef PC9800
      unsigned int sector_size = SECTOR_SIZE(buf_geom);
#endif

      while (len && !errnum)
	{
#ifdef PC9800
	  int blknum = TO_NR_SECTORS(buf_geom, filepos);
	  int tmp;
#endif
	  /* we may need to look for the right block in the list(s) */
#ifndef PC9800
	  if (filepos < BLK_CUR_FILEPOS)
#else
	  if (blknum < BLK_CUR_BLKNUM)
#endif
	    {
#ifndef PC9800
	      BLK_CUR_FILEPOS = 0;
#endif
	      BLK_CUR_BLKLIST = BLK_BLKLIST_START;
	      BLK_CUR_BLKNUM = 0;
	    }

#ifndef PC9800
	  /* run BLK_CUR_FILEPOS up to filepos */
	  while ( filepos > BLK_CUR_FILEPOS )
	    {
	      if ( (filepos - (BLK_CUR_FILEPOS & ~(SECTOR_SIZE - 1)))
		   >= SECTOR_SIZE )
		{
		  BLK_CUR_FILEPOS += SECTOR_SIZE;
		  BLK_CUR_BLKNUM++;

		  if ( BLK_CUR_BLKNUM >= BLK_BLKLENGTH(BLK_CUR_BLKLIST) )
		    {
		      BLK_CUR_BLKLIST += BLK_BLKLIST_INC_VAL;
		      BLK_CUR_BLKNUM = 0;
		    }
		}
	      else
		BLK_CUR_FILEPOS = filepos;
	    }

	  off = filepos & (SECTOR_SIZE - 1);
	  size = ( ( BLK_BLKLENGTH(BLK_CUR_BLKLIST) - BLK_CUR_BLKNUM )
		   * SECTOR_SIZE ) - off;

#else /* PC9800 */
	  while (blknum
		 >= (tmp = BLK_CUR_BLKNUM + BLK_BLKLENGTH(BLK_CUR_BLKLIST)))
	    {
	      BLK_CUR_BLKNUM = tmp;
	      BLK_CUR_BLKLIST += BLK_BLKLIST_INC_VAL;
	    }
	  off = filepos & (sector_size - 1);
	  size = NSECTORS_SIZE(buf_geom, tmp - blknum) - off;
#endif /* PC9800 */

	  if (size > len)
	    size = len;

#ifndef NO_FANCY_STUFF
	  debug_fs_func = debug_fs;
#endif  /* NO_FANCY_STUFF */

	  /* read current block and put it in the right place in memory */
#ifndef PC9800
	  devread(BLK_BLKSTART(BLK_CUR_BLKLIST) + BLK_CUR_BLKNUM,
		  off, size, (int)addr);
#else
	  devread(blknum - BLK_CUR_BLKNUM + BLK_BLKSTART(BLK_CUR_BLKLIST),
		  off, size, (int)addr);
#endif

#ifndef NO_FANCY_STUFF
	  debug_fs_func = NULL;
#endif  /* NO_FANCY_STUFF */

	  len -= size;
	  filepos += size;
	  ret += size;
	  addr += size;
	}

      if (errnum)
	ret = 0;

      return ret;
    }
#endif  /* NO_BLOCK_FILES */

  if (fsys_type == NUM_FSYS)
    {
      errnum = ERR_FSYS_MOUNT;
      return 0;
    }

  return (*(fsys_table[fsys_type].read_func))((int)addr, len);
}


#ifndef NO_FANCY_STUFF
int
dir(char *dirname)
{
#ifndef NO_DECOMPRESSION
  compressed_file = 0;
#endif  /* NO_DECOMPRESSION */

  if (!(dirname = setup_part(dirname)))
    return 0;

  if (*dirname != '/')
    errnum = ERR_BAD_FILENAME;

  if (fsys_type == NUM_FSYS)
    errnum = ERR_FSYS_MOUNT;

  if (errnum)
    return 0;

  /* set "dir" function to list completions */
  print_possibilities = 1;

  return (*(fsys_table[fsys_type].dir_func))(dirname);
}
#endif /* !NO_FANCY_STUFF */
