/*
 *  ISO 9660 filesystem backend for GRUB (GRand Unified Bootloader)
 *  including Rock Ridge Extensions support
 *
 *  Copyright (C) 1998, 1999  Kousuke Takai  <tak@kmc.kyoto-u.ac.jp>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 *  References:
 *	linux/fs/isofs/rock.[ch]
 *	mkisofs-1.11.1/diag/isoinfo.c
 *	mkisofs-1.11.1/iso9660.h
 *		(all are written by Eric Youngdale)
 */

#ifdef	TEST

#define	ERR_FSYS_CORRUPT	1
#define	ERR_FILE_NOT_FOUND	2
#define	ERR_BAD_FILETYPE	3
#define	ERR_READ		4

#define	DEBUG_FILESYSTEM	1

#define	DEBUG

#ifdef __ELF__
# define EXT_C(x)	x
# define L(x)		.L ## x
#else /* !ELF */
# define EXT_C(x)	_ ## x
# define L(x)		L ## x
#endif
#define ENTRY(x)	.global EXT_C(x); EXT_C(x) ## :

#include "pc_slice.h"

#else	/* !TEST */

#define ASM_FILE

#include "shared.h"
#include "pc_slice.h"

#endif	/* !TEST */

/*
 * ISO-9660 driver will use `block file' backend.
 */
#define	BLK_BLKLIST_START		(FSYS_BUF + 12)
#define	PRIMARY_DESCRIPTPOR_BUF		(FSYS_BUF + 2048)
#define	DIRECTORY_RECORD_BUF		(FSYS_BUF + 4096)
#define	ROCKRIDGE_EXTENT_BUF		(FSYS_BUF + 6144)
#define	ROCKRIDGE_NAME_BUF		(FSYS_BUF + 8192)

#define	ISO_REGULAR	1	/* regular file	*/
#define	ISO_DIRECTORY	2	/* directory	*/
#define	ISO_OTHER	0	/* other file (with Rock Ridge) */

#define	RR_FLAG_PX	0x01	/* have POSIX file attributes */
#define	RR_FLAG_NM	0x08	/* have alternate file name   */

/* POSIX file attributes for Rock Ridge extensions */
#define	POSIX_S_IFMT	0xF000
#define	POSIX_S_IFREG	0x8000
#define	POSIX_S_IFDIR	0x4000

/* volume descriptor types */
#define ISO_VD_PRIMARY 1
#define ISO_VD_END 255

#define ISO_STANDARD_ID "CD001"

/* primary descriptor offsets */
#define	ISO_PD_TYPE			   0	/* 711 */
#define	ISO_PD_ID			   1
#define	ISO_PD_VERSION			   6	/* 711 */
/*	ISO_PD_UNUSED1			   7 */
#define	ISO_PD_SYSTEM_ID		   8	/* achars */
#define	ISO_PD_VOLUME_ID		  40	/* dchars */
/*	ISO_PD_UNUSED2			  72 */
#define	ISO_PD_VOLUME_SPACE_SIZE	  80	/* 733 */
/*	ISO_PD_UNUSED3			  88 */
#define	ISO_PD_VOLUME_SET_SIZE		 120	/* 723 */
#define	ISO_PD_VOLUME_SEQUENCE_NUMBER	 124	/* 723 */
#define	ISO_PD_LOGICAL_BLOCK_SIZE	 128	/* 723 */
#define	ISO_PD_PATH_TABLE_SIZE		 132	/* 733 */
#define	ISO_PD_TYPE_L_PATH_TABLE	 140	/* 731 */
#define	ISO_PD_OPT_TYPE_L_PATH_TABLE	 144	/* 731 */
#define	ISO_PD_TYPE_M_PATH_TABLE	 148	/* 732 */
#define	ISO_PD_OPT_TYPE_M_PATH_TABLE	 152	/* 732 */
#define	ISO_PD_ROOT_DIRECTORY_RECORD	 156	/* 9.1 */
#define	ISO_PD_VOLUME_SET_ID		 190	/* dchars */
#define	ISO_PD_PUBLISHER_ID		 318	/* achars */
#define	ISO_PD_PREPARER_ID		 446	/* achars */
#define	ISO_PD_APPLICATION_ID		 574	/* achars */
#define	ISO_PD_COPYRIGHT_FILE_ID	 702	/* 7.5 dchars */
#define	ISO_PD_ABSTRACT_FILE_ID		 739	/* 7.5 dchars */
#define	ISO_PD_BIBLIOGRAPHIC_FILE_ID	 776	/* 7.5 dchars */
#define	ISO_PD_CREATION_DATE		 813	/* 8.4.26.1 */
#define	ISO_PD_MODIFICATION_DATE	 830	/* 8.4.26.1 */
#define	ISO_PD_EXPIRATION_DATE		 847	/* 8.4.26.1 */
#define	ISO_PD_EFFECTIVE_DATE		 864	/* 8.4.26.1 */
#define	ISO_PD_FILE_STRUCTURE_VERSION	 881	/* 711 */
/*	ISO_PD_UNUSED4			 882 */
#define	ISO_PD_APPLICATION_DATA		 883
/*	ISO_PD_UNUSED5			1395 */
/*				total	2048 bytes */

/* directory record offsets */
#define	ISO_DR_LENGTH			   0	/* 711 */
#define	ISO_DR_EXT_ATTR_LENGTH		   1	/* 711 */
#define	ISO_DR_EXTENT			   2	/* 733 */
#define	ISO_DR_SIZE			  10	/* 733 */
#define	ISO_DR_DATE			  18	/* 7 by 711 */
#define	ISO_DR_FLAGS			  25
#define	ISO_DR_FILE_UNIT_SIZE		  26	/* 711 */
#define	ISO_DR_INTERLEAVE		  27	/* 711 */
#define	ISO_DR_VOLUME_SEQUENCE_NUMBER	  28	/* 723 */
#define	ISO_DR_NAME_LEN			  32	/* 711 */
#define	ISO_DR_NAME			  33

#define	LONGWORD(b0,b1,b2,b3)	((b0)|((b1)<<8)|((b2)<<16)|((b3)<<24))
#define	WORD(b0,b1)		((b0)|((b1)<<8))

#define	ALIGN_FUNCTION	.balign 4
#define	ALIGN_LOOP	.balign 4
#define	ALIGN_JUMP	.balign	4

	ALIGN_FUNCTION
ENTRY(iso9660_mount)
	/*
	 *  Because there is no defined slice type ID for ISO-9660 filesystem,
	 *  this test will pass only either (1) if entire disk is used, or
	 *  (2) if current partition is BSD style sub-partition whose ID is
	 *  ISO-9660.
	 */

	cmpw	$-1,EXT_C(current_partition)+1
	je	1f
#ifndef PC9800
	cmpl	$(PC_SLICE_TYPE_BSD | (FS_ISO9660 << 8)), EXT_C(current_slice)
#else
	cmpl	$(PC_SLICE_TYPE_BSD | (FS_ISO9660 << 16)), EXT_C(current_slice)
#endif
	jne	L(mount_failed)
    1:	pushl	%ebx
	pushl	%esi
	movl	$16, %esi
	movl	$ PRIMARY_DESCRIPTPOR_BUF, %ebx
L(mount_loop):

	/*
	 * At this point, buf_geom will be Ok because real_open_partition()
	 * has already tried rawread() on the device.
	 */
#ifndef PC9800
# warning ISO-9660 driver on non-PC-9800 is currently not completed
	/*
	 * XXX sector size ?
	 */
	pushl	%ebx			/* load address */
	movl	%esi,%eax
	pushl	$2048
	shll	$2,%eax
#else
	movl	EXT_C(buf_geom),%eax
	movb	$11,%cl
	pushl	%ebx			/* load address */
	subb	(%eax),%cl
	movl	%esi,%eax
	pushl	$2048			/* byte length */
	shll	%cl,%eax
#endif
	pushl	$0			/* byte offset */
	pushl	%eax			/* real sector # */
	call	EXT_C(devread)
	addl	$16,%esp		/* pop off all arguments */
	testl	%eax,%eax
	jz	L(mount_failed_pop)
	/*
	 * check against ISO_VD_PRIMARY and ISO_STANDARD_ID ("CD001")
	 */
	cmpl	$ LONGWORD(ISO_VD_PRIMARY,'C','D','0'), (%ebx)
	jne	1f
	cmpw	$ WORD('0','1'), 4(%ebx)
	je	L(mount_success)
    1:	incl	%esi
	cmpl	$32,%esi
	jb	L(mount_loop)
L(mount_failed_pop):
	popl	%esi
	popl	%ebx
L(mount_failed):
	xorl	%eax,%eax
	ret

	ALIGN_JUMP
L(mount_success):
#ifdef DEBUG
# ifdef PC9800
	testb	$ DEBUG_FILESYSTEM,EXT_C(debug_flags)
	jz	0f
	pushl	%esi
	pushl	$1f
	call	EXT_C(dprintf)
	popl	%eax
	popl	%esi

	.section .rodata
    1:	.string	"iso9660: primary descriptor found at sector %u\n"
	.previous

    0:
# endif /* PC9800 */
#endif /* DEBUG */
	xorl	%eax,%eax
	popl	%esi
	popl	%ebx
	incl	%eax		/* %eax = 1 */
	ret

	.section .rodata
L(dotdot):
	.ascii	".."
	.previous

/* GAS's preprocessor treats '/' as coment starter... */
#define	SLASH	0x2F

	ALIGN_FUNCTION
ENTRY(iso9660_dir)
	pushl	%ebp
	pushl	%edi
	pushl	%esi
	pushl	%ebx

	movl	20(%esp),%edi		/* get (char *)DIRNAME */
	movl	$ PRIMARY_DESCRIPTPOR_BUF + ISO_PD_ROOT_DIRECTORY_RECORD, %esi

	/*
	 *  depth_loop is iterated when entering deeper sub-directory.
	 */
	ALIGN_LOOP
L(depth_loop):
	/* skip leading slashes of DIRNAME */
	xorl	%ebx,%ebx
#if 0
    1:	cmpb	$ SLASH,(%edi)
	jne	1f
	incl	%edi
	jmp	1b
	ALIGN_JUMP
#else
	cld
	movb	$ SLASH,%al
    1:	scasb
	je	1b
	decl	%edi
#endif
	/*
	 *  count the number of characters until next slash (or the end
	 *  of DIRNAME) in %ebx
	 *  (roughly equivarent to `%ebx = strcspn(%edi, "/\n\t ")')
	 */
    1:	movb	(%edi,%ebx),%al
	incl	%ebx			/* adjusted later */
	testb	%al,%al			/* end of DIRNAME */
	jz	1f
	cmpb	$ SLASH,%al
	je	1f
	cmpb	$' ',%al
	je	1f
	cmpb	$'\n',%al
	je	1f
	cmpb	$'\t',%al
	jne	1b
	ALIGN_JUMP
    1:	decl	%ebx			/* adjust length */
	movl	ISO_DR_SIZE(%esi),%eax
	movl	ISO_DR_EXTENT(%esi),%edx
	testl	%eax,%eax
	
	jle	L(sector_loop_exit)

	/*
	 *  sector_loop is for traversing all sectors that contain
	 *  directory-record of the same depth.
	 */
	ALIGN_LOOP
L(sector_loop):
	movl	$ DIRECTORY_RECORD_BUF,%esi
	pushl	%eax			/* save ISO_DR_SIZE */
	pushl	%edx			/* save ISO_DR_EXTENT */
	/* load one sector to buffer */
#ifndef PC9800
	pushl	%esi
	shll	$2,%edx			/* XXX */
	pushl	$2048
	pushl	$0
#else
	movl	EXT_C(buf_geom),%eax
	movb	$11,%cl
	pushl	%esi
	pushl	$2048
	subb	(%eax),%cl
	pushl	$0
	shll	%cl,%edx
#endif
	pushl	%edx
	call	EXT_C(devread)
	addl	$16,%esp		/* pop off all arguments */
	testl	%eax,%eax
	jz	L(dir_failed_pop2)	/* devread failed */

	/*
	 *  record_loop traverses all directory-records in one sector.
	 */
	ALIGN_LOOP
L(record_loop):
	cmpb	$0,(%esi)		/* == DR_LENGTH(%esi) */
	je	L(record_loop_exit)
	movl	%esi,%edx
	pushl	%esi			/* save pointer to directory record */
	pushl	%ebx			/* save PATHLEN */
	pushl	%edi			/* save DIRNAME */
	testb	$0x02,ISO_DR_FLAGS(%edx)
	leal	ISO_DR_NAME(%edx),%esi
	setnz	%bl
	movzbl	ISO_DR_NAME_LEN(%edx),%eax
	incb	%bl			/* %bl = file_type */
	movl	%eax,%ecx		/* save original name_len */
	cmpb	$1,%al			/* %eax never exceed 255 */
	jne	1f
	cmpb	$1,(%esi)		/* `^@' -> this directory (.) */
	ja	1f
	movl	$ L(dotdot)+1,%esi
	jne	1f
	incl	%eax			/* `^A' -> parent directory (..) */
	decl	%esi
    1:	cmpb	$2,%al
	jbe	1f
	cmpw	$ WORD(';','1'),-2(%esi,%eax)
	jne	1f
	decl	%eax			/* truncate trailing file version */
	decl	%eax
	cmpb	$1,%al
	jbe	1f
	cmpb	$'.',-1(%esi,%eax)
	jne	1f
	decl	%eax			/* truncate trailing dot */
    1:	pushl	%eax			/* save NAME_LEN */

	/*
	 *  Check if there is Rock-Ridge extension field and perse it if any.
	 */
	leal	ISO_DR_NAME(%edx,%ecx),%edi
	movzbl	(%edx),%edx		/* == DR_LENGTH(%esi) */
	subl	%ecx,%edx		/* subtract original NAME_LEN */
	movl	%edi,%eax
	subl	$ ISO_DR_NAME,%edx
	testb	$1,%al
	jz	1f
	decl	%edx			/* align to even address */
	incl	%edi
    1:	cmpl	$5,%edx
	jl	L(rr_exit)		/* too short to contain RR info */
	movb	$(RR_FLAG_NM | RR_FLAG_PX),%bh
	xorl	%ebp,%ebp		/* %ebp points continued extent */
	ALIGN_LOOP
L(rr_loop):
	movw	(%edi),%ax		/* get RR identifier in %ax */
	movzbl	2(%edi),%ecx		/* %ecx = length of this chunk */

	cmpb	$1,3(%edi)
	jne	L(rr_next)		/* invalid RockRidge version, ignore */
	cmpw	$ WORD('R','R'),%ax	/* Rock Ridge extension flags */
	jne	1f
	cmpb	$5,%cl
	jb	L(rr_next)
	andb	4(%edi),%bh
	jmp	3f

	ALIGN_JUMP
    1:	cmpw	$ WORD('N','M'),%ax	/* long filename */
	jne	1f
	movb	%cl,%al
	subb	$5,%al
	jbe	L(rr_next)
	movb	%al,(%esp)		/* alter NAME_LEN */
	leal	5(%edi),%esi		/* alter NAME */
	andb	$~RR_FLAG_NM,%bh
	jmp	3f

	ALIGN_JUMP
    1:	cmpw	$ WORD('P','X'),%ax	/* posix file attributes */
	jne	1f
	cmpb	$36,%cl
	jb	L(rr_next)
	movb	5(%edi),%al		/* MSB of file mode */
	movb	$ ISO_DIRECTORY,%bl
	andb	$0xF0,%al
	cmpb	$(POSIX_S_IFDIR >> 8),%al
	je	2f
	cmpb	$(POSIX_S_IFREG >> 8),%al
	sete	%bl			/* assume ISO_REGULAR == 1 */
  2:	andb	$~RR_FLAG_PX, %bh
  3:	jz	L(rr_exit)
	jmp	L(rr_next)

	ALIGN_JUMP
  1:	cmpw	$ WORD('C','E'),%ax	/* continuous extent */
	jne	L(rr_next)
	cmpb	$28,%cl
	jb	L(rr_next)
	movl	%edi,%ebp
L(rr_next):
	subl	%ecx,%edx		/* subtract from total length */
	addl	%ecx,%edi		/* advance record pointer */
	cmpl	$4,%edx
	jae	L(rr_loop)
	testl	%ebp,%ebp
	jz	L(rr_exit)		/* no continuous extent */
	/*
	 *  if NAME (currently on stack, NAME_LEN bytes) overlaps
	 *  ROCKRIDGE_EXTENT_BUF, we must save NAME on other place.
	 */
	cmpl	$ ROCKRIDGE_EXTENT_BUF,%esi
	jb	1f
	cmpl	$ ROCKRIDGE_EXTENT_BUF+2048,%esi
	jae	1f
	movl	$ ROCKRIDGE_NAME_BUF,%edi
	movl	(%esp),%ecx		/* get NAME_LEN */
	pushl	%edi			/* new NAME */
	cld
	rep; movsb
	popl	%esi			/* %esi = ROCKRIDGE_NAME_BUF */
    1:	movl	4(%ebp),%edx
	movl	12(%ebp),%edi
	movl	20(%ebp),%ebp
	movl	%edi,%eax
	addl	%ebp,%eax
	cmpl	$2048,%eax
	jae	L(rr_exit)		/* we cannot handle too large CE */
#ifndef PC9800
	pushl	$ ROCKRIDGE_EXTENT_BUF
	pushl	$2048
	shll	$2,%edx
#else
	movl	EXT_C(buf_geom),%eax
	movb	$11,%cl
	pushl	$ ROCKRIDGE_EXTENT_BUF
	subb	(%eax),%cl
	pushl	$2048
	shll	%cl,%edx
#endif
	pushl	$0
	pushl	%edx
	call	EXT_C(devread)
	addl	$16,%esp		/* pop off arguments */
	movb	$0,EXT_C(errnum)	/* error is not fatal */
	testl	%eax,%eax
	jz	L(rr_exit)		/* devread failed */
	addl	$ ROCKRIDGE_EXTENT_BUF,%edi
	movl	%ebp,%edx
	xorl	%ebp,%ebp
	cmpl	$5,%edx
	jae	L(rr_loop)
	ALIGN_JUMP
L(rr_exit):
	popl	%eax			/* name_len */
	popl	%edi			/* dirname */
	popl	%ecx			/* pathlen */
	cmpl	%ecx,%eax
	jb	L(next_file)		/* pathlen > name_len */
	pushl	%ecx
	pushl	%esi
	pushl	%edi
	cld
	testb	$0,%al			/* work even if %ecx == 0 */
	rep; cmpsb
	popl	%edi
	popl	%esi
	popl	%ecx
	jne	L(next_file)		/* no match, try the next file */
	cmpb	$ SLASH,(%edi,%ecx)
	jne	L(last_element)
	/*
	 *  %edi points names of intermediate directory. They must match
	 *  exactly and the file must be of directory type.
	 */
	cmpl	%ecx,%eax		/* exact match? */
	jne	L(next_file)		/* no, try the next file */
	popl	%esi			/* pointer to directory record */
	cmpb	$ ISO_DIRECTORY,%bl
	jne	L(bad_filetype)		/* not a directory */
	addl	%ecx,%edi		/* make %edi points next slash */
	popl	%ebx			/* discard extent */
	popl	%eax			/* discard size */
	jmp	L(depth_loop)		/* enter deeper sub directory */

	ALIGN_JUMP
	/*
	 *  No slash is following this pathname element:
	 *  if printing possible file name, print the name even if partial
	 *  match. Otherwise, specified pathname element and the name on
	 *  directory-record must match exactly.
	 */
L(last_element):
	cmpb	$0,EXT_C(print_possibilities)
	je	L(not_print_possibilities)
	pushl	%ecx
#if defined(TEST) || !defined(PC9800)
	movl	%eax,%ebx
	testl	%eax,%eax
	jz	2f			/* length == 0 (???) */
	pushl	$' '			/* put two whitespaces */
	call	EXT_C(putchar)
	popl	%eax
	pushl	$' '
	call	EXT_C(putchar)
	popl	%eax
    3:	movzbl	(%esi),%eax		/* print filename */
	incl	%esi
	pushl	%eax
	call	EXT_C(putchar)
	popl	%eax
	decl	%ebx
	jnz	3b
#else /* !TEST && PC9800 */
	movl	%eax,%ecx
	jecxz	2f
	movb	$' ',%al
	call	EXT_C(putchar)
	call	EXT_C(putchar)
    3:	lodsb
	call	EXT_C(putchar)
	loop	3b
#endif
	/* mark that we have listed at least one name */
	movb	$-1,EXT_C(print_possibilities)
    2:	popl	%ecx
	/* jmp	L(next_file) */
	ALIGN_JUMP
L(next_file):
	popl	%esi			/* pointer to directory record */
	movzbl	(%esi),%eax		/* get DR_LENGTH */
	movl	%ecx,%ebx		/* restore PATHLEN */
	addl	%eax,%esi		/* advance pointer */
	cmpl	$ DIRECTORY_RECORD_BUF+2048,%esi	/* sanity check */
	jb	L(record_loop)

	ALIGN_JUMP
L(record_loop_exit):
	popl	%edx			/* restore extent */
	popl	%eax			/* restore size */
	incl	%edx			/* next extent */
	subl	$2048,%eax
	ja	L(sector_loop)
	ALIGN_JUMP
	/*
	 *  We have parsed all directory records. If we had listed at least
	 *  one possible files, it is not an error. Otherwise, it means 
	 *  we cannot find the specified file.
	 */
L(sector_loop_exit):
	cmpb	$0,EXT_C(print_possibilities)
	jge	L(file_not_found)
#if defined(TEST) || !defined(PC9800)
	pushl	$'\n'
	call	EXT_C(putchar)
	popl	%eax
#else
	movb	$'\n',%al
	call	EXT_C(putchar)
#endif
	jmp	L(dir_success)

	ALIGN_JUMP
L(file_not_found):
	movb	$ ERR_FILE_NOT_FOUND,EXT_C(errnum)
	jmp	L(dir_failed)

	ALIGN_JUMP
L(not_print_possibilities):
	cmpl	%ecx,%eax		/* exact match? */
	jne	L(next_file)		/* no, try the next file */
	popl	%esi			/* pointer to directory record */
	cmpb	$ ISO_REGULAR,%bl
	jne	L(bad_filetype)		/* not a regular file, giving up */
	movb	$1,EXT_C(block_file)
	popl	%edx			/* discard extent */
	popl	%eax			/* discard size */
	movl	ISO_DR_EXTENT(%esi),%edx
#ifndef PC9800
	movl	$ BLK_BLKLIST_START,%edi
	shll	$2,%edx
	xorl	%ebx,%ebx
#else
	movl	EXT_C(buf_geom),%eax
	movb	$11,%cl
	movl	$ BLK_BLKLIST_START,%edi
	subb	(%eax),%cl
	xorl	%ebx,%ebx
	shll	%cl,%edx		/* %edx = real starting sector # */
#endif
	movl	%edx,(%edi)		/* BLK_BLKSTART(BLK_BLKLIST_START) */
#ifdef PC9800
	movb	(%eax),%cl
#endif
	movl	%ebx,-12(%edi)		/* clear BLK_CUR_FILEPOS */
	movl	ISO_DR_SIZE(%esi),%eax
	movl	%ebx,-4(%edi)		/* clear BLK_CUR_BLKNUM */
	movl	%eax,EXT_C(filemax)
	movl	%eax,EXT_C(fsmax)
#ifndef PC9800
	shrl	$9,%eax			/* %eax = size in unit of sectors */
#else
	shrl	%cl,%eax		/* %eax = size in unit of sectors */
#endif
	incl	%eax			/* for fractional portion
					 * (this will be sometimes one more
					 *  sector than needed, but that is Ok)
					 */
	movl	%edi,-8(%edi)		/* set BLK_CUR_BLKLIST */
	movl	%eax,4(%edi)		/* BLK_BLKLENGTH(BLK_BLKLIST_START) */

#if defined(TEST) || (defined(PC9800) && defined(DEBUG))
	testb	$ DEBUG_FILESYSTEM,EXT_C(debug_flags)
	jz	0f
	pushl	%eax
	pushl	%edx
	pushl	ISO_DR_SIZE(%esi)
	pushl	32(%esp)
	pushl	$1f
	call	EXT_C(dprintf)
	addl	$20,%esp

	.section .rodata
    1:	.string	"iso9660: %s: %u bytes, %u+%u\n"
	.previous
    0:
#endif
L(dir_success):
	xorl	%eax,%eax
	incl	%eax			/* %eax = 1 */
	jmp	L(dir_exit)

	ALIGN_JUMP
L(bad_filetype):
	movb	$ ERR_BAD_FILETYPE,EXT_C(errnum)
L(dir_failed_pop2):
	popl	%edx			/* discard extent */
	popl	%eax			/* discard size */
L(dir_failed):
	xorl	%eax,%eax

	ALIGN_JUMP
L(dir_exit):
	popl	%ebx
	popl	%esi
	popl	%edi
	popl	%ebp
	ret

/*
 * Local variables:
 * compile-command: "gcc -DTEST -DPC9800 -Wa,-a=/tmp/fsys_iso9660.list -o /tmp/fsys_iso9660.o -c fsys_iso9660.S"
 * End:
 */
