/*
 *  Debugging stuff for GRUB/98 (GRand Unified Bootloader for NEC PC-9800)
 *
 *  Copyright (C) 1999  Kousuke Takai  <tak@kmc.kyoto-u.ac.jp>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 *  This file is intended to be #include'd from ``cmdline.c'',
 *  so that it cannot compile by itself.
 */

static __inline__ void
list_geometries(void)
{
  int i;
  struct disk_geometry * cg = get_geometry(current_drive);
  struct disk_geometry * bg = get_geometry(boot_drive);

  printf(MSG2("\
 No  DA/UA  sector size  heads  sectors/track  total sectors\n\
 --  -----  -----------  -----  -------------  -------------\n", "\
 No  DA/UA    ͪ  ĪתΪ   \n\
 --  -----  -------  -----  -----------------  ------------\n"));
  for (i = 0 ; i < 19 ; i++)
    if (geometry_table[i].da_ua != 0xFF)
      {
	printf(MSG2("%3u   %2xh%14u%7u%15u%15u",
		    "%3u   %2xh%10u%7u%19u%14u"), i,
	       geometry_table[i].da_ua,
	       (geometry_table[i].log2_sector_size == 0xFF
		? 0 : SECTOR_SIZE(&geometry_table[i])),
	       geometry_table[i].heads, geometry_table[i].sectors,
	       geometry_table[i].total_sectors);
	if (&geometry_table[i] == bg)
	  printf(MSG2("  (boot)", "  (̎ގ)"));
	if (&geometry_table[i] == cg)
	  printf(MSG2("  (current)", "  (ڎݎ)"));
	putchar('\n');
      }
}

#define NIBBLE(x)							\
	({ int _ax;							\
	   __asm__ volatile ("cmp%B0 %1,%b0\n	"			\
			     "sbb%B0 %2,%b0\n	"			\
			     "das\n	"				\
			     "call %P3"					\
			     : "=a" (_ax)				\
			     : "i"(10), "i"(0x69), "g"(putchar),	\
			       "a"((unsigned char)((x) & 0xF))); })

static __inline__ void
dump_memory(void)
{
  static int start = 0;
  int len = 128, i, j, wait;
  unsigned char buf[16];

  if (*cur_cmdline
      && safe_parse_maxint(&cur_cmdline, &start)
      && *cur_cmdline)
    safe_parse_maxint(&cur_cmdline, &len);

  if (!errnum)
    {
      wait = (len > 16 * 20) ? 0 : -1;

      while (len)
	{
	  printf("%8x ", (start & ~0xF));
	  for (i = (start & 0xF) ; i ; i--)
	    printf("   ");
	  do
	    {

	      putchar((!(start & 7) && (start & 8)) ? '-' : ' ');
	      buf[i] = *(volatile unsigned char *)start++;
	      NIBBLE(buf[i] >> 4);
	      NIBBLE(buf[i]);
	      i++;
	    }
	  while (--len && (start & 0xF));

	  j = 18 - i + ((-start) & 0xF) * 2;
	  do
	    putchar(' ');
	  while (--j);
	  do
	    {
	      /*
	       *  EUC Single Shift 2 escapes all following characters
	       *  *for our putchar()*.
	       */
	      putchar('\x8E');
	      putchar(buf[j]);
	    }
	  while (++j < i);
	  putchar('\n');
	  if (len && wait >= 0 && ++wait == 16)
	    {
	      char c;
	      wait = 0;
	      putchar(':');
	      if ((c = ASCII_CHAR(getkey())) == 3 || c == 27)
		{
		  printf(MSG2("(abort)\n", "()\n"));
		  break;
		}
	      putchar_x--;
	    }
	}
    }
}

static __inline__ void
poke_memory(const char *cur_heap)
{
  int address, value;

  if (safe_parse_maxint(&cur_cmdline, &address))
    while (*(cur_cmdline = skip_to(0, cur_cmdline)))
      {
	if (!safe_parse_maxint(&cur_cmdline, &value))
	  break;
#if 0
	if (cur_heap[4] == 'l')
	  __asm__("stosl" : "=D"(address) : "a"(value), "0"(address));
	else if (cur_heap[4] == 'w')
	  __asm__("stosw" : "=D"(address) : "a"(value), "0"(address));
	else
	  __asm__("stosb" : "=D"(address) : "a"(value), "0"(address));
#else
	__asm__("cmp%B1 %2,%1\n\t"
		"je 1f\n\t"
		"cmp%B1 %3,%1\n\t"
		"je 2f\n\t"
		"stosb\n\t"
		"jmp 3f\n"
		"2:	.byte 0x66\n"
		"1:	stosl\n"
		"3:"
		: "=D"(address)
		: "q"(cur_heap[4]), "i"('l'), "i"('w'),
		  "a"(value), "0"(address));
#endif
      }
}

static __inline__ void
change_debug_flag(void)
{
  if (*cur_cmdline)
    {
      do
	{
	  unsigned char mask;

	  if (cur_cmdline[0] != '+' && cur_cmdline[0] != '-')
	    {
	      errnum = ERR_UNRECOGNIZED;
	      break;
	    }
	  if (cur_cmdline[1] == 'a')
	    mask = ~0;
	  else
	    {
	      const char *ptr = "becfgs";
	      mask = 1;
	      do
		{
		  if (cur_cmdline[1] == *ptr)
		    break;
		  mask <<= 1;
		}
	      while (*(++ptr));
	      if (!*ptr)
		{
		  errnum = ERR_UNRECOGNIZED;
		  break;
		}
	    }
	  debug_flags &= ~mask;
	  if (cur_cmdline[0] == '+')
	    debug_flags |= mask;
	}
      while (*(cur_cmdline = skip_to(0, cur_cmdline)));
    }
  else
    {
      const char * ptr =
	"\011Tracing biosdisk()\0"
	"\013Disk BIOS error\0"
	"\006Disk change\0"
	"\001Floppy detection\0"
	"\001Geometry probe\0"
	"\005Filesystem\0";
      unsigned char flags = debug_flags;

      do
	{
	  unsigned char highlight_x = (*ptr++) - 1;

	  do
	    {
	      if (putchar_x == highlight_x)
		putchar_attribute = 0xC1;
	      putchar(*ptr);
	      putchar_attribute = 0xE1;
	    }
	  while (*(++ptr));
	  do
	    putchar(' ');
	  while (putchar_x < 32);
	  printf("= o%s\n", (flags & 1) ? "n" : "ff");
	  flags >>= 1;
	}
      while (*(++ptr));
    }
}

static __inline__ void
message_command(char *cur_heap)
{
  if (open(cur_cmdline))
    {
      char c;

      while (read(&c, sizeof(c)))
	{
	  if (c == '\f')
	    {
	      gotoxy(12, 24);
	      printf(MSG2("Press any key to continue or ESC key to abort: ",
			  "򲡤³ɽޤ (ESC: ) "));
	      if ((c = getc()) == 0x1B || c == 0x03)
		break;
	      cls();
	    }
	  else if(c != '\r')
	    {
	      if (cur_heap[7] == 's')
		putchar_flags = PUTCHAR_SJIS;
	      putchar(c);
	    }
	}
      *(unsigned short *)&putchar_flags = 0;
    }
}

static void
load_func (char *arg)
{
  int sector;

  if (open (arg) && read (LINUX_STAGING_AREA, -1))
    return 1;
  return 0;
}
