/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  Generic defines to use anywhere
 */

/*
 *  Integer sizes
 */

#define MAXINT     0x7FFFFFFF

/*
 *  This is the location of the raw device buffer.  It is 31.5K
 *  in size.
 */

#define BUFFERSEG    0x7000
#define BUFFERADDR   0x70000
#ifndef	PC9800
/* 512-byte scratch area */
#define SCRATCHSEG   0x77e0
#define SCRATCHADDR  0x77e00
#else
#define	BUFFERSIZE	0x8000
#endif

/*
 *  BIOS disk defines
 */
#ifndef	PC9800
# define BIOSDISK_SUBFUNC_READ       0x2
# define BIOSDISK_SUBFUNC_WRITE      0x3
#else
# define BIOSDISK_SUBFUNC_READ		0x06
# define BIOSDISK_SUBFUNC_WRITE		0x05
# define BIOSDISK_ERROR_NOTREADY	0x60
# define BIOSDISK_ERROR_UNITATTENSION	0x68
# define BIOSDISK_ERROR_TIMEOUT		0x90
#endif
#define BIOSDISK_ERROR_GEOMETRY     0x100

/*
 *  This is the location of the filesystem (not raw device) buffer.
 *  It is 32K in size, do not overrun!
 */

#define FSYS_BUF 0x68000

/*
 *  Linux setup parameters
 */

#define LINUX_STAGING_AREA        0x100000
#define LINUX_SETUP               0x90000
#define LINUX_SETUP_MAXLEN        0x1E00
#define LINUX_KERNEL              0x10000
#define LINUX_KERNEL_MAXLEN       0x7F000
#define LINUX_SETUP_SEG           0x9020
#define LINUX_INIT_SEG            0x9000
#define LINUX_KERNEL_LEN_OFFSET   0x1F4
#define LINUX_SETUP_LEN_OFFSET    0x1F1
#define LINUX_SETUP_STACK         0x3FF4

#define LINUX_SETUP_LOADER        0x210
#define LINUX_SETUP_LOAD_FLAGS    0x211
#define LINUX_FLAG_BIG_KERNEL     0x1
#define LINUX_SETUP_CODE_START    0x214
#define LINUX_SETUP_INITRD        0x218

#define CL_MY_LOCATION  0x92000
#define CL_MY_END_ADDR  0x920FF
#define CL_MAGIC_ADDR   0x90020
#define CL_MAGIC        0xA33F
#define CL_BASE_ADDR    0x90000
#define CL_OFFSET       0x90022

/*
 *  General disk stuff
 */

#ifndef	PC9800
#define SECTOR_SIZE          0x200

#define BIOS_FLAG_FIXED_DISK 0x80

#define BOOTSEC_LOCATION     0x7C00
#else

#define	MAX_SECTOR_SIZE		2048
#endif
#define BOOTSEC_SIGNATURE    0xAA55
#define BOOTSEC_BPB_OFFSET   0x3
#define BOOTSEC_BPB_LENGTH   0x3B
#define BOOTSEC_PART_OFFSET  0x1BE
#define BOOTSEC_PART_LENGTH  0x40
#define BOOTSEC_SIG_OFFSET   0x1FE

/*
 *  GRUB specific information
 *    (in LSB order)
 */

#define COMPAT_VERSION_MAJOR 2
#define COMPAT_VERSION_MINOR 0
#define COMPAT_VERSION       ((COMPAT_VERSION_MINOR<<8)|COMPAT_VERSION_MAJOR)

#ifndef PC9800
#define STAGE1_VER_MAJ_OFFS	0x1bc
#define STAGE1_INSTALLSEG	0x1ba
#define STAGE1_INSTALLADDR	0x1b8
#define STAGE1_FIRSTLIST	0x1b5
#else
#define STAGE1_VER_MAJ_OFFS	0x1F6
#define STAGE1_INSTALLSEG	0x1F4
#define STAGE1_INSTALLADDR	0x1F2
#define STAGE1_FIRSTLIST	0x1F0
#endif
#define	STAGE1_SIZE		0x200

#define STAGE2_VER_MAJ_OFFS  0x6
#define STAGE2_INSTALLPART   0x8
#ifndef	PC9800
#define STAGE2_VER_STR_OFFS  0xc
#else
#define	STAGE2_SIZE_OFFS	0xC
#define	STAGE2_VER_STR_OFFS	0x10
#endif

/*
 *  defines for use when switching between real and protected mode
 */

#define data32		.byte 0x66
#define addr32		.byte 0x67
#define CR0_PE_ON	0x1
#define CR0_PE_OFF	0xfffffffe
#define PROT_MODE_CSEG	0x8
#define PROT_MODE_DSEG  0x10
#define PSEUDO_RM_CSEG	0x18
#define PSEUDO_RM_DSEG	0x20
#define STACKOFF	0x2000 - 0x10
#define PROTSTACKINIT   FSYS_BUF - 0x10

/*
 * Assembly code defines
 *
 * "EXT_C" is assumed to be defined in the Makefile by the configure
 *   command.
 */

#ifndef	EXT_C	/* fallback defines by K.Takai, Feb-Mar 1999 */
# ifdef	__ELF__
#  define EXT_C(sym)	sym
# else	/* traditional a.out */
#  define EXT_C(sym)	_ ## sym
# endif
#endif

#ifdef __ELF__
# define L(sym)	.L ## sym
#else
# define L(sym) L ## sym
#endif

#define ENTRY(x) .globl EXT_C(x) ; EXT_C(x) ## :
#define VARIABLE(x) ENTRY(x)

/* ModR/M bytes */
#define	MODRM(mod,spare,r_m) (((mod)&3)<<6 | ((spare)&7) <<3 | ((r_m)&7))

#ifndef	PC9800

#define K_RDWR  	0x60		/* keyboard data & cmds (read/write) */
#define K_STATUS	0x64		/* keyboard status */
#define K_CMD		0x64		/* keybd ctlr command (write-only) */

#define K_OBUF_FUL 	0x01		/* output buffer full */
#define K_IBUF_FUL 	0x02		/* input buffer full */

#define KC_CMD_WIN	0xd0		/* read  output port */
#define KC_CMD_WOUT	0xd1		/* write output port */
#define KB_OUTPUT_MASK  0xdd		/* enable output buffer full interrupt
					   enable data line
					   enable clock line */
#define KB_A20_ENABLE   0x02

#else	/* PC9800 */

#define	PUTCHAR_SJIS		1	/* use Shift-JIS coding system     */
#define	PUTCHAR_XZENKAKU	2	/* inhibit displaying zenkaku char */

#define KEY_ESC		0x001B
#define KEY_ENTER	0x1C0D
#define	KEY_DELETE	0x3900
#define	KEY_UP		0x3A00
#define	KEY_LEFT	0x3B00
#define	KEY_RIGHT	0x3C00
#define	KEY_DOWN	0x3D00
#define	KEY_HOMECLR	0x3E00
#define	KEY_HELP	0x3F00
#define	KEY_HOME	0x5E00

#ifdef	ASM_FILE
# define __C(type, address)	address
#else
# define __C(type, address)	((type *)address)
typedef	unsigned char	Byte;
typedef	unsigned short	Word;
struct disk_result { Byte st0, st1_pcn, st2, c, h, r, n, ncn; };
struct f2dd_result { Byte st0, pcn; };
#endif

#define	EXPMMSZ		__C(Byte, 0x0401)
#define	SCSI_PARAM	__C(Byte, 0x0460)
#define	DISK_EQUIPS	__C(Byte, 0x0482)
#define	DISK_EQUIP2	__C(Byte, 0x0494)
#define DISK_XROM	__C(Byte, 0x04B0)
#define	BIOS_FLAG	__C(Byte, 0x0501)
#define	KB_KY_STS	__C(Byte, 0x052A)
#define	KB_SHFT_STS	__C(Byte, 0x053A)
#define	DISK_EQUIP	__C(Word, 0x055C)
#define	DISK_INT	__C(Word, 0x055E)
#define	DISK_RESULT	__C(struct disk_result, 0x0564)
#define	DISK_BOOT	__C(Byte, 0x0584)
#define	EXPMMSZ16M	__C(Word, 0x0594)
#define	F2DD_RESULT	__C(struct f2dd_result, 0x05D8)

#define	GEOMETRY_TABLE		0x0600
#define	DRIVE_LOOKUP_TABLE	0x0700

#define DEBUG_BIOSDISK		0x01	/* trace calling biosdisk() */
#define DEBUG_BIOS_ERROR	0x02	/* verbose disk BIOS error reporting */
#define DEBUG_DISK_CHANGE	0x04	/* display disk change status */
#define DEBUG_FLOPPY_PROBE	0x08	/* display floppy probe status */
#define DEBUG_GEOMETRY		0x10	/* display raw SCSI geometry */
#define DEBUG_FILESYSTEM	0x20	/* report file-system internals */
#define DEBUG_ATTRIBUTE		0xA1	/* text attribute for debug message */

#define MSG	PC9800

#endif	/* PC9800 */

#ifndef	MSG
# define MSG	0
#endif
#if MSG == 0
# define MSG2(msg0, msg1)	msg0
# define MSG3(msg0, msg1, msg2)	msg0
#elif MSG == 1
# define MSG2(msg0, msg1)	msg1
# define MSG3(msg0, msg1, msg2)	msg1
#else
# define MSG2(msg0, msg1)	msg1
# define MSG3(msg0, msg1, msg2)	msg2
#endif

#define ERR_NONE            0
#define ERR_WONT_FIT        (ERR_NONE + 1)
#define ERR_NO_DISK         (ERR_WONT_FIT + 1)
#define ERR_READ            (ERR_NO_DISK + 1)
#define ERR_WRITE           (ERR_READ + 1)
#define ERR_GEOM            (ERR_WRITE + 1)
#define ERR_OUTSIDE_PART    (ERR_GEOM + 1)
#define ERR_BAD_PART_TABLE  (ERR_OUTSIDE_PART + 1)
#define ERR_NO_PART         (ERR_BAD_PART_TABLE + 1)
#define ERR_BAD_FILENAME    (ERR_NO_PART + 1)
#define ERR_BAD_FILETYPE    (ERR_BAD_FILENAME + 1)
#define ERR_FILE_NOT_FOUND  (ERR_BAD_FILETYPE + 1)
#define ERR_FSYS_MOUNT      (ERR_FILE_NOT_FOUND + 1)
#define ERR_FSYS_CORRUPT    (ERR_FSYS_MOUNT + 1)
#define ERR_FILELENGTH      (ERR_FSYS_CORRUPT + 1)
#define ERR_NUMBER_PARSING  (ERR_FILELENGTH + 1)
#define ERR_DEV_FORMAT      (ERR_NUMBER_PARSING + 1)
#define ERR_DEV_VALUES      (ERR_DEV_FORMAT + 1)
#define ERR_EXEC_FORMAT     (ERR_DEV_VALUES + 1)
#define ERR_BELOW_1MB       (ERR_EXEC_FORMAT + 1)
#define ERR_BOOT_FEATURES   (ERR_BELOW_1MB + 1)
#define ERR_BOOT_FAILURE    (ERR_BOOT_FEATURES + 1)
#define ERR_NEED_MB_KERNEL  (ERR_BOOT_FAILURE + 1)
#define ERR_NEED_LX_KERNEL  (ERR_NEED_MB_KERNEL + 1)
#define ERR_BOOT_COMMAND    (ERR_NEED_LX_KERNEL + 1)
#define ERR_UNRECOGNIZED    (ERR_BOOT_COMMAND + 1)
#define ERR_BAD_GZIP_HEADER (ERR_UNRECOGNIZED + 1)
#define ERR_BAD_GZIP_DATA   (ERR_BAD_GZIP_HEADER + 1)
#define ERR_BAD_VERSION     (ERR_BAD_GZIP_DATA + 1)
/* GNU GRUB compatible error-codes follow.  */
#define ERR_SYMLINK_LOOP    (ERR_BAD_VERSION + 1)
#define ERR_BAD_ARGUMENT    (ERR_SYMLINK_LOOP + 1)
#define MAX_ERR_NUM         (ERR_BAD_ARGUMENT + 1)

#ifndef ASM_FILE

/*
 *  Below this should be ONLY defines and other constructs for C code.
 */


/* added by K.Takai */

#define _REGPARM1	__attribute__ ((__regparm__ (1)))
#define _REGPARM2	__attribute__ ((__regparm__ (2)))

/*
 *  typedefs to match assembler's `.long', etc. pseudo ops.
 */

/*
typedef	unsigned char	Byte;
typedef	unsigned short	Word;
*/
typedef	unsigned int	Long;

#define MAKEWORD(byte...)	(((union { Byte b[2]; Word w; }){{ byte }}).w)
#define MAKELONG(byte...)	(((union { Byte b[4]; Long l; }){{ byte }}).l)

/* added portion by K.Takai ends here */

/* Modified by K.Takai, Feb-Mar 1999.
 * - Excessive underscores to keep namespace clean
 * - Use short immediate form of I/O instruction
 */

static __inline__ unsigned char
inb(unsigned short __port)
{
        unsigned char __data;

        __asm__ __volatile__("inb %1,%0" : "=a" (__data) : "Nd" (__port));
        return __data;
}

static __inline__ void
outb(unsigned short __port, unsigned char __val)
{
        __asm__ __volatile__("outb %0,%1" : :"a" (__val), "Nd" (__port));
}

/* multiboot stuff */

#include "mb_header.h"
#include "mb_info.h"

extern char end[];    /* will be the end of the bss */

/* this function must be called somewhere... */
void cmain(void) __attribute__ ((noreturn));

#define NULL         ((void *) 0)

/*
 *  From "asm.S"
 */

extern unsigned long install_partition;
#ifndef PC9800
extern unsigned long boot_drive;
#else
#define	boot_drive	(*DISK_BOOT)

/*
 *  INT 09h (hardware interrupt from keyboard) hooking stuff
 *
 *  This will emulate IBM PC's CTRL+GRPH+DEL.
 *
 *  Note: While grub is in 32-bit protected mode, all maskable interrupts
 *	  are disabled. So we can alter interrupt vector in simple way.
 */
extern unsigned char int09h_handler[];
extern unsigned long int09h_orig_handler;
#define	INTERRUPT_VECTOR	((unsigned long *)0)

#define	hook_int09h()	__asm__("xchg%L0 %0,%1"			\
				: "=r"(int09h_orig_handler)	\
				: "m"(INTERRUPT_VECTOR[9]),	\
				  "0"(int09h_handler)		\
				: "memory")
#define	restore_int09h()(INTERRUPT_VECTOR[9] = int09h_orig_handler)

#endif
extern char version_string[];
extern char config_file[];

#ifdef PC9800
void panic(const char *, ...) __attribute__((noreturn,format(printf,1,2)));
# ifdef DEBUG
extern unsigned char debug_flags;
void dprintf(const char *, ...) __attribute__((format(printf,1,2)));
#  define DPRINTF(flag, fmt, arg...) \
	({ if (debug_flags & (flag)) dprintf(fmt , ## arg ); })
# else
#  define DPRINTF(flag, fmt, arg...) ((void)0)
# endif

extern unsigned char force_check_disk_change;
#endif

void stop(void) __attribute__ ((noreturn));

/* calls for direct boot-loader chaining */
#ifndef PC9800
void chain_stage1(int segment, int offset, int part_table_addr)
     __attribute__ ((noreturn));
#else
void chain_stage1(void) __attribute__((noreturn));
#endif

#if 0	/* by K.Takai */
void chain_stage2(int segment, int offset) __attribute__ ((noreturn));
#else
static __inline__ void
chain_stage2(int segment, int offset)
{
  extern char prot_to_real[];

  *(long *)(segment * 16 + offset + STAGE2_INSTALLPART) = install_partition;

  __asm__("call %P0\n\t"
	  ".code16\n\t"
	  "mov%B1 %1,%%al\n\t"
	  "ljmp %2,%3\n\t"
	  ".code32"
	  : /* no output (and no return) */
	  : "g"(prot_to_real), "m"(*DISK_BOOT),
	    "g"(segment), "g"(offset));
}
#endif

/* do some funky stuff, then boot linux */
void linux_boot(void) __attribute__ ((noreturn));
void big_linux_boot(void) __attribute__ ((noreturn));

/* booting a multiboot executable */
void multi_boot(int start, int mbi) __attribute__ ((noreturn));

#ifndef	PC9800

/* sets it to linear or wired A20 operation */
void gateA20(int linear);

/* memory probe routines */
int get_memsize(int type);
int get_eisamemsize(void);
int get_mmap_entry(int buf, int cont);
#else /* PC9800 */

/* sets it to linear or wired A20 operation */
static __inline__ void gateA20(int __linear)
{
	if( __linear )
		__asm__("outb %%al,%0" : : "N"(0xF2));
	else
		__asm__("out%B0 %b0,%1" : : "a"((char)0x03), "N"(0xF6));
}

#endif /* PC9800 */

/* low-level timing info */
int getrtsecs(void);

/* low-level character I/O */
void cls(void);
#ifndef	PC9800
int getxy(void);     /* returns packed values, LSB+1 is x, LSB is y */
void gotoxy(int x, int y);

/* displays an ASCII character.  IBM displays will translate some
   characters to special graphical ones */
#define DISP_UL         218
#define DISP_UR         191
#define DISP_LL         192
#define DISP_LR         217
#define DISP_HORIZ      196
#define DISP_VERT       179
#define DISP_LEFT       0x1b
#define DISP_RIGHT      0x1a
#define DISP_UP         0x18
#define DISP_DOWN       0x19

void putchar(int c);
#else

extern volatile unsigned char putchar_x, putchar_y, putchar_save;
extern unsigned char putchar_flags, putchar_attribute;
extern unsigned int vram;
#define	VRAM_ADDRESS(x, y) (vram + ((x) + (y) * 80) * 2)

extern void putchar(int) __attribute__((regparm(1)));

/* putchar() does not clobber any register */
#define	putchar(c) __asm__("call %P0" : : "g"(putchar), "a"((char)(c)))

#define	getxy()	(*(unsigned short *)&putchar_y)
#define	gotoxy(x, y) (__builtin_constant_p(x) && __builtin_constant_p(y) ?    \
		      (*(unsigned long *)&putchar_y = (y) | ((x) << 8)) :     \
		      __builtin_constant_p(y) ?				      \
		      (*(unsigned long *)&putchar_y = (y), putchar_x = (x)) : \
		      (putchar_x = (x), putchar_y = (y),		      \
		       *(unsigned short *)&putchar_flags = 0))

#define DISP_UP		0x1E
#define DISP_DOWN	0x1F
#define DISP_HORIZ	0x95
#define DISP_VERT	0x96
#if 1
# define DISP_UL	0x98
# define DISP_UR	0x99
# define DISP_LL	0x9A
# define DISP_LR	0x9B
#else
# define DISP_UL	0x9C
# define DISP_UR	0x9D
# define DISP_LL	0x9E
# define DISP_LR	0x9F
#endif

#endif

/* returns packed BIOS/ASCII code */
#define BIOS_CODE(x)    ((x) >> 8)
#define ASCII_CHAR(x)   ((x) & 0xFF)
#ifndef	PC9800
#define KEY_LEFT        0x4B00
#define KEY_RIGHT       0x4D00
#define KEY_UP          0x4800
#define KEY_DOWN        0x5000
#define KEY_INSERT      0x5200
#define KEY_DELETE      0x5300
#define KEY_HOME        0x4700
#define KEY_END         0x4F00
#define KEY_PGUP        0x4900
#define KEY_PGDN        0x5100
#endif
int asm_getkey(void);

/* returns 0 if non-ASCII character */
#define getc()  ASCII_CHAR(getkey())

/* like 'getkey', but doesn't wait, returns -1 if nothing available */
int checkkey(void);

/* sets text mode character attribute at the cursor position */
#define ATTRIB_NORMAL    0x7
#define ATTRIB_INVERSE   0x70
void set_attrib(int attr);

/* low-level disk I/O */
#ifndef PC9800
int get_diskinfo(int drive);

int biosdisk(int subfunc, int drive, int geometry,
	     int sector, int nsec, int segment);

void stop_floppy(void);

#else /* PC9800 */

struct disk_geometry {
	unsigned char	log2_sector_size;
	unsigned char	da_ua;
	unsigned char	sectors;
	unsigned char	heads;
	unsigned int	total_sectors;
};

#define GEOMETRY_DRIVE_EXIST_P(geom)	((geom) && (geom)->da_ua != 0xFF)
#define GEOMETRY_IDE_P(geom)		(((geom)->da_ua & 0x70) == 0x00)
#define GEOMETRY_SCSI_P(geom)		(((geom)->da_ua & 0x70) == 0x20)
#define GEOMETRY_CHS_AVAIL_P(geom)	(((geom)->da_ua & 0x80) != 0x00)

int biosdisk(int __ah, struct disk_geometry * __geom,
	     int __sector, int __nsec, int __segment);

#define DAUA_HARDDISK_P(daua)	(((daua) & 0x50) == 0)
#define DAUA_FLOPPY_P(daua)	({ Byte __da = (daua) & 0x70; \
			(__da == 0x10 || __da == 0x30 || __da == 0x70); })
#define DRIVE_HARDDISK_P(drive)	DAUA_HARDDISK_P((drive))

#if 0
#define	NSECTORS_SIZE(geom, n)		((n) << (geom)->log2_sector_size)
#define	TO_NR_SECTORS(geom, size)	((size) / SECTOR_SIZE(geom))
#else /* GCC code generating workaround */
#define NSECTORS_SIZE(geom, n) ({					\
	unsigned int __s;						\
	__asm__("shl%L0 %b1,%0"						\
		: "=r"(__s) : "c"((geom)->log2_sector_size), "0"((n)));	\
	__s;							\
})
#define	TO_NR_SECTORS(geom, size) ({					   \
	unsigned int __n;						   \
	__asm__("shr%L0 %b1,%0"						   \
		: "=r"(__n) : "c"((geom)->log2_sector_size), "0"((size))); \
	__n;								   \
})
#endif
#define	SECTOR_SIZE(geom)		NSECTORS_SIZE(geom, 1U)
#define SECTOR_SIZE_SHIFT(geom)	((geom)->log2_sector_size)
#define	TOTAL_SECTORS(geom)	((geom)->total_sectors)

#define __REGPARM	__attribute__((regparm(1)))

extern int get_bios_geometry(struct disk_geometry *__geom)	__REGPARM;
extern int read_capacity(struct disk_geometry *__geom)		__REGPARM;
extern int test_unit_ready(struct disk_geometry *__geom)	__REGPARM;
extern void check_interrupt(void);
extern int simple_int1B(int __eax)				__REGPARM;
extern struct disk_geometry * get_geometry(int __drive)		__REGPARM;

extern struct disk_geometry geometry_table[256 / sizeof(struct disk_geometry)];
extern unsigned char	    drive_lookup_table[256];

#define	geometry_table	((struct disk_geometry *)GEOMETRY_TABLE)
#define GEOMETRY_BASE	(GEOMETRY_TABLE / sizeof(struct disk_geometry))
#define drive_lookup_table	((unsigned char *)DRIVE_LOOKUP_TABLE)

#define	MAX_NR_GEOMETRY		(256 / sizeof(struct disk_geometry))

#define	FD1MB_GEOMETRY_INDEX	(0)
#define	MAX_NR_FD1MB		(4)
#define	FD640KB_GEOMETRY_INDEX	(0 + MAX_NR_FD1MB)
#define	MAX_NR_FD640KB		(4)
#define	IDE_GEOMETRY_INDEX	(FD640KB_GEOMETRY_INDEX + MAX_NR_FD640KB)
#define	MAX_NR_IDE		(4)
#define	SD_GEOMETRY_INDEX	(HD_GEOMETRY_INDEX + MAX_NR_IDE)
#define	MAX_NR_SD		(7)
#define HD_GEOMETRY_INDEX	IDE_GEOMETRY_INDEX
#define MAX_NR_HD		(MAX_NR_IDE + MAX_NR_SD)

extern struct disk_geometry * current_geometry;

#endif	/* PC9800 */

/*
 *  From "cmdline.c"
 */

#ifndef _CMDLINE_C

extern int fallback;
extern char *password;
extern char commands[];

#endif  /* _CMDLINE_C */

enum kernel_type
  {
    KERNEL_TYPE_NONE		= 0,
    KERNEL_TYPE_MULTIBOOT	= 'm',
    KERNEL_TYPE_LINUX		= 'l',
    KERNEL_TYPE_BIG_LINUX	= 'L',
    KERNEL_TYPE_FREEBSD		= 'f',
    KERNEL_TYPE_NETBSD		= 'n',
    KERNEL_TYPE_CHAINLOADER	= 'c',
    KERNEL_TYPE_PTEDITOR	= 'p',
  };

extern enum kernel_type kernel_type;

#if 0	/* by K.Takai  Mar. 22 1999 */
char *skip_to(int after_equal, char *cmdline);
#else
# if 0
#define skip_to(after_equal, cmdline) \
	__skip_to_asm((cmdline),((after_equal) ? '=' : ' '))
extern char *__skip_to(char *, char) __attribute__((regparm(2)));
#define __skip_to_asm(cmdline, delim)	({				\
	register char *__p;						\
	register int _cx, _dx;						\
	__asm__ ("call %P3"						\
		 : "=a"(__p), "=c" (_cx), "=d" (_dx)			\
		 : "g"(__skip_to), "0"((cmdline)), "2"((char)(delim)));	\
	__p;								\
})
# else
#define skip_to(after_equal, cmdline) \
	((after_equal) ? first_arg ((cmdline)) : next_arg ((cmdline)))
extern char *first_arg (const char *) _REGPARM1;
extern char *next_arg (const char *) _REGPARM1;
# endif
#endif

void init_cmdline(void);
int enter_cmdline(char *script, char *heap);


/*
 *  From "char_io.c"
 */

#ifndef _CHAR_IO_C

/* int special_attribute; */

#endif  /* _CHAR_IO_C */

/* Same as memset in libc except for return value.
   Address is NOT checked unlike bzero.  */
static __inline__ void
memset (void *start, int c, unsigned int n)
{
  int ecx;

  c = (c & 0xFF) * 0x01010101;
  asm volatile ("cld\n	rep\n	stos%L2"
		: "=D" (start), "=c" (ecx)
		: "a" (c), "0" (start), "1" (n >> 2));
  if (n & 2)
    asm volatile ("stos%W1" : "=D" (start) : "a" (c), "0" (start));
  if (n & 1)
    asm volatile ("stos%B1" : "=D" (start) : "a" (c), "0" (start));
}

static __inline__ void
memset4 (void *start, int c, unsigned int n)
{
  asm volatile ("cld\n	rep\n	stos%L2"
		: "=D" (start), "=c" (n)
		: "a" ((c * 0xFF) * 0x01010101), "0" (start), "1" (n >> 2));
}

/* returns packed BIOS/ASCII code */
#define BIOS_CODE(x)    ((x) >> 8)
#define ASCII_CHAR(x)   ((x) & 0xFF)
#ifndef	PC9800
#define KEY_LEFT        0x4B00
#define KEY_RIGHT       0x4D00
#define KEY_UP          0x4800
#define KEY_DOWN        0x5000
#define KEY_INSERT      0x5200
#define KEY_DELETE      0x5300
#define KEY_HOME        0x4700
#define KEY_END         0x4F00
#define KEY_PGUP        0x4900
#define KEY_PGDN        0x5100
#endif
int getkey(void);  /* actually just calls asm_getkey and invalidates the
		      disk buffer */

void init_page(void);
void print_error(void);
char *convert_to_ascii(char *buf, int c, ...);
void printf(const char *format, ... )	__attribute__((format(printf,1,2)));
int get_cmdline(char *prompt, char *commands, char *cmdline, int maxlen);
int tolower(int c);
int isspace(int c);
int strncat(char *s1, char *s2, int n);
int strcmp(char *s1, char *s2);
char *strstr(char *s1, char *s2);
int bcopy(const char *from, char *to, int len);
int bzero(char *start, int len);
#if 0
int get_based_digit(int c, int base);
#endif
int safe_parse_maxint(char **str_ptr, int *myint_ptr);
int memcheck(int start, int len);

/* Forward compatibility with GNU GRUB: */
int substring (const char *s1, const char *s2);
static __inline void *
memmove (void *to, const void *from, __SIZE_TYPE__ len)
{
  return bcopy (from, to, len) ? to : NULL;
}

static __inline void *
memcpy_const_l (void *dest, const void *src, __SIZE_TYPE__ len)
{
  void *d1;
  __SIZE_TYPE__ l1;

  __asm __volatile ("cld\n\trep\n\tmovsl"
		    : "=S" (src), "=D" (d1), "=c" (l1)
		    : "0" (src), "1" (dest), "2" (len / 4)
		    : "memory");
  if (len & 2)
    __asm __volatile ("movsw"
		      : "=S" (src), "=D" (d1)
		      : "0" (src), "1" (d1)
		      : "memory");
  if (len & 1)
    __asm __volatile ("movsb"
		      : "=S" (src), "=D" (d1)
		      : "0" (src), "1" (d1)
		      : "memory");
  return dest;
}

static __inline void *
memcpy_variable_l (void *dest, const void *src, __SIZE_TYPE__ len)
{
  void *d1;

  __asm __volatile ("cld\n\trep\n\tmovsb"
		    : "=S" (src), "=D" (d1), "=c" (len)
		    : "0" (src), "1" (dest), "2" (len)
		    : "memory");
  return dest;
}

#define memcpy(dest, src, len)	(__builtin_constant_p ((len))		     \
				 ? memcpy_const_l ((dest), (src), (len))     \
				 : memcpy_variable_l ((dest), (src), (len)))


/*
 *  From "gunzip.c"
 */

#ifndef _GUNZIP_C

extern int no_decompression;
extern int compressed_file;

#endif  /* _GUNZIP_C */

int gunzip_test_header(void);
int gunzip_read(void * addr, int len);


/*
 *  From "disk_io.c"
 */

#ifndef _DISK_IO_C

#ifndef NO_FANCY_STUFF
/* instrumentation variables */
extern void (*debug_fs)(int);
extern void (*debug_fs_func)(int);
#endif  /* NO_FANCY_STUFF */

extern unsigned long current_drive;
extern unsigned long current_partition;
#define	ENTIRE_DISK_P(partition)	\
	(*(unsigned short *)((char *)&(partition) + 1) == 0xFFFF)

extern int fsys_type;

#ifndef NO_BLOCK_FILES
extern int block_file;
#endif  /* NO_BLOCK_FILES */

extern long part_start;
extern long part_length;

extern int current_slice;

#ifndef PC9800
extern int buf_drive;
extern int buf_track;
extern int buf_geom;
#else
extern struct disk_geometry *buf_geom;
extern int buf_sectors;
#endif

/* these are the current file position and maximum file position */
extern int filepos;
extern int filemax;

#endif  /* _DISK_IO_C */

/* For forward compatibility with GNU GRUB... */
union int_ptr { int i; void *p; } __attribute__ ((transparent_union));

int rawread(int drive, int sector, int byte_offset, int byte_len, union int_ptr addr);
int devread(int sector, int byte_offset, int byte_len, union int_ptr addr);

char *set_device(char *device);  /* this gets a device from the string and
				    places it into the global parameters */
int open_device(void);
int make_saved_active(void);   /* sets the active partition to the that
				  represented by the "saved_" parameters */

int open(char *filename);

/* Read LEN bytes into BUF from the file that was opened with
   open().  If LEN is -1, read all the remaining data in the file.  */
int grub_read (void *buf, int len);

/* GRUB 0.5 backward compatibility */
static __inline int
read (void *addr, int len)
{
  return grub_read (addr, len);
}

int dir(char *dirname);       /* list directory, printing all completions */

int set_bootdev(int hdbias);
void print_fsys_type(void);   /* this prints stats on the currently
				 mounted filesystem */
void print_completions(char *filename); /* this prints device and filename
					   completions */
void copy_current_part_entry(int addr); /* copies the current partition data
					   to the desired address */
#ifdef PC9800
extern void print_partition_name (const void *name) _REGPARM1;
#endif

/*
 *  From "boot.c"
 */

/* for the entry address */
typedef void
(*entry_func)(int, int, int, int, int, int) __attribute__ ((noreturn));

#ifndef _BOOT_C

extern char *cur_cmdline;
extern entry_func entry_addr;

#endif  /* _BOOT_C */

void bsd_boot(int type, int bootdev) __attribute__ ((noreturn));
int load_image(void);
int load_module(void);
int load_initrd(void);


/*
 *  From "common.c"
 */


#ifndef _COMMON_C

/*
 *  Common BIOS/boot data.
 */

extern struct multiboot_info mbi;
extern unsigned long saved_drive;
extern unsigned long saved_partition;
extern unsigned long saved_mem_upper;

/*
 *  Error variables.
 */

extern int errnum;
extern char *err_list[];

#endif  /* _COMMON_C */

void init_bios_info(void) __attribute__ ((noreturn));

#endif /* ASM_FILE */
