/*
 *  GRUB  --  GRand Unified Bootloader
 *  Copyright (C) 1996   Erich Boleyn  <erich@uruk.org>
 *
 *  GRUB for NEC PC-9800 series PCs (GRUB/98)
 *  This file is the source code of GRUB/98 boot sector.
 *  Copyright (C) 1997-1999  Kousuke Takai  <tak@kmc.kyoto-u.ac.jp>
 *			     Kyoto University Microcomputer Club, Kyoto Japan.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *  defines for the code go here
 */

#define BPBEND		0x03E
#define	FIRSTLIST	0x1F0
#define	AUTO_BOOT	0x1FA	/* automatic boot partition (NEC's menu uses)*/
#define LISTSIZ		8	/* size of sector list */
#define REALSTACK	0x2000	/* stack for this code and BIOS calls */

	/* System-common areas */
#define	SCSI_PARAM	0x460	/* SCSI device parameter table */
#define	DISK_EQUIPS	0x482	/* SCSI disk existence flags */
#define	BIOS_FLAG	0x501	/* various BIOS information flags */
#define	DISK_BOOT	0x584	/* holding Device Address/Unit Address (DA/UA)
				 * of the boot device */

	.text

	/*
	 *  The ".code16" directive only works in GAS, the GNU assembler!
	 *  This adds 32-bit data or addressing directives so that this
	 *  code will work in real mode!
	 */
	.code16

	.globl	_start

_start:
	/*
	 * _start is loaded at 0x1FC00 (or 0x20000 - <sector size> * 2)
	 * and is jumped to with CS:IP 0x1FC0:0x0000
	 */

	/*
	 * Beginning of the sector is compatible with the FAT/HPFS BIOS
	 * parameter block.
	 */

	jmp	after_BPB
	nop	/* do I care about this ??? */

	.ascii	"GRUB/98 "	/* `OEM name' for FAT boot sector */
				/* this is also magic number of stage1 */

	/*
	 * This space is for the BIOS parameter block!!!!  Don't change
	 * the first jump, nor start the code anywhere but right after
	 * this area.
	 */

	. = _start + BPBEND

	/*
	 * End of BIOS parameter block.
	 */
after_BPB:
	/*
	 *  Input assumption:
	 *	%al = boot device (= DISK_BOOT)
	 *  for SCSI device:
	 *	%bx = block size x 2
	 *  when booted via other loader (especially via menu program in MBR)
	 *	%si = index to partition table
	 */

	movb	$0x07, %ah		/* use later for Recalibrate FDD */
	cwtd				/* %dx <- 0 */
	movw	%dx, %ss		/* 80x86 disables interrupts from
					 * `mov reg,%ss' until next instruction
					 * automatically. */
	movw	$REALSTACK, %sp		/* set up the REAL stack */

	pushw	%cs
	popw	%ds

	/*
	 *  Check if we have a forced disk reference here
	 */
	movw	$firstlist, %di
	cmpb	$0xFF, (%di)
	je	1f
	movb	(%di), %al
  1:	testb	$0x50, %al
	jnz	floppy_probe

	/*
	 *  %AL will be 0x8X or 0xAX.
	 *  That means we are using hard disk.
	 */
	andb	$0x7F, %al	/* force LBA addressing mode */
	jz	hd_probe_ok	/* when IDE/SASI hard disk */

	/*
	 *  When booting SCSI device, BIOS sets %BX to double of the block
	 *  size. Determine block size from it and fake the BIOS.
	 */
	movw	%ax, %bp	/* %bp = 0x0720 + (SCSI ID) */
	shlw	$3, %bx
	jz	to_probe_error
	movb	$0x30, %ah
    2:	shlw	%bx
	jc	3f
	subb	$0x10, %ah
	jnb	2b
to_probe_error:
	jmp	probe_error

    3:	jnz	to_probe_error

	shlw	$2, %bp
	movw	(SCSI_PARAM + 2 - (0x0720 << 2))(%bp), %cx
	pushw	%cx
	pushw	%bp
	andb	$0xCF, %ch
	orb	%ah, %ch
	movw	%cx, (SCSI_PARAM + 2 - (0x0720 << 2))(%bp)
hd_probe_ok:
	movb	$0x8E, %ah	/* disk BIOS `Mode Set' */
	int	$0x1B
	movb	$0x84, %ah	/* disk BIOS `New SENSE' (get geometry) */
	int	$0x1B
	jmp	final_init

floppy_probe:
	/*
	 *  Determine the floppy disk geometry
	 */
				/* (%ah = 0x07 from startup) */
	int	$0x1B		/* recalibrate the drive before probe */
	movb	$2, %ch		/* parameter for 512 bytes/sector */
	movb	$18, %dl	/* (%dh = 0x00 from startup) */
	pushw	%ax
	andb	$0x70, %al
	cmpb	$0x10, %al
	je	try_floppy
	cmpb	$0x30, %al
	je	fd1440k
	cmpb	$0x70, %al
	jne	to_probe_error
try_floppy:
	popw	%ax
	testb	%al,%al
#ifdef SUPPORT_640K
	jns	try640
#else
	movb	$9, %dl			/* 9 sectors/track (720K) */
	jns	fd_ok
#endif
try1M:	movb	$0x4A, %ah		/* `READ ID' opcode */
	int	$0x1B
	movb	$8, %dl
	cmpb	$0x03, %ch		/* 1024 b/s ? */
	je	fd_ok			/* yes -> 1232K */
	movb	$15, %dl		/* no -> 1200K */
#ifdef SUPPORT_640K
	jmp	fd_ok
try640:	movb	$0x51, %ah		/* `VERIFY' opcode */
	movb	$0x00, %cl		/* 512 b/s, cylinder 0 (%ch == 0) */
	movb	$9, %dl			/* head 0, sector 9 (%dh == 0) */
	movw	$512,%bx		/* read 512 bytes */
	int	$0x1B
	/* jnc	fd_ok */		/* Ok -> 720K */
	.byte	0x73; RELB(fd_ok)
	movb	$8, %dl			/* failed -> 640K */
#endif
	.byte	0xB4			/* opcode byte for `movb $imm8,%ah' */
fd1440k:
	popw	%ax
fd_ok:	movb	%ch, %cl
	movw	$128, %bx
	shlw	%cl, %bx		/* %bx = sector size */
	push	%dx
	push	%cx

final_init:
	pushw	%bx
	pushw	%si			/* %si holds partition information
					 * (valid only when booted from menu)
					 */
	movw	%sp, %bp

	/*
	 * Here %DI holds $firstlist, %BP holds frame pointer as following
	 */
#define	SECTSIZE	2
#define	FDCSSIZE	5	/* floppy only */
#define	NSECTORS	6	/* floppy only */

	/* this is the loop for reading the secondary boot-loader in */
bootloop:

	/* update position to load from */
	subw	$LISTSIZ, %di

	cmpw	$0, 6(%di)	/* check the number of sectors to read */

	je	finish_load	/* if zero, go to the start function */

setup_sectors:
	movw	(%di), %cx	/* load logical sector start (bottom half) */
	movw	2(%di), %dx	/* load logical sector start (top half) */

	movw	4(%di), %bx
	movw	SECTSIZE(%bp), %si
	movw	%bx, %es
	shlw	$4, %bx
	negw	%bx
	jnz	1f
	subw	%si, %bx
    1:	pushw	%bx
	shrw	$4, %bx
	cmpw	6(%di), %bx
	popw	%bx
	jbe	2f
	movw	6(%di), %bx
	shlw	$4, %bx
	decw	%si		/* round-up %bx to multiple of sector size */
	addw	%si, %bx
	notw	%si
	andw	%si, %bx
    2:	movb	$0x06,%ah	/* disk BIOS `READ DATA' opcode for HD */
	testb	$0x50,%al
	jz	3f
	movb	$0xD6,%ah	/* disk BIOS `READ DATA' opcode for FD */

	xchgw	%ax, %cx	/* %dx:%ax = logical sector start */
	divw	NSECTORS(%bp)	/* divide by number of sectors */
	shrw	%ax		/* divide by 2 (number of heads) */
	adcb	%dh, %dh	/* head # (0 - 1) */
	xchgw	%ax, %cx	/* %cx <- cylinder # */

	pushw	%ax
	movb	NSECTORS(%bp), %al
	jnz	4f		/* is head # == 0 ? */
	addb	%al, %al
    4:	subb	%dl, %al
	incw	%dx		/* %dl (sector #) is 1-base */
	mulb	SECTSIZE+1(%bp)
	cmpb	%al, %bh
	jb	5f
	movb	%al, %bh
    5:	popw	%ax
	cmpw	$80, %cx	/* do we need too many cylinders? */
	jae	geometry_error
	movb	FDCSSIZE(%bp), %ch

    3:	/*
	 *  Ok, we have following parameters prepared here:
	 *	%al = device address / unit address
	 *	%bx = transfer length
	 *	%es = transfer segment
	 *  for floppy disk:
	 *	%ah = 0xD6 (Read Data / MFM / Seek / MultiTrack)
	 *	%ch = N byte for FDC (log2(sector size) - 7)
	 *	%cl = cylinder number
	 *	%dh = head number
	 *	%dl = sector number (start from 1)
	 *  for hard disk/other devices:
	 *	%ah = 0x06 (Read Data / no error report)
	 *	%dx:%cx = logical block address
	 */

	/*
	 *  Prepare more parameters:
	 *	%bp = transfer offset address
	 *
	 *  Then call "INT 0x1B" (Disk BIOS).
	 *  Return value:
	 *	%ah < 0x30 on success; otherwise error code on failure
	 *	CF = 0 on success; 1 on failure
	 *
	 *  Registers other than %ah and %flags will not be clobbered.
	 */

	pushw	%bp
	xorw	%bp,%bp			/* %ES:%BP = transfer address */
	int	$0x1B
	popw	%bp

	jc	read_error

	movw	%bx, %si
	shrw	$4, %si			/* %si = length in unit of paragraph */
	addw	%si, 4(%di)
	subw	%si, 6(%di)
	jbe	bootloop
	cwtd				/* %dx = 0 because %ah < 0x30 */
	xchgw	%ax, %bx
	divw	SECTSIZE(%bp)		/* %ax = number of sectors just read */
	xchgw	%ax, %bx
	addw	%bx, (%di)
	adcw	$0, 2(%di)
	jmp	setup_sectors

/* END OF MAIN LOOP */

finish_load:
	popw	%si		/* restore boot partition index */
	testb	$0x50, %al
	jnz	1f
	testb	$0x20, %al
	jz	1f
	popw	%bx
	popw	%bp
	popw	(SCSI_PARAM + 2 - (0x0720 << 2))(%bp)
    1:	jmp	bootit

/*
 * BIOS Geometry translation error (past the end of the disk geometry!).
 */
geometry_error:
	movw	$geometry_error_string, %si
	.byte	0x8D		/* `.byte 0x8D; movw $<imm16>,%si' will be
				 * taken as `leaw <disp16>(%bp),%di'
				 */

/*
 * Disk probe failure.
 */
probe_error:
	movw	$probe_error_string, %si
	.byte	0x8D		/* `.byte 0x8D; movw $<imm16>,%si' will be
				 * taken as `leaw <disp16>(%bp),%di'
				 */

/*
 * Read error on the disk.
 */
read_error:
	movw	$read_error_string, %si

general_error:
	movw	$0xA000, %ax
	ss
	testb	$0x08, BIOS_FLAG
	jz	1f
	movb	$0xE0, %ah
    1:	movw	%ax, %es
	movw	$(1 * 80 + 1) * 2, %di		/* write position */
	cld
	call	message
	movw	$general_error_string, %si
	call	message

reboot_wait:
	movb	$0x17, %ah
	int	$0x18
	movb	$0, %ah
	int	$0x18
	movb	$0, %al
	outb	%al, $0xF0
/* go here when you need to stop the machine hard after an error condition */
stop:	jmp	stop

geometry_error_string:	.string "Geom"
probe_error_string:	.string "Disk Probe"
read_error_string:	.string "Read"
general_error_string:	.string " Error"

/*
 * message: write the string pointed to by %si
 *
 *   WARNING: trashes %si, %di and %ax
 */

    1:	movb	$0, %ah
	stosw
message:
	lodsb
	cmpb	$0, %al
	jne	1b
	ret

#if 0
	/* I reached here :-) */
bootit:	movw	$0x0A00, %ax
	int	$0x18
	movb	$0x16, %ah
	movw	$0xE123, %dx
	ss; movb 0x8000,%dl
	int	$0x18
	pushw	$0xA000
	popw	%es
	xorw	%ax, %ax
	xorw	%di, %di
	movw	%ax, %ds
	movw	$0x8000, %si
	movw	$160, %cx
1:	lodsb
	stosw
	loop	1b
	jmp	reboot_wait
#endif

lastlist:

/*
 *  This area is an empty space between the main body of code below which
 *  grows up (fixed after compilation, but between releases it may change
 *  in size easily), and the lists of sectors to read, which grows down
 *  from a fixed top location.
 */

        /*
         *  This data area is for keeping general parameters.
         */
#ifndef	KILL_ORG
	. = _start + FIRSTLIST - LISTSIZ - 4
#endif

	/* this next data area before the partition area is specifically
	   sized, you should update "MINPARAMSIZ" to reflect any additions
	   or deletions to this area */

	.word 0
	.word 0

        /* fill the first data listing with the default */
#ifdef  FFS_STAGE1_5
	.long	2	/* this is the sector start parameter, in logical
			   sectors from the start of the disk, sector 0 */
	.word	0x0200	/* this is the segment of the starting address
			   to load the data into */
	.word	448	/* this is the size to read in unit of paragraph */
#else
	.long	1	/* this is the sector start parameter, in logical
			   sectors from the start of the disk, sector 0 */
	.word	0x0800	/* this is the segment of the starting address
			   to load the data into */
# ifdef STAGE2_PARAGRAPHS
	.word	STAGE2_PARAGRAPHS	/* size to read in */
# else
	.word	3072	/* this is the size to read in unit of paragraph */
# endif
#endif
firstlist:	/* this label has to be after the list data!!! */

	.byte	0xFF	/* the disk to load stage2 from */
			/* 0xff means use the boot drive */

	/*
	 *  Jump here when all data loading is done.  This
	 *  goes to the second stage bootloader.
	 */

bootit:	/* ljmp	$myoffset, $myseg */
	.byte	0xEA
#ifdef  STAGE1_5
	.word	0x2000, 0
#else
	.word	0x8000, 0
#endif

	/*
	 *  This is the compatibility version number.
	 *
	 *  DO NOT MOVE THIS!!!
	 */
	.byte	2, 0

#ifndef	KILL_ORG
	. = _start + AUTO_BOOT
#endif
	.byte	0, 0		/* automatic boot */
	.byte	0		/* menu program version (seen as very old) */
	.byte	0		/* padding? */
/* the last 2 bytes in the sector 0 contain the signature */
	.word	0xAA55

/*
 * Local variables:
 * compile-command: "make -C .. all-stage1"
 * End:
 */
